/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmApplicationMac.h"

#include "EmDocumentMac.h"		// EmDocumentMac
#include "EmClassic.h"			// EmClassicInit, PPCInit
#include "EmMenusMac.h"			// HostCreateMenubar
#include "EmPatchState.h"		// EmPatchState, UIInitialized
#include "Logging.h"			// gLogMisc, LogAppendMsg
#include "Preferences.h"		// EmulatorPreferences
#include "RegisterClasses.h"	// RegisterClasses

#define GUSI_SOURCE
#include "GUSIBasics.h"			// GUSIHook, etc.
#undef GUSI_SOURCE

#if __profile__
#include <Profiler.h>
#endif


// Over in Platform.cpp (for now).
Boolean				ShouldConvertHighAscii	(void);
void 				ConvertHighAscii 		(char* d, const char* s);
void 				ConvertHighAscii 		(LStr255& str);

static bool			PrvSpinner				(bool wait);
static Bool			PrvCheckCarbon			(void);
static EmCommandID	PrvCommandT2EmCommandID (CommandT inCommand);

EmApplicationMac*	gHostApplication;

#if _DEBUG
typedef map<Ptr, Handle>	HandleMap;
HandleMap&			PrvGetHandleMap (void);	// In Platform_Mac.cpp
extern Bool			gExittedMain;
#endif

/***********************************************************************
 *
 * FUNCTION:	main
 *
 * DESCRIPTION: Application entry point.  Initializes PowerPlant,
 *				creates the application object, and enters the main
 *				event loop.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

int main (int argc, char** argv)
{
	// Perform initialization that needs to be done before creating
	// our LApplication object.

	// Initialize the Mac.

	::InitializeHeap (3); 					// Memory initialization.
	UQDGlobals::InitializeToolbox ();		// Toolbox initialization.
	UEnvironment::InitEnvironment ();		// Scope out the system.

	// Check to see if we have a sufficient version of the CarbonLib.
	// If not, PrvCheckCarbon displays an error.

	if (!::PrvCheckCarbon ())
	{
		return 0;
	}

#if __profile__
	// If we're profiling, initialize the system
	::ProfilerInit (collectDetailed, bestTimeBase, 4000, 40);
	::ProfilerSetStatus (false);
#endif

	// Set a GUSI spin hook to prevent GUSI from trying to handle
	// events.  Note that it's important that we do this very early on!
	// EmulatorPreferences will read in the preference data with fread.
	// GUSI gets its mitts into fread, and will issue an asynchronous
	// PBRead to carry it out.  While waiting for the PBRead to
	// complete, it calls the GUSISpinHook.  The default hoook will
	// call WaitNextEvent, calling AEProcessAppleEvent for any
	// high-level events.  We haven't yet installed our event handlers,
	// so any events will get tossed on the ground.  This includes our
	// 'oapp' event.  In order to not lose that event, install a spin
	// hook that doesn't try to read or handle events.
	
	::GUSISetHook (GUSI_SpinHook, (GUSIHook) ::PrvSpinner);

	// Clear out the default commander.  Otherwise, the application
	// will be a sub-commander of any statically or globally declared
	// commanders out there.  We probably don't have any of those any
	// more, but we did at one time, which is why this line is here!

	LCommander::SetDefaultCommander (NULL);

	// Enter a new scope to make sure that EmulatorPreferences and
	// EmApplicationMac are destroyed before we set gExittedMain to
	// TRUE.  Otherwise, if the destructors for those classes call
	// Platform::DisposeMemory, we might get a false assert.

	{
		// Construct the preferences.

		EmulatorPreferences	prefs;

		// Construct the application.

		EmApplicationMac	theApp;

		try
		{
			if (theApp.Startup (argc, argv))
			{
				// Run() calls MakeMenuBar, MakeModelDirector and Initialize.
				// If that all works, it calls ProcessNextEvent until the state
				// changes to something other than "processing events".

				theApp.Run ();
			}
		}
		catch (...)
		{
			::StopAlert (129, NULL);
		}

		theApp.Shutdown ();
	}

#if __profile__
	::ProfilerDump ("\pEmulator.prof");
	::ProfilerTerm ();
#endif

	// We specify LCleanupTask_Terminator as the termination function,
	// so we don't need to explicitly call CleanUpAtExit.  As well,
	// with EmApplicationMac descending from LCleanupTask, we run into
	// a problem with calling CleanUpAtExit here.  LCleanupTask
	// descends from LInterruptSafeListMember, which contains a
	// mParentList member.  LInterruptSafeListMember's destructor will
	// try to access mParentList.  However, LCleanupTask::CleanUpAtExit
	// cleans up and destroys that list, but without notifying anyone
	// containing pointers to it (as mParentList is).  Thus, accessing
	// mParentList causes a crash.
	//
	// We could explicitly scope EmApplicationMac so that it is
	// destroyed before we call CleanUpAtExit.  But we could also take
	// the current approach of calling CleanUpAtExit from
	// LCleanupTask_Terminator, by which time the application is
	// destroyed.
	//
	// Calls DoCleanup() for each registered LCleanupTask.  This should
	// execute the same stuff in the application's destructor, but do
	// only the critical stuff, and do it safely.

//	LCleanupTask::CleanUpAtExit ();

#if _DEBUG
	// !!! Not a good assert for now.  I see that there are three objects
	// in the handle map at this point.  I don't know where they're
	// from, and not sure I have the time to track them down at this
	// point.  At least whatever owns them doesn't try to clean them
	// up later after the handle map is destroyed! But I'm pretty sure
	// they're the pixel depth translation tables used in EmPixMap.

//	EmAssert (::PrvGetHandleMap ().size () == 0);
	gExittedMain = true;
#endif

	return
		gErrorHappened ? 2 :
		gWarningHappened ? 1 : 0;
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac constructor
 *
 * DESCRIPTION: Initializes data members
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

EmApplicationMac::EmApplicationMac (void) :
	LDocApplication (),
	LPeriodical (),
	LCleanupTask (),
	EmApplication (),
	fPSN (),
	fPalmUIInitialized (false)
{
	EmAssert (gHostApplication == NULL);
	gHostApplication = this;

	(void) ::GetCurrentProcess (&fPSN);
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac destructor
 *
 * DESCRIPTION: Cleans up when the application is quitting.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

EmApplicationMac::~EmApplicationMac ()
{
	EmAssert (gHostApplication == this);
	gHostApplication = NULL;
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::Startup
 *
 * DESCRIPTION: Performs one-time startup initialization.
 *
 *				Note that this is called after the EmApplicationMac
 *				constructor, but before the MakeMenuBar,
 *				MakeModelDirector and Initialize methods.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

Bool EmApplicationMac::Startup (int argc, char** argv)
{
	// Register our view classes.

	::RegisterClasses ();

	// Initialize our thunks that allow us to call non-Carbon functions
	// while still being a Carbon application.
	//
	// Do this before calling EmApplication::Startup.  That will
	// attempt to create our debugger sockets, which may need access to
	// the PPC functions in EmClassic.

	::EmClassicInit ();

	// Initialize PPCToolbox.  This may fail under Mac OS X.

	if (::EmClassicHasPPCToolbox ())
	{
		::PPCInit ();
	}

	// Initialize PPCToolbox.  This may fail under Mac OS X.

	if (::EmClassicHasSerialManager ())
	{
		CRMErr	err = ::InitCRM ();
		EmAssert (err == 0);
	}

	// Initialize the base system.

	if (!EmApplication::Startup (argc, argv))
		return false;

	// Initialize our menus.  Use different layouts depending on
	// whether or not we are running under the Aqua UI.

	::MenuInitialize (UEnvironment::HasFeature (env_HasAquaTheme));

#if (UNIVERSAL_INTERFACES_VERSION >= 0x0340)

	// Turn on the Preferences menu item in the Application menu

	::EnableMenuCommand (NULL, kHICommandPreferences);

	// Install an event handler to get the Preferences command

	EventHandlerUPP	prefsCmdHandlerUPP = ::NewEventHandlerUPP (
						EmApplicationMac::PreferencesCmdEventHandler);

	EventTypeSpec	eventTypeSpec;
	eventTypeSpec.eventClass	= kEventClassCommand;
	eventTypeSpec.eventKind		= kEventCommandProcess;

	::InstallEventHandler (::GetApplicationEventTarget (),
							prefsCmdHandlerUPP, 1,
							&eventTypeSpec, NULL, NULL);

#endif

	// Start the application's idling facilities.

	this->StartIdling ();

	return true;
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::Shutdown
 *
 * DESCRIPTION: Cleans up when the application is quitting.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::Shutdown (void)
{
	EmApplication::Shutdown ();
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::OpenOrPrintDocList
 *
 * DESCRIPTION: Respond to an AppleEvent to open or print a Document.
 *
 *				We override this method so that we can call HandleListList
 *				to examine and handle all the files at once, rather than
 *				seeing them one at a time in OpenDocument as the default
 *				OpenOrPrintDocList processes them.
 *
 * PARAMETERS:	inDocList - the direct parameter from the Apple event
 *
 *				inAENumber - ae_OpenDoc or ae_PrintDoc
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void
EmApplicationMac::OpenOrPrintDocList(
	const AEDescList&	inDocList,
	SInt32				inAENumber)
{
	// We don't have a print command.

	if (inAENumber == ae_PrintDoc)
		return;

	SInt32	numDocs;
	OSErr	err = ::AECountItems (&inDocList, &numDocs);
	ThrowIfOSErr_(err);

		// Loop through all items in the list
			// Extract descriptor for the document
			// Coerce descriptor data into a FSSpec
			// Add each spec to a list to be processed later.

	EmFileRefList	fileList;

	for (SInt32 ii = 1; ii <= numDocs; ++ii)
	{
		AEKeyword	theKey;
		DescType	theType;
		Size		theSize;
		FSSpec		theFileSpec;

		// Get the nth guy as an FSSpec.

		err = ::AEGetNthPtr (&inDocList, ii, typeFSS, &theKey, &theType,
							(Ptr) &theFileSpec, sizeof(FSSpec), &theSize);
		ThrowIfOSErr_(err);

		// Add this file to a list for later.

		EmFileRef	theFileRef (theFileSpec);
		fileList.push_back (theFileRef);
	}

	// Process all the collected files.

	this->HandleFileList (fileList);
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::OpenDocument
 *
 * DESCRIPTION: Open a Document specified by an FSSpec.  Called in
 *				response to an Open or Print Document Apple event, or
 *				if SendAEOpenDoc can't create the Apple event.
 *
 * PARAMETERS:	inMacFSSpec - spec of the file to open
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::OpenDocument (FSSpec* inMacFSSpec)
{
	EmFileRef	file (*inMacFSSpec);

	// Create the new cross-platform document object.

	this->HandleOpenFromFile (file);
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::MakeNewDocument
 *
 * DESCRIPTION: Create a new Document and pass back an AppleEvent Model
 *				object representing that Document.	Called when
 *				HandleCreateElementEvent is asked to create a cDocument,
 *				or from SendAECreateDocument.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

LModelObject* EmApplicationMac::MakeNewDocument (void)
{
	// Create the new cross-platform document object.

	EmDocument*	newDoc = this->HandleNewFromUser (NULL);

	// If we got one, fetch and return the PowerPlant document object.

	if (newDoc)
		return dynamic_cast <EmDocumentMac*> (newDoc);

	return NULL;
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::ChooseDocument
 *
 * DESCRIPTION: Allow the user to pick a Document (usually for opening).
 *				Called in response to the user selecting the "Open"
 *				menu item.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::ChooseDocument (void)
{
	// Create the new cross-platform document object.

	this->HandleOpenFromUser (kFileTypeSession);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::ShowAboutBox
 *
 * DESCRIPTION: Show our About box.  We override the standard method
 *				because we use more than a simple ALRT.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::ShowAboutBox (void)
{
	this->HandleCommand (kCommandAbout);
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::DoQuit
 *
 * DESCRIPTION: Prepare the application for quitting.
 *
 * PARAMETERS:	inSaveOption - parameter describing how to close the
 *					document.  This is kAEAsk if the use selected the
 *					Quit menu item, but can be kAEAsk, kAEYes, or
 *					kAENo if we received an Apple event.  From Apple's
 *					IAC documentation, "Factoring the Quit Command and
 *					the New Command":
 *
 *				The input to the MySendAEQuit routine is a constant that
 *				indicates whether to save dirty windows without asking
 *				the user (kAEYes), quit without saving dirty windows
 *				(kAENo), or ask the user whether each dirty window
 *				should be saved (kAEAskUser).
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::DoQuit (SInt32 /* inSaveOption */)
{
	EmApplication::DoQuit (kCommandQuit);
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::StartUp
 *
 * DESCRIPTION: StartUp is called if the application is launched without
 *				any documents.
 *
 *				Note that this is NOT the same Startup method that's
 *				defined by EmApplication.  That's "Startup", this is
 *				"StartUp".  The former gets called to perform one-time
 *				initialization of the system (much like the LApplication
 *				"Initialize" method); the latter is called when the
 *				application receives an 'oapp' Apple event.
 *
 *				We override this standard PowerPlant method.  If there
 *				was a last-used session document, try reloading it.  If
 *				not, but there is information on the last kind of
 *				created session, create a new document based on the same
 *				settings.  If either of those fails, or this is the
 *				first time the user's run the application, or the Caps
 *				Lock key is down, bring up the "New Configuration"
 *				dialog.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::StartUp (void)
{
	LDocApplication::StartUp ();

	this->HandleStartupActions ();

	if (!gDocument)
	{
		this->ObeyCommand (cmd_New, NULL);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::DoReopenApp
 *
 * DESCRIPTION: Respond to Reopen Application AppleEvent
 *
 *				The system sends the reopen application AppleEvent when
 *				the user resumes an application that has no open windows.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::DoReopenApp (void)
{
	if (!gDocument)
	{
		this->ObeyCommand (cmd_New, NULL);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::DoPreferences
 *
 * DESCRIPTION: Handle Preferences settings
 *
 *				Many programs have a "Preferences" menu item to let the
 *				user specify application-level settings. On Mac OS 9 and
 *				earlier, this item is often in the Edit menu. On Mac OS
 *				X, the system provides a Preferences item in the
 *				Application menu.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::DoPreferences (void)
{
	EmApplication::HandleCommand (kCommandPreferences);
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::MakeMenuBar
 *
 * DESCRIPTION: Create MenuBar object for this Application.
 *
 *				We override this method to create our cross-platform
 *				menus.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::MakeMenuBar (void)
{
	LDocApplication::MakeMenuBar ();

	EmMenuItemList*	menubar = ::MenuFindMenu (kMenuMenubarPreferred);
	EmAssert (menubar);

	::HostCreateMenubar (*menubar);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::ObeyCommand
 *
 * DESCRIPTION: Override of standard PowerPland method to handle
 *				commands (usually as a result of selecting a menu item).
 *
 * PARAMETERS:	inCommand - selected command
 *
 *				ioParam - any associated parameters
 *
 * RETURNED:	True if we handled the command.
 *
 ***********************************************************************/

Boolean EmApplicationMac::ObeyCommand ( CommandT	inCommand,
										void*		ioParam)
{
	// If this is a PowerPlant-defined command number, translate it
	// into a cross-platform command number.  In our menu code, we use
	// only cross-platform numbers, but parts of PowerPlant may issue
	// commands using the numbers it knows about.

	EmCommandID	id = ::PrvCommandT2EmCommandID (inCommand);

	// Let the application try to handle the command.

	Boolean	wasHandled = false;

	if (id)
	{
		wasHandled = this->HandleCommand (id);
	}

	// Otherwise, let PowerPlant try to handle the command.

	if (!wasHandled)
	{
		wasHandled = LDocApplication::ObeyCommand (inCommand, ioParam);
	}

	return wasHandled;
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::FindCommandStatus
 *
 * DESCRIPTION: Enable/disable menu items, set checkmarks, etc.
 *
 * PARAMETERS:	inCommand - command to operate on.
 *
 *				outEnabled - return true in this variable to enable the
 *					command (usually enables a menu item).
 *
 *				outUsesMark - return true here to indicate that oMark
 *					has a value in it that should be heeded.  Passing false
 *					here does *not* remove any marks.  In order to do that,
 *					pass true here and "noMark" in oMark.
 *
 *				outMark - any mark to be used on the menu item.
 *
 *				outName - name of the command in case you need to change
 *					the text of the menu item.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::FindCommandStatus (	CommandT	inCommand,
											Boolean&	outEnabled,
											Boolean&	outUsesMark,
											UInt16& 	outMark,
											Str255		outName)
{
	// If this is a PowerPlant-defined command number, translate it
	// into a cross-platform command number.  In our menu code, we use
	// only cross-platform numbers, but parts of PowerPlant may issue
	// commands using the numbers it knows about.

	EmCommandID	id = ::PrvCommandT2EmCommandID (inCommand);

	// If this is a cross-platform command, look up the menu item
	// and see if it's enabled.

	if (id)
	{
		EmMenu*		menubar		= ::MenuFindMenu (kMenuMenubarPreferred);
		EmAssert (menubar);

		EmMenuItem*	menuItem	= ::MenuFindMenuItemByCommandID (*menubar, id, true);

		if (menuItem)
		{
			outEnabled = menuItem->GetIsActive ();

			if (menuItem->GetIsChecked ())
			{
				outUsesMark = true;
				outMark = checkMark;
			}

			return;
		}
	}

	// Handle some commands that might come our way, but that aren't in
	// our menus.

	switch (inCommand)
	{
		case cmd_About:			outEnabled = true;	return;
		case cmd_Preferences:	outEnabled = true;	return;
		case cmd_Quit:			outEnabled = true;	return;
	}

	// If this is some command number we don't know about, pass it on
	// to PowerPlant.

	LDocApplication::FindCommandStatus (
		inCommand, outEnabled, outUsesMark, outMark, outName);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::UpdateMenus
 *
 * DESCRIPTION: LEventDispatcher method to update the status of all
 *				menu items.
 *
 *				We override this method to rebuild our host menubar
 *				from our cross-platform data structures.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/
 
void EmApplicationMac::UpdateMenus (void)
{
	EmMenu*		menubar		= ::MenuFindMenu (kMenuMenubarPreferred);
	EmAssert (menubar);

	// Update the contents and states of our menu items.

	::MenuUpdateMruMenus (*menubar);
	::MenuUpdateMenuItemStatus (*menubar);

	// Install them in the Mac menubar.

	::HostCreateMenubar (*menubar);

	// Have PowerPlant "update" them all.  Our menu items will already
	// be in the correct state after HostCreateMenubar.  But we need
	// to let PowerPlant grovel over all the menus so that it can take
	// care of any system menus.  While doing this, it will call
	// our FindCommandStatus override.  We'll re-return the same states
	// that the menu items are already in, so that should be harmless.

	LDocApplication::UpdateMenus ();
}


/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::ClickMenuBar
 *
 * DESCRIPTION: LEventDispatcher method to handle clicks on the menubar.
 *
 *				We override this method to make sure the menubar is
 *				up-to-date, regardless of whether or not something
 *				has called SetUpdateCommandStatus(true).
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/
 
void
EmApplicationMac::ClickMenuBar(
	const EventRecord&	inMacEvent)
{
	this->UpdateMenus ();
	LDocApplication::ClickMenuBar (inMacEvent);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::SpendTime
 *
 * DESCRIPTION: Classes derived from LPeriodical must define a SpendTime
 *				function, as this is a pure virtual function in LPeriodical.
 *
 *				We override this method in order to periodically service
 *				the serial ports and update our menus.	We also take this
 *				time to adjust our WaitNextEvent sleep value.
 *
 * PARAMETERS:	inMacEvent - current Mac event record (should be an
 *				"idle" event.
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::SpendTime (const EventRecord& inMacEvent)
{
#pragma unused (inMacEvent)

#if __profile__ && 0
	short	oldStatus = ::ProfilerGetStatus ();
	::ProfilerSetStatus (true);
#endif

//	if (gLogMisc)
//		LogAppendMsg ("Entering SpendTime.");
//	LogAppendMsg ("Enter EmApplicationMac::SpendTime");

	// Set a default sleep time.  The document may override this if
	// there is one.

	this->SetSleepTime (6);

	this->HandleIdle ();

	// If the initialized state of the Palm changed, update our
	// menus.

	if (fPalmUIInitialized != EmPatchState::UIInitialized ())
	{
		fPalmUIInitialized = EmPatchState::UIInitialized ();
		this->UpdateMenus ();
	}

//	LogAppendMsg ("Exit EmApplicationMac::SpendTime");
//	if (gLogMisc)
//		LogAppendMsg ("Leaving SpendTime.");

#if __profile__ && 0
	::ProfilerSetStatus (oldStatus);
#endif
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::DoCleanup
 *
 * DESCRIPTION: Override to perform any task that must be performed
 *				before the app quits.
 *
 *				This method is called by LCleanUpTask::CleanUpAtExit.  It
 *				is called because ExitToShell is being called.
 *
 *				This method is called to perform any cleanup tasks that
 *				we *really* need to perform, even if we crash (hence the
 *				patch to ExitToShell).	For the application, we make sure
 *				the serial and PPC ports are closed.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::DoCleanup (void)
{
	EmTransport::CloseAllTransports ();
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::BringToFront
 *
 * DESCRIPTION: Utility routine to make this application the frontmost
 *				application.  Called by routines like ExitDebugger so
 *				that the emulator is foremost when the user says to
 *				continue emulation.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

void EmApplicationMac::BringToFront (void)
{
	::SetFrontProcess (&fPSN);
}
	

#if (UNIVERSAL_INTERFACES_VERSION >= 0x0340)

/***********************************************************************
 *
 * FUNCTION:	EmApplicationMac::PreferencesCmdEventHandler
 *
 * DESCRIPTION: .
 *
 * PARAMETERS:	none
 *
 * RETURNED:	nothing
 *
 ***********************************************************************/

pascal OSStatus EmApplicationMac::PreferencesCmdEventHandler (
									EventHandlerCallRef	/*inHandlerCallRef*/,
									EventRef			inEvent,
									void*				/*inUserData*/)
{
	HICommand command;
	OSStatus status = ::GetEventParameter(
			inEvent,
			kEventParamDirectObject,
			typeHICommand, NULL,
			sizeof (HICommand), NULL,
			&command);

	if (status == noErr && command.commandID == kHICommandPreferences)
	{
		gHostApplication->ObeyCommand (cmd_Preferences, NULL);
		return noErr;
	}

	return eventNotHandledErr;
}

#endif


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmApplicationMac::SetTimeToQuit
// ---------------------------------------------------------------------------
// We override this method so that we can tell PowerPlant to quit after
// we finish the cross-platform quit process.

void EmApplicationMac::SetTimeToQuit (Bool quit)
{
	EmApplication::SetTimeToQuit (quit);

	if (this->GetTimeToQuit ())
	{
		LDocApplication::DoQuit (kAENo);
	}
}


#pragma mark -

// Should be of type GUSISpinFn, even though it gets installed as GUSIHook.

bool PrvSpinner (bool /*wait*/)
{
	EventRecord	event;
	::EventAvail (0, &event);

	return true;
}


Bool PrvCheckCarbon (void)
{
	// Make sure we're using at least version 1.2.5 of CarbonLib. The
	// version result is in BCD.  E.g., version 1.2.5 is returned as
	// 0x00000125.

	Int32	result;
	OSErr	err = ::Gestalt (gestaltCarbonVersion, &result);

	if (result < 0x00000125)
	{
		const char		kNumbers[] = "0123456789ABCDEF";
		Str255			buffer;
		buffer[0] = 0;

		for (int ii = 0; ii < 32; ii += 4)
		{
			uint8	nybble = (((uint32) result) >> (28 - ii)) & 0x0F;
			if (nybble)
			{
				if (buffer[0] > 0)
				{
					buffer[++buffer[0]] = '.';
				}

				buffer[++buffer[0]] = kNumbers[nybble];
			}
		}

		if (buffer[0] == 1)
		{
			buffer[0] = 3;
			buffer[2] = '.';
			buffer[3] = '0';
		}

		::ParamText (buffer, "\p", "\p", "\p");
		::StopAlert (128, NULL);
	
		return false;
	}

	return true;
}


EmCommandID PrvCommandT2EmCommandID (CommandT inCommand)
{
	EmCommandID	id = kCommandNone;

	if (inCommand >= kCommandFirst)
	{
		id = (EmCommandID) inCommand;
	}
	else
	{
		switch (inCommand)
		{
			case cmd_About:			id = kCommandAbout;				break;
			case cmd_New:			id = kCommandSessionNew;		break;
			case cmd_Open:			id = kCommandSessionOpenOther;	break;
			case cmd_Save:			id = kCommandSessionSave;		break;
			case cmd_SaveAs:		id = kCommandSessionSaveAs;		break;
			case cmd_Quit:			id = kCommandQuit;				break;

			case cmd_Undo:			id = kCommandUndo;				break;
			case cmd_Cut:			id = kCommandCut;				break;
			case cmd_Copy:			id = kCommandCopy;				break;
			case cmd_Paste:			id = kCommandPaste;				break;
			case cmd_Clear:			id = kCommandClear;				break;

			case cmd_Preferences:	id = kCommandPreferences;		break;
		}
	}

	return id;
}
