/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmClassic.h"

typedef UniversalProcPtr (*NewRoutineDescriptorProc)(ProcPtr, ProcInfoType, ISAType);
typedef THz (*ApplicationZoneProc)(void);
typedef void (*SetZoneProc)(THz);
typedef THz (*GetZoneProc)(void);
typedef OSErr (*PPCInitProc)(void);
typedef OSErr (*PPCOpenSyncProc)(PPCOpenPBPtr);
typedef OSErr (*PPCInformAsyncProc)(PPCInformPBPtr);
typedef OSErr (*PPCStartSyncProc)(PPCStartPBPtr);
typedef OSErr (*PPCWriteAsyncProc)(PPCWritePBPtr);
typedef OSErr (*PPCReadAsyncProc)(PPCReadPBPtr);
typedef OSErr (*PPCEndSyncProc)(PPCEndPBPtr);
typedef OSErr (*PPCEndAsyncProc)(PPCEndPBPtr);
typedef OSErr (*PPCCloseSyncProc)(PPCClosePBPtr);
typedef CRMErr (*InitCRMProc)(void);
typedef CRMRecPtr (*CRMSearchProc)(CRMRecPtr);
typedef OSErr (*SerResetProc)(short, short);
typedef OSErr (*SerSetBufProc)(short, Ptr, short);
typedef OSErr (*SerGetBufProc)(short, long *);
typedef OSErr (*OpenDriverProc)(ConstStr255Param, short *);
typedef OSErr (*CloseDriverProc)(short);
typedef OSErr (*ControlProc)(short, short, const void*);
typedef OSErr (*StatusProc)(short, short, void*);

NewRoutineDescriptorProc	sNewRoutineDescriptor;
ApplicationZoneProc			sApplicationZone;
SetZoneProc					sSetZone;
GetZoneProc					sGetZone;
PPCInitProc					sPPCInit;
PPCOpenSyncProc				sPPCOpenSync;
PPCInformAsyncProc			sPPCInformAsync;
PPCStartSyncProc			sPPCStartSync;
PPCWriteAsyncProc			sPPCWriteAsync;
PPCReadAsyncProc			sPPCReadAsync;
PPCEndSyncProc				sPPCEndSync;
PPCEndAsyncProc				sPPCEndAsync;
PPCCloseSyncProc			sPPCCloseSync;
InitCRMProc					sInitCRM;
CRMSearchProc				sCRMSearch;
SerResetProc				sSerReset;
SerSetBufProc				sSerSetBuf;
SerGetBufProc				sSerGetBuf;
OpenDriverProc				sOpenDriver;
CloseDriverProc				sCloseDriver;
ControlProc					sControl;
StatusProc					sStatus;

// ---------------------------------------------------------------------------
//		 PrvGetSharedLibrary
// ---------------------------------------------------------------------------

static CFragConnectionID PrvGetSharedLibrary (ConstStr255Param libName)
{
	CFragConnectionID	connID;

	OSErr	theErr;
	Str255	theErrName;
	Ptr		theMainAddr;

	theErr = ::GetSharedLibrary (libName, kCompiledCFragArch,
					kLoadCFrag, &connID, &theMainAddr, theErrName);

	if (theErr)
		connID = 0;

	return connID;
}


// ---------------------------------------------------------------------------
//		 PrvFindSymbol
// ---------------------------------------------------------------------------

static Ptr PrvFindSymbol (CFragConnectionID connID, ConstStr255Param symName)
{
	unsigned char	symClass;
	Ptr				addr;

	if (connID)
	{
		OSErr	theErr = ::FindSymbol (connID, symName, (Ptr*) &addr, &symClass);
		EmAssert (theErr == 0);

		if (theErr)
			addr = NULL;
	}
	else
	{
		addr = NULL;
	}

	return addr;
}


// ---------------------------------------------------------------------------
//		 PrvGetInterfaceLib
// ---------------------------------------------------------------------------

static CFragConnectionID PrvGetInterfaceLib (void)
{
	static CFragConnectionID	sConnID;

	if (sConnID == 0)
	{
		sConnID = ::PrvGetSharedLibrary ("\pInterfaceLib");
	}

	return sConnID;
}


// ---------------------------------------------------------------------------
//		 PrvFindInterfaceLibSymbol
// ---------------------------------------------------------------------------

static Ptr PrvFindInterfaceLibSymbol (ConstStr255Param symName)
{
	return ::PrvFindSymbol (::PrvGetInterfaceLib (), symName);
}


// ---------------------------------------------------------------------------
//		 EmClassicInit
// ---------------------------------------------------------------------------

void EmClassicInit (void)
{
#define GET_FUNC(name)		\
	do { if (s##name == NULL)		\
		s##name = (name##Proc) ::PrvFindInterfaceLibSymbol ("\p" #name);	\
	} while (0)


	GET_FUNC (NewRoutineDescriptor);
	GET_FUNC (ApplicationZone);
	GET_FUNC (SetZone);
	GET_FUNC (GetZone);
	GET_FUNC (PPCInit);
	GET_FUNC (PPCOpenSync);
	GET_FUNC (PPCInformAsync);
	GET_FUNC (PPCStartSync);
	GET_FUNC (PPCWriteAsync);
	GET_FUNC (PPCReadAsync);
	GET_FUNC (PPCEndSync);
	GET_FUNC (PPCEndAsync);
	GET_FUNC (PPCCloseSync);
	GET_FUNC (InitCRM);
	GET_FUNC (CRMSearch);
	GET_FUNC (SerReset);
	GET_FUNC (SerSetBuf);
	GET_FUNC (SerGetBuf);
	GET_FUNC (OpenDriver);
	GET_FUNC (CloseDriver);
	GET_FUNC (Control);
	GET_FUNC (Status);
}


Bool	EmClassicHasPPCToolbox (void)
{
	return sPPCInit != NULL;
}


Bool	EmClassicHasSerialManager (void)
{
	return	sInitCRM != NULL &&
			sCRMSearch != NULL &&
			sSerReset != NULL &&
			sOpenDriver != NULL;
}


Bool	EmClassicHasZones (void)
{
	return sApplicationZone != NULL;
}


// ---------------------------------------------------------------------------
//		 NewRoutineDescriptor
// ---------------------------------------------------------------------------

DEFINE_API( UniversalProcPtr )
NewRoutineDescriptor            (ProcPtr                theProc,
                                 ProcInfoType           theProcInfo,
                                 ISAType                theISA)
{
	EmAssert (sNewRoutineDescriptor);

	if (!sNewRoutineDescriptor)
		return NULL;

	return sNewRoutineDescriptor (theProc, theProcInfo, theISA);
}


// ---------------------------------------------------------------------------
//		 ApplicationZone
// ---------------------------------------------------------------------------

DEFINE_API( THz )
ApplicationZone                 (void)
{
	EmAssert (sApplicationZone);

	if (!sApplicationZone)
		return NULL;

	return sApplicationZone ();
}


// ---------------------------------------------------------------------------
//		 SetZone
// ---------------------------------------------------------------------------

DEFINE_API( void )
SetZone                         (THz                    hz)
{
	EmAssert (sSetZone);

	if (!sSetZone)
		return;

	sSetZone (hz);
}


// ---------------------------------------------------------------------------
//		 GetZone
// ---------------------------------------------------------------------------

DEFINE_API( THz )
GetZone                         (void)
{
	EmAssert (sGetZone);

	if (!sGetZone)
		return NULL;

	return sGetZone ();
}


// ---------------------------------------------------------------------------
//		 PPCInit
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
PPCInit							(void)
{
	EmAssert (sPPCInit);

	if (!sPPCInit)
		return cfragNoSymbolErr;

	return sPPCInit ();
}


// ---------------------------------------------------------------------------
//		 PPCOpenSync
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
PPCOpenSync						(PPCOpenPBPtr 			pb)
{
	EmAssert (sPPCOpenSync);

	if (!sPPCOpenSync)
		return cfragNoSymbolErr;

	return sPPCOpenSync (pb);
}


// ---------------------------------------------------------------------------
//		 PPCInformAsync
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
PPCInformAsync					(PPCInformPBPtr 		pb)
{
	EmAssert (sPPCInformAsync);

	if (!sPPCInformAsync)
		return cfragNoSymbolErr;

	return sPPCInformAsync (pb);
}


// ---------------------------------------------------------------------------
//		 PPCStartSync
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
PPCStartSync					(PPCStartPBPtr 			pb)
{
	EmAssert (sPPCStartSync);

	if (!sPPCStartSync)
		return cfragNoSymbolErr;

	return sPPCStartSync (pb);
}


// ---------------------------------------------------------------------------
//		 PPCWriteAsync
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
PPCWriteAsync					(PPCWritePBPtr 			pb)
{
	EmAssert (sPPCWriteAsync);

	if (!sPPCWriteAsync)
		return cfragNoSymbolErr;

	return sPPCWriteAsync (pb);
}


// ---------------------------------------------------------------------------
//		 PPCReadAsync
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
PPCReadAsync					(PPCReadPBPtr 			pb)
{
	EmAssert (sPPCReadAsync);

	if (!sPPCReadAsync)
		return cfragNoSymbolErr;

	return sPPCReadAsync (pb);
}


// ---------------------------------------------------------------------------
//		 PPCEndSync
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
PPCEndSync						(PPCEndPBPtr 			pb)
{
	EmAssert (sPPCEndSync);

	if (!sPPCEndSync)
		return cfragNoSymbolErr;

	return sPPCEndSync (pb);
}


// ---------------------------------------------------------------------------
//		 PPCEndAsync
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
PPCEndAsync						(PPCEndPBPtr 			pb)
{
	EmAssert (sPPCEndAsync);

	if (!sPPCEndAsync)
		return cfragNoSymbolErr;

	return sPPCEndAsync (pb);
}


// ---------------------------------------------------------------------------
//		 PPCCloseSync
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
PPCCloseSync					(PPCClosePBPtr 			pb)
{
	EmAssert (sPPCCloseSync);

	if (!sPPCCloseSync)
		return cfragNoSymbolErr;

	return sPPCCloseSync (pb);
}


// ---------------------------------------------------------------------------
//		 NewPPCCompUPP
// ---------------------------------------------------------------------------

DEFINE_API(PPCCompUPP)
NewPPCCompUPP					(PPCCompProcPtr			userRoutine)
{
	/* pascal no_return_value Func(4_bytes) */
    enum { uppPPCCompProcInfo = 0x000000C0 };

	return (PPCCompUPP) NewRoutineDescriptor ((ProcPtr) userRoutine,
							uppPPCCompProcInfo, GetCurrentArchitecture ());
}


// ---------------------------------------------------------------------------
//		 InitCRM
// ---------------------------------------------------------------------------

EXTERN_API( CRMErr )
InitCRM                         (void)
{
	EmAssert (sInitCRM);

	if (!sInitCRM)
		return -1;

	return sInitCRM ();
}


// ---------------------------------------------------------------------------
//		 CRMSearch
// ---------------------------------------------------------------------------

DEFINE_API( CRMRecPtr )
CRMSearch                       (CRMRecPtr              crmReqPtr)
{
	EmAssert (sCRMSearch);

	if (!sCRMSearch)
		return NULL;

	return sCRMSearch (crmReqPtr);
}


// ---------------------------------------------------------------------------
//		 SerReset
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
SerReset                        (short                  refNum,
                                 short                  serConfig)
{
	EmAssert (sSerReset);

	if (!sSerReset)
		return cfragNoSymbolErr;

	return sSerReset (refNum, serConfig);
}


// ---------------------------------------------------------------------------
//		 SerSetBuf
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
SerSetBuf                       (short                  refNum,
                                 Ptr                    serBPtr,
                                 short                  serBLen)
{
	EmAssert (sSerSetBuf);

	if (!sSerSetBuf)
		return cfragNoSymbolErr;

	return sSerSetBuf (refNum, serBPtr, serBLen);
}


// ---------------------------------------------------------------------------
//		 SerGetBuf
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
SerGetBuf                       (short                  refNum,
                                 long *                 count)
{
	EmAssert (sSerGetBuf);

	if (!sSerGetBuf)
		return cfragNoSymbolErr;

	return sSerGetBuf (refNum, count);
}


// ---------------------------------------------------------------------------
//		 MacOpenDriver
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
MacOpenDriver                   (ConstStr255Param       name,
                                 short *                drvrRefNum)
{
	EmAssert (sOpenDriver);

	if (!sOpenDriver)
		return cfragNoSymbolErr;

	return sOpenDriver (name, drvrRefNum);
}


// ---------------------------------------------------------------------------
//		 MacCloseDriver
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
MacCloseDriver                  (short                  refNum)
{
	EmAssert (sCloseDriver);

	if (!sCloseDriver)
		return cfragNoSymbolErr;

	return sCloseDriver (refNum);
}


// ---------------------------------------------------------------------------
//		 Control
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
Control                         (short                  refNum,
                                 short                  csCode,
                                 const void *           csParamPtr)
{
	EmAssert (sControl);

	if (!sControl)
		return cfragNoSymbolErr;

	return sControl (refNum, csCode, csParamPtr);
}


// ---------------------------------------------------------------------------
//		 Status
// ---------------------------------------------------------------------------

DEFINE_API( OSErr )
Status                          (short                  refNum,
                                 short                  csCode,
                                 void *                 csParamPtr)
{
	EmAssert (sStatus);

	if (!sStatus)
		return cfragNoSymbolErr;

	return sStatus (refNum, csCode, csParamPtr);
}
