/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 2000-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmDirRefMac.h"

#include "ErrorHandling.h"		// Errors::ThrowIfError


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::EmDirRef
 *
 * DESCRIPTION:	Various ways to make a directory reference.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

EmDirRef::EmDirRef (void) :
	fDirAlias (NULL),
	fDirSpec ()
{
	::PrvUnspecify (&fDirSpec);
}


EmDirRef::EmDirRef (const EmDirRef& other) :
	fDirAlias (NULL),
	fDirSpec (other.fDirSpec)
{
	Handle	newHandle = (Handle) other.fDirAlias;

	if (newHandle)
	{
		OSErr	err = ::HandToHand (&newHandle);
		Errors::ThrowIfError (err);
	}

	fDirAlias = (AliasHandle) newHandle;
}


EmDirRef::EmDirRef (const char* s) :
	fDirAlias (NULL)
{
	LStr255 name (s);
	::PrvMakeSpec (name, &fDirSpec);
}


EmDirRef::EmDirRef (const string& s) :
	fDirAlias (NULL)
{
	LStr255 name (s.c_str ());
	::PrvMakeSpec (name, &fDirSpec);
}


EmDirRef::EmDirRef (const EmDirRef& parent, const char* s) :
	fDirAlias (NULL)
{
	LStr255 name (s);
	::PrvMakeSpec (parent, name, &fDirSpec);
}


EmDirRef::EmDirRef (const EmDirRef& parent, const string& s) :
	fDirAlias (NULL)
{
	LStr255 name (s.c_str ());
	::PrvMakeSpec (parent, name, &fDirSpec);
}


EmDirRef::EmDirRef (const unsigned char* s) :
	fDirAlias (NULL)
{
	LStr255 name (s);
	::PrvMakeSpec (name, &fDirSpec);
}


EmDirRef::EmDirRef (const FSSpec& spec) :
	fDirAlias (NULL),
	fDirSpec (spec)
{
	// Get an alias now, before the spec can be invalidated.

	this->UpdateAlias ();
}


EmDirRef::EmDirRef (AliasHandle alias) :
	fDirAlias (alias),
	fDirSpec ()
{
	::PrvUnspecify (&fDirSpec);
	this->UpdateSpec ();
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::~EmDirRef
 *
 * DESCRIPTION: EmDirRef destructor.  Nothing special to do...
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

EmDirRef::~EmDirRef (void)
{
	::DisposeHandle ((Handle) fDirAlias);
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::operator=
 *
 * DESCRIPTION: Assignment operator.  If "other" is not the same as
 *				the controlled object, copy the contents.
 *
 * PARAMETERS:	other - object to copy.
 *
 * RETURNED:	reference to self.
 *
 ***********************************************************************/

EmDirRef&
EmDirRef::operator= (const EmDirRef& other)
{
	if (&other != this)
	{
		Handle	newHandle = (Handle) other.fDirAlias;

		if (newHandle)
		{
			OSErr	err = ::HandToHand (&newHandle);
			Errors::ThrowIfError (err);
		}

		::DisposeHandle ((Handle) fDirAlias);
		fDirAlias = (AliasHandle) newHandle;
		fDirSpec = other.fDirSpec;
	}

	return *this;
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::IsSpecified
 *
 * DESCRIPTION: Returns whether or not the controlled object has been
 *				pointed to a (possibly non-existant) file, or if it's
 *				empty (that it, it was created with the default ctor).
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	True if the object points to a file.
 *
 ***********************************************************************/

Bool
EmDirRef::IsSpecified (void) const
{
	return (fDirAlias != NULL) || ::PrvIsSpecified (fDirSpec);
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::Exists
 *
 * DESCRIPTION: Returns whether or not the controlled object points to
 *				an existing file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	True if the referenced file exists.
 *
 ***********************************************************************/

Bool
EmDirRef::Exists (void) const
{
	if (this->IsSpecified ())
	{
		FSSpec	spec = this->GetFSSpec ();

		return ::PrvSpecExists (spec);
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::Create
 *
 * DESCRIPTION:	Attempt to create the managed directory.  Does nothing
 *				if the directory already exists.  Throws an exception
 *				if the attempt to create the directory fails.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void
EmDirRef::Create (void) const
{
	if (!this->Exists () && this->IsSpecified ())
	{
		// Make sure all directories down to us are created, too.

		EmDirRef	parent = this->GetParent ();
		parent.Create ();

		long	dirID;
		OSErr	err = ::FSpDirCreate (&fDirSpec, smSystemScript, &dirID);
		if (err != noErr)
		{
			// !!! throw...
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::GetName
 *
 * DESCRIPTION: Returns the name of the referenced file.  Only the file
 *				*name* is returned, not the full path.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	A string containing the name.  If the file is not
 *				specified, an empty string is returned.  No checks are
 *				made to see if the file actually exists.
 *
 ***********************************************************************/

string
EmDirRef::GetName (void) const
{
	FSSpec	mySpec	= this->GetFSSpec ();
	return string ((char*) &mySpec.name[1], mySpec.name[0]);
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::GetParent
 *
 * DESCRIPTION:	Returns an object representing the parent (or container)
 *				of the managed file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	An object representing the file's parent.
 *
 ***********************************************************************/

EmDirRef
EmDirRef::GetParent (void) const
{
	EmDirRef	result;

	if (this->IsSpecified ())
	{
		FSSpec	mySpec	= this->GetFSSpec ();
		FSSpec	parSpec	= ::PrvGetParentSpec (mySpec);

		if (parSpec.vRefNum != 0)
		{
			result = EmDirRef (parSpec);
		}
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::GetFullPath
 *
 * DESCRIPTION: 
 *
 *				Lifted from MoreFiles 1.4.9.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

string
EmDirRef::GetFullPath (void) const
{
	return ::PrvGetFullPath (fDirSpec);
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::GetChildren
 *
 * DESCRIPTION:	Get a full (platform-specific) path to the object.  The
 *				path is canonicalized in that it will always have a
 *				trailing slash.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	An string representing the file's path.
 *
 ***********************************************************************/

void
EmDirRef::GetChildren (EmFileRefList* fileList, EmDirRefList* dirList) const
{
	if (!this->Exists ())
		return;

	// Get our dirID.
	
	short	vRefNum;
	long	dirID;
	::PrvGetDirVolDir (*this, &vRefNum, &dirID);
	
	// Start iterating over it.

	OSErr		err;
	long		index = 1;
	CInfoPBRec	pb;
	LStr255 	fileName;	

	pb.hFileInfo.ioNamePtr = fileName;
	pb.hFileInfo.ioVRefNum = vRefNum;

	do
	{
		// ioDrDirID contains dirID of named directory when PBGetCatInfo returns;
		// ioDirParID ends up holding what we originally placed in ioDrDirID.

		pb.hFileInfo.ioDirID = dirID;
		pb.hFileInfo.ioFDirIndex = index++;
		
		err = PBGetCatInfoSync (&pb);
		if (err == noErr)
		{
			FSSpec	spec;
			err = ::FSMakeFSSpec (vRefNum, dirID, fileName, &spec);
			if (err == noErr)
			{
				if ((pb.hFileInfo.ioFlAttrib & ioDirMask) == 0)
				{
					if (fileList)
						fileList->push_back (EmFileRef (spec));
				}
				else
				{
					if (dirList)
						dirList->push_back (EmDirRef (spec));
				}
			}
		}
	}
	while (err == noErr);
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::operator==
 * FUNCTION:	EmDirRef::operator!=
 * FUNCTION:	EmDirRef::operator>
 * FUNCTION:	EmDirRef::operator<
 *
 * DESCRIPTION: Bogus operators for wiggy VC++ compiler which won't let
 *				us instantiate STL containers without them.
 *
 * PARAMETERS:	other - object to compare ourself to.
 *
 * RETURNED:	True if the requested condition is true.  Comparisons
 *				are based on the file's full path.
 *
 ***********************************************************************/

bool
EmDirRef::operator== (const EmDirRef& other) const
{
	FSSpec	mySpec = const_cast<EmDirRef*>(this)->GetFSSpec ();
	FSSpec	hisSpec = const_cast<EmDirRef&>(other).GetFSSpec ();

	return ::PrvEqualFSSpec (mySpec, hisSpec);
}


bool
EmDirRef::operator!= (const EmDirRef& other) const
{
	return !(*this == other);
}


/***********************************************************************
 *
 * FUNCTION:	
 *
 * DESCRIPTION: 
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

bool
EmDirRef::FromPrefString (const string& s)
{
	if (fDirAlias)
	{
		::DisposeHandle ((Handle) fDirAlias);
		fDirAlias = NULL;
	}

	::PrvUnspecify (&fDirSpec);

	if (s.size() == 0)
	{
		return true;
	}

	short	vol;
	long	dir;
	::PrvGetAppVolDir (&vol, &dir);

	LStr255 name (s.c_str ());
	OSErr	err = ::FSMakeFSSpec (vol, dir, name, &fDirSpec);

	return (err == noErr) || (err == fnfErr) || (err == nsvErr);
}


/***********************************************************************
 *
 * FUNCTION:	
 *
 * DESCRIPTION: 
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

string
EmDirRef::ToPrefString (void) const
{
	return this->GetFullPath ();
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::GetEmulatorDirectory
 *
 * DESCRIPTION:	Return an EmDirRef for Poser's directory.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	The desired EmDirRef.
 *
 ***********************************************************************/

EmDirRef
EmDirRef::GetEmulatorDirectory (void)
{
	EmDirRef	result;

	short	vRefNum;
	long	dirID;
	::PrvGetAppVolDir (&vRefNum, &dirID);

	Str255	name = {0};
	FSSpec	spec;
	OSErr	err = ::FSMakeFSSpec (vRefNum, dirID, name, &spec);

	if (err == noErr || err == fnfErr)
	{
		result = EmDirRef (spec);
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::GetPrefsDirectory
 *
 * DESCRIPTION:	Return an EmDirRef for Poser's preferences directory.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	The desired EmDirRef.
 *
 ***********************************************************************/

EmDirRef
EmDirRef::GetPrefsDirectory (void)
{
	EmDirRef	result;

	short	vRefNum;
	long	dirID;
	OSErr	err = ::FindFolder (kOnSystemDisk, kPreferencesFolderType,
								kDontCreateFolder, &vRefNum, &dirID);

	Str255	name = {0};
	FSSpec	spec;
	err = ::FSMakeFSSpec (vRefNum, dirID, name, &spec);

	if (err == noErr || err == fnfErr)
	{
		result = EmDirRef (spec);
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::GetAlias
 *
 * DESCRIPTION: Returns the alias of the referenced file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	Alias for the referenced file.	If the EmDirRef
 *				was created with a file spec, we attempt to create an
 *				alias and return it.  If an alias cannot be created,
 *				this function could return NULL.
 *
 ***********************************************************************/

AliasHandle
EmDirRef::GetAlias (void) const
{
	if (!fDirAlias && ::PrvIsSpecified (fDirSpec))
	{
		const_cast<EmDirRef*>(this)->UpdateAlias ();
	}

	return fDirAlias;
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::GetFSSpec
 *
 * DESCRIPTION: Returns the filespec of the referenced file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	Filespec for the referenced file.  If the EmDirRef
 *				was created with an alias, we attempt to resolve it and
 *				return the resulting filespec.	If an alias cannot be
 *				resolved, this function could return an empty spec.
 *
 ***********************************************************************/

FSSpec
EmDirRef::GetFSSpec (void) const
{
	if (!::PrvIsSpecified (fDirSpec) && fDirAlias)
	{
		const_cast<EmDirRef*>(this)->UpdateSpec ();
	}

	return fDirSpec;
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::UpdateAlias
 *
 * DESCRIPTION: Creates an alias from the filespec for the controlled
 *				file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void
EmDirRef::UpdateAlias (void)
{
	OSErr	err = ::NewAlias (NULL, &fDirSpec, &fDirAlias);

// I think there are too many "soft errors" (fnfErr, userCanceledErr)
// that can occur here that we don't want to throw an exception.
//
//	Errors::ThrowIfError (err);
}


/***********************************************************************
 *
 * FUNCTION:	EmDirRef::UpdateSpec
 *
 * DESCRIPTION: Resolves an alias and saves the resulting filespec.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void
EmDirRef::UpdateSpec (void)
{
	Boolean wasChanged;
	short	matchCount = 1;

	// Use MatchAlias instead of ResolveAlias to avoid any UI.

//	OSErr	err = ::ResolveAlias (NULL, fDirAlias, &fDirSpec, &wasChanged);
	OSErr	err = ::MatchAlias (NULL, kARMSearch | kARMNoUI, fDirAlias,
									&matchCount, &fDirSpec, &wasChanged, NULL, NULL);

// I think there are too many "soft errors" (fnfErr, userCanceledErr)
// that can occur here that we don't want to throw an exception.
//
//	Errors::ThrowIfError (err);
}


/***********************************************************************
 *
 * FUNCTION:	PrvEqualAlias
 *
 * DESCRIPTION: Utility to compare two aliases for equality.
 *
 * PARAMETERS:	alias1, alias2 - aliases to compare
 *
 * RETURNED:	True if the two aliases refer to the same file.
 *
 ***********************************************************************/

Bool PrvEqualAlias (AliasHandle alias1, AliasHandle alias2)
{
	Boolean wasChanged;

	FSSpec	target1;
	OSErr	err1 = ::ResolveAlias (NULL, alias1, &target1, &wasChanged);

	FSSpec	target2;
	OSErr	err2 = ::ResolveAlias (NULL, alias2, &target2, &wasChanged);

	if (err1 || err2)
	{
		return false;
	}

	return ::PrvEqualFSSpec (target1, target2);
}


/***********************************************************************
 *
 * FUNCTION:	PrvEqualFSSpec
 *
 * DESCRIPTION: Utility to compare two FSSpec's for equality.
 *
 * PARAMETERS:	spec1, spec2 - FSSpec's to compare
 *
 * RETURNED:	True if the two FSSpec's refer to the same file.
 *
 ***********************************************************************/

Bool PrvEqualFSSpec (const FSSpec& spec1, const FSSpec& spec2)
{
	if ((spec1.vRefNum != spec2.vRefNum)	||
		(spec1.parID != spec2.parID)		||
		(::RelString (spec1.name, spec2.name, false, false) != 0))
	{
		return false;
	}

	return true;
}


/***********************************************************************
 *
 * FUNCTION:	PrvIsType
 *
 * DESCRIPTION: Determine if the given file is of the given tpe.
 *
 * PARAMETERS:	
 *
 * RETURNED:	.
 *
 ***********************************************************************/

Bool PrvIsType (const FSSpec& inSpec, OSType type, const char* suffix)
{
	FSSpec	spec = inSpec;

	FInfo	fndrInfo;
	OSErr	err = ::FSpGetFInfo (&spec, &fndrInfo);
	if (err != noErr)
		return false;

	if (fndrInfo.fdType == type)
		return true;

	// Some PRC files have incorrect file types.
	// Special-case those by looking at their file extensions

	LStr255 fileName (spec.name);
	fileName.SetCompareFunc (LString::CompareIgnoringCase);

	if (fileName.EndsWith (LStr255(suffix)))
		return true;

	// Add special hacks for ROM files.
	if (strcmp (".rom", suffix) == 0)
	{
		if (fileName.EndsWith (LStr255(" rom")))
			return true;

		if (fileName.BeginsWith (LStr255("rom.")))
			return true;
	}

	return false;
}

Bool PrvIsSpecified (const FSSpec& inSpec)
{
	return inSpec.vRefNum != 0;
}

void PrvUnspecify (FSSpec* ioSpec)
{
	memset (ioSpec, 0, sizeof (FSSpec));
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetParentSpec
 *
 * DESCRIPTION:	Returns an FSSpec for the parent of the file system
 *				object represented by the input FSSpec.  If an error
 *				occurs or we're at the root level, an FSSpec with
 *				vRefNum == 0 is returned.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	An FSSpec representing the object's parent.
 *
 ***********************************************************************/

FSSpec PrvGetParentSpec (const FSSpec& inSpec)
{
	FSSpec	result;
	::PrvUnspecify (&result);

	if (inSpec.vRefNum != fsRtParID)
	{
		OSErr	err = ::FSMakeFSSpec (	inSpec.vRefNum,
										inSpec.parID,
										NULL, &result);

		if (err != noErr)
		{
			::PrvUnspecify (&result);
		}
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetFullPath
 *
 * DESCRIPTION: Generate a full path for the given FSSpec.
 *
 * PARAMETERS:	inSpec - the FSSpec for which a full path should be
 *					created.
 *
 * RETURNED:	The full path, in Macintosh format.  An empty string
 *				is returned if the FSSpec is empty or refers to an
 *				invalid parent directory.
 *
 ***********************************************************************/

string
PrvGetFullPath (const FSSpec& inSpec)
{
	if (!::PrvIsSpecified (inSpec))
		return string();

	// Start off with the name of the file system object.

	string	fullPath ((char*) &inSpec.name[1], inSpec.name[0]);

	// Is the object a file or a directory?

	FSSpec		curSpec = inSpec;
	CInfoPBRec	pb;

	pb.dirInfo.ioNamePtr	= curSpec.name;
	pb.dirInfo.ioVRefNum	= curSpec.vRefNum;
	pb.dirInfo.ioDrDirID	= curSpec.parID;
	pb.dirInfo.ioFDirIndex	= 0;

	OSErr		err = PBGetCatInfoSync (&pb);

	// Allow file/directory name at end of path to not exist.

	if ((err == noErr) || (err == fnfErr))
	{
		// If the object is a directory, append a colon so full pathname
		// ends with colon.

		if ((err == noErr) && ((pb.hFileInfo.ioFlAttrib & ioDirMask) != 0))
		{
			fullPath += ':';
		}

		// Loop up the hierarchy, getting information about parent
		// directories, until we reach the top.

		while (curSpec.parID != fsRtParID)
		{
			OSErr	err = ::FSMakeFSSpec (curSpec.vRefNum, curSpec.parID,
											NULL, &curSpec);
			if (err)
			{
				fullPath.clear ();
				break;
			}

			string	name ((char*) &curSpec.name[1], curSpec.name[0]);
			
			fullPath = name + ":" + fullPath;
		}
	}
	else
	{
		fullPath.clear ();
	}

	return fullPath;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetAppVolDir
 *
 * DESCRIPTION: Utility routine to return the location of the application.
 *				This function is the same as the one in CEmulatorApp.
 *				However, we can't call that, as we can't count on the
 *				application object being around when we need it.  For
 *				instance, the prefs stuff is created before the app, and
 *				it may need to create file references.
 *
 * PARAMETERS:	vol - pointer to where the application volume should be
 *					stored.
 *
 *				dir - pointer to where the application directory should
 *					be stored.
 *
 * RETURNED:	nothing.  If an error occurs, an exception is thrown.
 *
 ***********************************************************************/

void PrvGetAppVolDir (short* vol, long* dir)
{
	ProcessSerialNumber	psn;
	ProcessInfoRec		pInfo;
	FSSpec				us;
	
	(void) ::GetCurrentProcess (&psn);

	pInfo.processInfoLength = sizeof (pInfo);
	pInfo.processName		= NULL;
	pInfo.processAppSpec	= &us;

	OSErr	err = ::GetProcessInformation (&psn, &pInfo);
	Errors::ThrowIfError (err);

	if (vol)
		*vol = us.vRefNum;

	if (dir)
		*dir = us.parID;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetDirVolDir
 *
 * DESCRIPTION: Utility routine to return the vRefNum and dirID of a
 *				directory, given its name, parent dirID and vRefNum.
 *
 * PARAMETERS:	vRefNum - pointer to where the application volume should
 *					be stored.
 *
 *				dirID - pointer to where the application directory
 *					should be stored.
 *
 * RETURNED:	nothing.  If an error occurs, an exception is thrown.
 *
 ***********************************************************************/

void PrvGetDirVolDir (const EmDirRef& dir, short* vRefNum, long* dirID)
{
	FSSpec		spec = dir.GetFSSpec ();
	CInfoPBRec	pb;

	pb.dirInfo.ioNamePtr	= spec.name;
	pb.dirInfo.ioVRefNum	= spec.vRefNum;
	pb.dirInfo.ioDrDirID	= spec.parID;
	pb.dirInfo.ioFDirIndex	= 0;

	OSErr	err = PBGetCatInfoSync (&pb);
	if (err == noErr)
	{
		*vRefNum	= pb.dirInfo.ioVRefNum;
		*dirID		= pb.dirInfo.ioDrDirID;
	}
	else
	{
		*vRefNum	= 0;
		*dirID		= 0;
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvSpecExists
 *
 * DESCRIPTION: Utility routine to determine if a file system entity
 *				exists or not.
 *
 * PARAMETERS:	spec - the FSSpec for the entity to test.
 *
 * RETURNED:	True if it exists.
 *
 ***********************************************************************/

Bool PrvSpecExists (const FSSpec& spec)
{
	CInfoPBRec	pb;
	LStr255		name (spec.name);	// Make a non-const copy

	pb.dirInfo.ioNamePtr	= name;
	pb.dirInfo.ioVRefNum	= spec.vRefNum;
	pb.dirInfo.ioDrDirID	= spec.parID;
	pb.dirInfo.ioFDirIndex	= 0;

	OSErr	err = PBGetCatInfoSync (&pb);
	return err == 0;
}


/***********************************************************************
 *
 * FUNCTION:	PrvMakeSpec
 *
 * DESCRIPTION: Utility routines to create a new FSSpec for a file
 *				system object, given the object's name and optional
 *				location information.  If no location information is
 *				included, the file system object is assumed to be in
 *				the emulator's directory.
 *
 * PARAMETERS:	result - pointer to FSSpec where result is to be stored.
 *
 *				name - name of the file system object.
 *
 *				parent - EmDirRef of the parent directory of this object
 *
 *				vRefNum, dirID - location information of the parent
 *					directory of this object
 *
 * RETURNED:	nothing.  If an error occurs, an exception is thrown.
 *
 ***********************************************************************/

void PrvMakeSpec (const LStr255& name, FSSpec* result)
{
	short	vol;
	long	dir;
	::PrvGetAppVolDir (&vol, &dir);
	::PrvMakeSpec (vol, dir, name, result);
}


void PrvMakeSpec (const EmDirRef& parent, const LStr255& name, FSSpec* result)
{
	short	vol;
	long	dir;
	::PrvGetDirVolDir (parent, &vol, &dir);
	::PrvMakeSpec (vol, dir, name, result);
}


void PrvMakeSpec (short vRefNum, long dirID, const LStr255& name, FSSpec* result)
{
	OSErr	err = ::FSMakeFSSpec (vRefNum, dirID, name, result);

	if (err != noErr && err != fnfErr && err != dirNFErr && err != nsvErr)
	{
		::PrvUnspecify (result);
		return;
	}

	// If this is an alias, resolve it.

	Boolean	isFolder;
	Boolean	isAlias;
	err = ::ResolveAliasFile (result, true, &isFolder, &isAlias);

	if (err != noErr && err != fnfErr && err != dirNFErr && err != nsvErr)
	{
		::PrvUnspecify (result);
		return;
	}
}
