/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1999-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmDlgMac.h"

#include "CURLPushButton.h"		// GetURL
#include "EmApplicationMac.h"	// gHostApplication
#include "EmFileRefMac.h"		// PrvTypeToOSType
#include "EmPatchState.h"		// EmPatchState::GetROMCharEncoding
#include "ErrorHandling.h"		// Errors::SetParameter, ReplaceParameters, etc.
#include "Palm OS Emulator.r.h"	// wnd_About
#include "Platform.h"			// GetString, GetShortVersionString
#include "Strings.r.h"			// kStr_AppAndVers

#include <LEditText.h>
#include <LPane.h>
#include <LPopupButton.h>
#include <LScrollerView.h>
#include <LStaticText.h>
#include <LTabsControl.h>
#include <LTextColumn.h>
#include <LWindow.h>
#include <PP_Messages.h>
#include <UEventMgr.h>
#include <UInternetConfig.h>
#include <UNavServicesDialogs.h>


typedef vector<OSType>		HostFileTypeList;

EmFileTypeList				gFileTypes;
HostFileTypeList			gHostFileTypes;

static pascal Boolean		PrvFileFilterProc	(AEDesc* theItem, void* info,
												 NavCallBackUserData callBackUD,
												 NavFilterModes filterMode);
static HostFileTypeList		PrvConvertTypeList	(const EmFileTypeList& typeList);


static void LinkListenerToPanes (LListener* listener, LPane* pane)
{
	//
	// Link this pane to the listener if it is a broadcaster.
	//

	LBroadcaster* broadcaster (dynamic_cast<LBroadcaster*> (pane));
	if (broadcaster)
	{
		broadcaster->AddListener (listener);
	}

	//
	// If this pane is a view, recurse for the sub-panes.
	//

	LView* view (dynamic_cast<LView*> (pane));
	if (view)
	{
		TArrayIterator<LPane*> iterator (view->GetSubPanes ());
		LPane* sub;
		while (iterator.Next (sub))
		{
			::LinkListenerToPanes (listener, sub);
		}
	}
}
    

static void PrvInitializeLColumnViews (LView* curParent)
{
	TArray<LPane*>&	children = curParent->GetSubPanes ();

	for (UInt32 ii = 1; ii <= children.GetCount (); ++ii)
	{
		LPane*	child = children[ii];
		EmAssert (child);

		//
		// Recurse if we find other LViews.
		//

		LView*	childView = dynamic_cast<LView*> (child);
		if (childView)
		{
			::PrvInitializeLColumnViews (childView);
		}

		//
		// If we find LColumnViews, hook them up to the LListener.
		// As well, adjust the row heights, and adjust the view
		// height to be an integral number of rows.  If the
		// LColumnView is in an LScroller, adjust its height, too.
		//

		LColumnView*	childColumnView = dynamic_cast<LColumnView*> (child);
		if (childColumnView)
		{
			//
			// Set the cell height based on the font.
			//

			short	cellHeight = ::GetListCellHeight ();
			childColumnView->SetRowHeight (cellHeight, 0, 1);

			//
			// Resize the view to have an integral number of cells
			//

			SDimension16	frameSize;
			childColumnView->GetFrameSize (frameSize);

			int	delta = frameSize.height % cellHeight;

			//
			// If the list is in a scroller, resize that, assuming
			// that the embedded list view is bound to its edges.
			// Otherwise, resize the list view directly.
			//

			LView*	scroller = childColumnView->GetSuperView ();
			if (scroller && dynamic_cast<LScrollerView*>(scroller))
			{
				scroller->ResizeFrameBy (0, -delta, false);
			}
			else
			{
				childColumnView->ResizeFrameBy (0, -delta, false);
			}
		}
	}
}


static LWindow* PrvCreateDialog (EmDlgFn fn, void* data, EmDlgID dlgID)
{
	//
	// Create the dialog.
	//

	LWindow*		w = LWindow::CreateWindow (::PrvFromDlgID (dlgID), gHostApplication);
	EmDialogBox*	dlg = dynamic_cast <EmDialogBox*> (w);
	EmAssert (dlg);

	//
	// Get a handy reference to the dialog's context.
	//
	
	EmDlgContext&	context = dlg->GetContext ();

	//
	// Fill in our context fields.
	//

	context.fFn			= fn;
	context.fDlg		= (EmDlgRef) dlg;
	context.fDlgID		= dlgID;
	context.fUserData	= data;

	//
	// Host-specific initialization.
	//

	EmDlg::HostDialogInit (context);

	//
	// Cross-platform initialization.
	//

	{
		StStopListening	stopper (dlg);

		if (context.Init () == kDlgResultClose)
		{
			delete dlg;
			return NULL;
		}
	}

	return w;
}


static void PrvShowDialog (LWindow* dlg)
{
	//
	// Show the window.
	//

	dlg->Show ();

	//
	// Activate and focus the dialog NOW.  In progress dialogs (like the
	// "Import database" progress dialog), the items aren't drawn immediately
	// unless we call Draw, and they'll appear inactive unless we call Activate.
	//

	dlg->Activate ();
	dlg->FocusDraw (dlg);
	dlg->Draw (NULL);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunGetFile
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to load
 *				the file.
 *
 * PARAMETERS:	typeList - types of files user is allowed to select.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunGetFile (const void* parameters)
{
	EmAssert (parameters);
	DoGetFileParameters&	data = *(DoGetFileParameters*) parameters;

	UNavServicesDialogs::LFileChooser	chooser;

	// Set the filter proc

	gFileTypes		= data.fFilterList;
	gHostFileTypes	= ::PrvConvertTypeList (data.fFilterList);
	chooser.SetObjectFilterProc (PrvFileFilterProc);

	// Set the default path is one is specified.

	if (data.fDefaultPath.Exists ())
	{
		chooser.SetDefaultLocation (data.fDefaultPath.GetFSSpec (), false);
	}

	// Set the prompt if one is specified.

	NavDialogOptions*	options = chooser.GetDialogOptions();

	if (data.fPrompt.size ())
	{
		if (options)
		{
			LStr255		message (data.fPrompt.c_str ());
			LString::CopyPStr (message, options->message);
		}
	}

	if (options)
	{
		options->dialogOptionFlags |= kNavNoTypePopup;
		options->dialogOptionFlags |= kNavDontAutoTranslate;
		options->dialogOptionFlags |= kNavDontAddTranslateItems;
		options->dialogOptionFlags &= ~kNavAllFilesInPopup;
		options->dialogOptionFlags &= ~kNavAllowStationery;
		options->dialogOptionFlags &= ~kNavAllowPreviews;
		options->dialogOptionFlags &= ~kNavAllowMultipleFiles;
		options->dialogOptionFlags &= ~kNavAllowInvisibleFiles;
		options->dialogOptionFlags &= ~kNavDontResolveAliases;
		options->dialogOptionFlags &= ~kNavSelectDefaultLocation;
		options->dialogOptionFlags &= ~kNavSelectAllReadableItem;
	}

	// Ask the user for the file.

	LFileTypeList	list (fileTypes_All);
	bool			answer = chooser.AskOpenFile (list);

	// If they pressed OK, get the file spec and turn it into an EmFileRef.

	if (answer)
	{
		FSSpec	spec;
		chooser.GetFileSpec (1, spec);
		data.fResult = EmFileRef (spec);

		return kDlgItemOK;
	}

	return kDlgItemCancel;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunGetFileList
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to load
 *				the file.
 *
 * PARAMETERS:	typeList - types of files user is allowed to select.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunGetFileList (const void* parameters)
{
	EmAssert (parameters);
	DoGetFileListParameters&	data = *(DoGetFileListParameters*) parameters;

	UNavServicesDialogs::LFileChooser	chooser;

	// Set the filter proc

	gFileTypes		= data.fFilterList;
	gHostFileTypes	= ::PrvConvertTypeList (data.fFilterList);
	chooser.SetObjectFilterProc (PrvFileFilterProc);

	// Set the default path is one is specified.

	if (data.fDefaultPath.Exists ())
	{
		chooser.SetDefaultLocation (data.fDefaultPath.GetFSSpec (), false);
	}

	// Set the prompt if one is specified.

	NavDialogOptions*	options = chooser.GetDialogOptions();

	if (data.fPrompt.size ())
	{
		if (options)
		{
			LStr255		message (data.fPrompt.c_str ());
			LString::CopyPStr (message, options->message);
		}
	}

	if (options)
	{
		options->dialogOptionFlags |= kNavNoTypePopup;
		options->dialogOptionFlags |= kNavDontAutoTranslate;
		options->dialogOptionFlags |= kNavDontAddTranslateItems;
		options->dialogOptionFlags &= ~kNavAllFilesInPopup;
		options->dialogOptionFlags &= ~kNavAllowStationery;
		options->dialogOptionFlags &= ~kNavAllowPreviews;
		options->dialogOptionFlags |= kNavAllowMultipleFiles;
		options->dialogOptionFlags &= ~kNavAllowInvisibleFiles;
		options->dialogOptionFlags &= ~kNavDontResolveAliases;
		options->dialogOptionFlags &= ~kNavSelectDefaultLocation;
		options->dialogOptionFlags &= ~kNavSelectAllReadableItem;
	}

	// Ask the user for the file.

	LFileTypeList	list (fileTypes_All);
	bool			answer = chooser.AskOpenFile (list);

	// If they pressed OK, get the file specs and turn it into an
	// EmFileRefList.

	if (answer)
	{
		SInt32	numFiles = chooser.GetNumberOfFiles ();
		for (SInt32 ii = 0; ii < numFiles; ++ii)
		{
			FSSpec	spec;
			chooser.GetFileSpec (ii + 1, spec);
			data.fResults.push_back (EmFileRef (spec));
		}

		return kDlgItemOK;
	}

	return kDlgItemCancel;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunPutFile
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to save
 *				the file.
 *
 * PARAMETERS:	defName - default name for the file to be saved.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunPutFile (const void* parameters)
{
	EmAssert (parameters);
	DoPutFileParameters&	data = *(DoPutFileParameters*) parameters;

	UNavServicesDialogs::LFileDesignator	designator;

	// Set the type (if we can).

	HostFileTypeList	hostFileTypeList = ::PrvConvertTypeList (data.fFilterList);
	if (hostFileTypeList.size () > 0)
	{
		designator.SetFileType (hostFileTypeList[0]);
	}

	// Set the default path is one is specified.

	if (data.fDefaultPath.Exists ())
	{
		designator.SetDefaultLocation (data.fDefaultPath.GetFSSpec (), false);
	}

	// Set the prompt if one is specified.

	NavDialogOptions*	options = designator.GetDialogOptions();

	if (data.fPrompt.size ())
	{
		if (options)
		{
			LStr255		message (data.fPrompt.c_str ());
			LString::CopyPStr (message, options->message);
		}
	}

	if (options)
	{
		options->dialogOptionFlags |= kNavNoTypePopup;
		options->dialogOptionFlags |= kNavDontAutoTranslate;
		options->dialogOptionFlags |= kNavDontAddTranslateItems;
		options->dialogOptionFlags &= ~kNavAllFilesInPopup;
		options->dialogOptionFlags &= ~kNavAllowStationery;
		options->dialogOptionFlags &= ~kNavAllowPreviews;
		options->dialogOptionFlags &= ~kNavAllowMultipleFiles;
		options->dialogOptionFlags &= ~kNavAllowInvisibleFiles;
		options->dialogOptionFlags &= ~kNavDontResolveAliases;
		options->dialogOptionFlags &= ~kNavSelectDefaultLocation;
		options->dialogOptionFlags &= ~kNavSelectAllReadableItem;
	}

	// Ask the user for the file.

	LStr255	defaultNamePStr (data.fDefaultName.c_str ());
	bool	answer = designator.AskDesignateFile (defaultNamePStr);

	// If they pressed OK, get the file spec and turn it into an EmFileRef.

	if (answer)
	{
		FSSpec	spec;
		designator.GetFileSpec (spec);
		data.fResult = EmFileRef (spec);

		return kDlgItemOK;
	}

	return kDlgItemCancel;
}

	
/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunGetDirectory
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				directory from the user.
 *
 * PARAMETERS:	result - chosen directory is returned in this.
 *				prompt - prompt to display in the dialog box.
 *				defaultPath - starting position.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunGetDirectory (const void* parameters)
{
	EmAssert (parameters);
	DoGetDirectoryParameters&	data = *(DoGetDirectoryParameters*) parameters;

	UNavServicesDialogs::LFileChooser	chooser;

	// Set the default path is one is specified.

	if (data.fDefaultPath.Exists ())
	{
		chooser.SetDefaultLocation (data.fDefaultPath.GetFSSpec (), true);
	}

	// Set the prompt if one is specified.

	if (data.fPrompt.size ())
	{
		NavDialogOptions*	options = chooser.GetDialogOptions();

		if (options)
		{
			LStr255		message (data.fPrompt.c_str ());
			LString::CopyPStr (message, options->message);
		}
	}

	// Ask the user for the folder.

	FSSpec	outFileSpec;
	SInt32	outFolderDirID;
	bool	answer = chooser.AskChooseFolder (outFileSpec, outFolderDirID);

	if (answer)
	{
		data.fResult = EmDirRef (outFileSpec);
		return kDlgItemOK;
	}

	return kDlgItemCancel;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunAboutBox
 *
 * DESCRIPTION:	Platform-specific routine for displaying the About Box.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunAboutBox (const void* parameters)
{
	UNUSED_PARAM (parameters);

	CAboutDialog::DoAboutBox (gHostApplication);

	return kDlgItemOK;
}


/***********************************************************************
 *
 * FUNCTION:	HostRunSessionSave
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgItemID	EmDlg::HostRunSessionSave (const void* parameters)
{
	EmAssert (parameters);
	DoSessionSaveParameters&	data = *(DoSessionSaveParameters*) parameters;

	LStr255	appNamePStr (data.fAppName.c_str ());
	LStr255	docNamePStr (data.fDocName.c_str ());

	SInt16	result = UNavServicesDialogs::AskSaveChanges (docNamePStr,
														  appNamePStr,
														  data.fQuitting);

	if (result == kNavAskSaveChangesSave)
		return kDlgItemYes;

	if (result == kNavAskSaveChangesCancel)
		return kDlgItemCancel;

	// else kNavAskSaveChangesDontSave
	return kDlgItemNo;
}


/***********************************************************************
 *
 * FUNCTION:	HostRunDialog
 *
 * DESCRIPTION:	Common routine that handles the creation of a dialog,
 *				initializes it (via the dialog handler), fetches events,
 *				handles events (via the dialog handler), and closes
 *				the dialog.
 *
 * PARAMETERS:	fn - the custom dialog handler
 *				userData - custom data passed back to the dialog handler.
 *				dlgID - ID of dialog to create.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/
 
EmDlgItemID EmDlg::HostRunDialog (const void* parameters)
{
	//
	// Get the information on the dialog to create.
	//

	EmAssert (parameters);
	RunDialogParameters&	data = *(RunDialogParameters*) parameters;

	//
	// Create the dialog.
	//

	EmDialogBox*			dlg = (EmDialogBox*) ::PrvCreateDialog (
									data.fFn, data.fUserData, data.fDlgID);
	if (!dlg)
	{
		return kDlgItemNone;
	}

	//
	// Create the handler for it.  Be sure to associate the handler with
	// the dialog *after* the dialog is created bug *before* the dialog
	// is made visible.  If the dialog is visible before handing it off
	// to the handler, the commander chain will have already been
	// established.  When the handler inserts itself into that chain, it
	// doesn't do anything to put itself on duty.  With an off-duty
	// commander in the chain, problems can occur.  In particular, because
	// IsOnDuty returns false for the handler, it won't call UpdateMenus
	// in its Dodialog method.
	//

	EmDialogHandler			handler (dlg, gHostApplication);

	//
	// Tell the dialog about the handler so that the handler can
	// be notified when the dialog wants to close.
	//

	dlg->SetHandler (&handler);

	// Show the dialog.

	::PrvShowDialog (dlg);

	//
	// Go into a loop, handling events for this dialog.
	//

	while (1)
	{
		MessageT	msg = handler.DoDialog ();

		if (msg == msg_Cancel)
		{
			break;
		}
	}

	//
	// Get the result and return.  The EmDialogHandler will
	// dispose of the dialog (which will also send the Destroy
	// message to its callback).
	//

	EmDlgItemID	itemID = dlg->GetContext ().fItemID;

	return itemID;
}


/***********************************************************************
 *
 * FUNCTION:	HostDialogOpen
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

EmDlgRef EmDlg::HostDialogOpen (EmDlgFn fn, void* data, EmDlgID dlgID)
{
	LWindow*	dlg = ::PrvCreateDialog (fn, data, dlgID);
	::PrvShowDialog (dlg);
	return (EmDlgRef) dlg;
}


/***********************************************************************
 *
 * FUNCTION:	HostDialogClose
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	.
 *
 ***********************************************************************/

void EmDlg::HostDialogClose (EmDlgRef dlg)
{
	EmDialogBox*	dialog = static_cast <EmDialogBox*> (dlg);
	if (!dialog)
		return;

	dialog->ProcessCommand (cmd_Close);
}


/***********************************************************************
 *
 * FUNCTION:	HostStartIdling
 *
 * DESCRIPTION:	Enable the dialog to receive idle time.
 *
 * PARAMETERS:	context - dialog context data.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmDlg::HostStartIdling (EmDlgContext& context)
{
	EmDialogBox*	dialog = static_cast <EmDialogBox*> (context.fDlg);
	if (!dialog)
		return;

	dialog->StartIdling ();
}


/***********************************************************************
 *
 * FUNCTION:	HostStopIdling
 *
 * DESCRIPTION:	Disable the dialog from receiving idle time.
 *
 * PARAMETERS:	context - dialog context data.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmDlg::HostStopIdling (EmDlgContext& context)
{
	EmDialogBox*	dialog = static_cast <EmDialogBox*> (context.fDlg);
	if (!dialog)
		return;

	dialog->StopIdling ();
}


/***********************************************************************
 *
 * FUNCTION:	HostDialogInit
 *
 * DESCRIPTION:	Perform any platform-specific initialization.
 *
 * PARAMETERS:	context - dialog context data.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmDlg::HostDialogInit (EmDlgContext& context)
{
	EmDialogBox*	dialog = static_cast <EmDialogBox*> (context.fDlg);
	if (!dialog)
		return;

	if (context.fDlgID == kDlgCommonDialog)
	{
		dialog->AddAttachment (new LCmdCAttachment (dialog, kDlgItemCmnButtonCopy, true));
	}

	//
	// Ensure that everybody is hooked up.  When StDialogHandler calls
	// LinkListenerToBroadcasters, it hooks up only panes whose IDs are
	// in the 'RidL' resource.  That resource contains only LControl sub-classes.
	// Since the 'RidL' resource is an artifact of the days before dynamic
	// casting, we can implement a better solution that doesn't require our
	// panes to be LControl sub-classes (as LTextViews are not).
	//
	// Calling LinkListenerToPanes duplicates much of the work already
	// done by StDialogHandler when it called UReanimator::
	// LinkListenerToBroadcasters.  However, LinkListenerToPanes does
	// more, and it's OK for the duplicated work to be done twice.
	//

	::LinkListenerToPanes (dialog, dialog);

	//
	// Adjust row heights, etc.
	//

	::PrvInitializeLColumnViews (dialog);
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetDlgBounds
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetDlgBounds (EmDlgRef dlg, const EmRect& bounds)
{
	EmDialogBox*	dialog = static_cast <EmDialogBox*> (dlg);
	if (!dialog)
		return;

	EmRect	oldBounds = EmDlg::GetDlgBounds (dlg);

	if (oldBounds.TopLeft () != bounds.TopLeft ())
	{
		dialog->MoveWindowTo (bounds.fLeft, bounds.fTop);
	}

	if (oldBounds.Size () != bounds.Size ())
	{
		dialog->ResizeWindowTo (bounds.Width (), bounds.Height ());
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetDlgBounds
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmRect EmDlg::GetDlgBounds (EmDlgRef dlg)
{
	EmRect	result (0, 0, 0, 0);

	EmDialogBox*	dialog = static_cast <EmDialogBox*> (dlg);
	if (!dialog)
		goto Exit;

	Rect	bounds;
	dialog->GetGlobalBounds (bounds);
	
	result = bounds;

Exit:
	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::CenterDlg
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::CenterDlg (EmDlgRef dlg)
{
	EmDialogBox*	dialog = static_cast <EmDialogBox*> (dlg);
	if (!dialog)
		return;

	BitMap			screenBits;
	::GetQDGlobalsScreenBits (&screenBits);

	EmRect			screenBounds (screenBits.bounds);

	SDimension16	frameSize;
	dialog->GetFrameSize (frameSize);
	
	SPoint32		location;
	location.h = (screenBounds.Width () - frameSize.width) / 2;
	location.v = (screenBounds.Height () - frameSize.height) / 3;

	dialog->MoveWindowTo (location.h, location.v);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetDlgDefaultButton
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetDlgDefaultButton (EmDlgContext& context, EmDlgItemID item)
{
	EmDialogBox*	dialog = static_cast <EmDialogBox*> (context.fDlg);
	if (!dialog)
		return;

	dialog->SetDefaultButton (::PrvFromDlgItemID (item));
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetDlgCancelButton
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetDlgCancelButton (EmDlgContext& context, EmDlgItemID item)
{
	EmDialogBox*	dialog = static_cast <EmDialogBox*> (context.fDlg);
	if (!dialog)
		return;

	dialog->SetCancelButton (::PrvFromDlgItemID (item));
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemMin
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemMin (EmDlgRef dlg, EmDlgItemID item, long minValue)
{
	LControl*	control;
	::FindPane (dlg, item, control);

	if (control)
	{
		if (minValue != control->GetMinValue ())
			control->SetMinValue (minValue);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemValue
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemValue (EmDlgRef dlg, EmDlgItemID item, long value)
{
	LPane*	pane;
	::FindPane (dlg, item, pane);

	if (pane)
	{
		LPopupButton*	popup = dynamic_cast<LPopupButton*>(pane);

		if (popup)
		{
			++value;	// menus are 1-based.

			if (value != pane->GetValue ())
				popup->SetValue (value);
		}
		else
		{
			if (value != pane->GetValue ())
				pane->SetValue (value);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemMax
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemMax (EmDlgRef dlg, EmDlgItemID item, long maxValue)
{
	LControl*	control;
	::FindPane (dlg, item, control);

	if (control)
	{
		if (maxValue != control->GetMaxValue ())
			control->SetMaxValue (maxValue);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemBounds
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemBounds (EmDlgRef dlg, EmDlgItemID item, const EmRect& bounds)
{
	LPane*	pane;
	::FindPane (dlg, item, pane);

	if (pane)
	{
		pane->PlaceInSuperFrameAt (bounds.fLeft, bounds.fTop, true);
		pane->ResizeFrameTo (bounds.Width (), bounds.Height (), true);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemText
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemText (EmDlgRef dlg, EmDlgItemID item, string str)
{
	LPane*	pane;
	::FindPane (dlg, item, pane);
	if (!pane)
		return;

	// Check for LTextEditViews, which have a special method
	// for allowing > 255 bytes of text.

	LTextEditView*	textEdit = dynamic_cast<LTextEditView*> (pane);

	if (textEdit)
	{
		textEdit->SetTextPtr ((Ptr) str.c_str (), str.size (), NULL);
	}
	else
	{
		string	oldText = EmDlg::GetItemText (dlg, item);
		if (oldText != str)
		{
			LStaticText*	staticText = dynamic_cast<LStaticText*> (pane);

			if (staticText)
			{
				staticText->SetText (const_cast<char*>(str.c_str ()),
									str.size ());
			}
			else
			{
				LStr255	newText (str.c_str ());
				pane->SetDescriptor (newText);
			}
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetItemValue
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

long EmDlg::GetItemValue (EmDlgRef dlg, EmDlgItemID item)
{
	long	result = 0;

	LPane*	pane;
	::FindPane (dlg, item, pane);
	if (!pane)
		goto Exit;

	// These views use StringToNum in order to convert a text
	// number representation into a binary one.  However, that
	// supports only decimal numbers.  In order to support
	// hex and octal numbers, convert it ourself.

	if (dynamic_cast<LStaticText*>(pane)	||
		dynamic_cast<LEditText*>(pane)		||
		dynamic_cast<LTextEditView*>(pane)	||
		dynamic_cast<LEditField*>(pane))
	{
		string	text = EmDlg::GetItemText (dlg, item);
		long	num;
		if (EmDlg::StringToLong (text.c_str (), &num))
			return num;
	}

	result = pane->GetValue ();

	// If a menu, make the result 0-based instead of 1-based.
	if (dynamic_cast<LPopupButton*>(pane) != NULL)
	{
		--result;
	}

Exit:
	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetItemBounds
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmRect EmDlg::GetItemBounds (EmDlgRef dlg, EmDlgItemID item)
{
	EmRect	result (0, 0, 0, 0);

	LPane*	pane;
	::FindPane (dlg, item, pane);
	if (!pane)
		goto Exit;

	SPoint32		location;
	pane->GetFrameLocation (location);

	SDimension16	size;
	pane->GetFrameSize (size);

	result.Set (location.h, location.v,
				location.h + size.width,
				location.v + size.height);

Exit:
	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetItemText
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

string EmDlg::GetItemText (EmDlgRef dlg, EmDlgItemID item)
{
	string	result;

	LPane*	pane;
	::FindPane (dlg, item, pane);
	if (!pane)
		goto Exit;

	// Check for LTextEditViews, which have a special method
	// for allowing > 255 bytes of text.
	// LControlPane also has a method for >255 character strings.

	LTextEditView*		textEdit = dynamic_cast<LTextEditView*> (pane);
	LStaticText*		staticText = dynamic_cast<LStaticText*> (pane);

	if (textEdit)
	{
		Handle	theText = textEdit->GetTextHandle ();
		if (!theText)
			return result;

		StHandleLocker	lock (theText);
		result.assign ((const char*) *theText,  ::GetHandleSize (theText));
	}

	else if (staticText)
	{
		// Get the size by passing in a zero-size buffer.

		Size	stringSize;
		staticText->GetText (&result[0], 0, &stringSize);

		// Now get the text.

		result.resize (stringSize);
		staticText->GetText (&result[0], stringSize, NULL);
	}

	else
	{
		LStr255	str;
		pane->GetDescriptor (str);

		result.assign ((char*) &str[1], str[0]);
	}

Exit:
	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::EnableItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::EnableItem (EmDlgRef dlg, EmDlgItemID item)
{
	LPane*	pane;
	::FindPane (dlg, item, pane);

	if (pane)
	{
		pane->Enable ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::DisableItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::DisableItem (EmDlgRef dlg, EmDlgItemID item)
{
	LPane*	pane;
	::FindPane (dlg, item, pane);

	if (pane)
	{
		pane->Disable ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::ShowItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::ShowItem (EmDlgRef dlg, EmDlgItemID item)
{
	LPane*	pane;
	::FindPane (dlg, item, pane);

	if (pane)
	{
		pane->Show ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HideItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::HideItem (EmDlgRef dlg, EmDlgItemID item)
{
	LPane*	pane;
	::FindPane (dlg, item, pane);

	if (pane)
	{
		pane->Hide ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::ClearMenu
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::ClearMenu (EmDlgRef dlg, EmDlgItemID item)
{
	LPopupButton*	popup;
	::FindPane (dlg, item, popup);

	if (popup)
	{
		popup->DeleteAllMenuItems ();
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::AppendToMenu
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::AppendToMenu (EmDlgRef dlg, EmDlgItemID item, const StringList& strList)
{
	LPopupButton*	popup;
	::FindPane (dlg, item, popup);

	if (popup)
	{
		StringList::const_iterator	iter = strList.begin ();
		while (iter != strList.end ())
		{
			if (iter->empty ())
			{
				popup->AppendMenu ("\p(-", false);
			}
			else
			{
				LStr255	str (iter->c_str());
				popup->AppendMenu (str, true);
			}

			++iter;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::EnableMenuItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::EnableMenuItem (EmDlgRef dlg, EmDlgItemID item, long menuItem)
{
	LPopupButton*	popup;
	::FindPane (dlg, item, popup);

	if (popup)
	{
		::EnableMenuItem (popup->GetMacMenuH (), menuItem + 1);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::DisableMenuItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::DisableMenuItem (EmDlgRef dlg, EmDlgItemID item, long menuItem)
{
	LPopupButton*	popup;
	::FindPane (dlg, item, popup);

	if (popup)
	{
		::DisableMenuItem (popup->GetMacMenuH (), menuItem + 1);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::ClearList
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::ClearList (EmDlgRef dlg, EmDlgItemID item)
{
	LTextColumn*	listView;
	::FindPane (dlg, item, listView);

	if (listView)
	{
		listView->RemoveRows (max_Int32, 0, true);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::AppendToList
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::AppendToList (EmDlgRef dlg, EmDlgItemID item, const StringList& strList)
{
	LTextColumn*	listView;
	::FindPane (dlg, item, listView);

	if (listView)
	{
		TableIndexT	rows;
		TableIndexT	cols;
		listView->GetTableSize (rows, cols);

		StringList::const_iterator	iter = strList.begin ();
		while (iter != strList.end ())
		{
			listView->InsertRows (1, rows++, iter->c_str (), iter->size (), true);

			++iter;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SelectListItems
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SelectListItems (EmDlgRef dlg, EmDlgItemID item, const EmDlgListIndexList& itemList)
{
	LTextColumn*	listView;
	::FindPane (dlg, item, listView);

	if (listView)
	{
		EmDlgListIndexList::const_iterator	iter = itemList.begin ();
		while (iter != itemList.end ())
		{
			STableCell	cell (*iter + 1, 1);
			listView->SelectCell (cell);
			listView->ScrollCellIntoFrame (cell);
			++iter;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::UnselectListItems
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::UnselectListItems (EmDlgRef dlg, EmDlgItemID item, const EmDlgListIndexList& itemList)
{
	LTextColumn*	listView;
	::FindPane (dlg, item, listView);

	if (listView)
	{
		EmDlgListIndexList::const_iterator	iter = itemList.begin ();
		while (iter != itemList.end ())
		{
			STableCell	cell (*iter + 1, 1);
			listView->UnselectCell (cell);
			++iter;
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetSelectedItems
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::GetSelectedItems (EmDlgRef dlg, EmDlgItemID item, EmDlgListIndexList& itemList)
{
	LTextColumn*	listView;
	::FindPane (dlg, item, listView);

	if (listView)
	{
		STableCell	cell (0, 0);
		while (listView->GetNextSelectedCell (cell))
		{
			itemList.push_back (cell.row - 1);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	GetTextHeight
 *
 * DESCRIPTION:	Determine the height the text would be if fitted into a
 *				rectangle with the given width.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	New text height
 *
 ***********************************************************************/

int EmDlg::GetTextHeight (EmDlgRef dlg, EmDlgItemID item, const string& s)
{
	SInt16			textHeight = 0;
	LControlPane*	control;
	::FindPane (dlg, item, control);

	if (control)
	{
		EmDlg::SetItemText (dlg, item, s);

		control->GetDataTag(0, kControlStaticTextTextHeightTag,
					sizeof (textHeight), &textHeight, NULL);
	}

	return textHeight;
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	PrvFileFilterProc
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

pascal Boolean PrvFileFilterProc (AEDesc* theItem, void* info,
							NavCallBackUserData /*callBackUD*/,
							NavFilterModes /*filterMode*/)
{
	NavFileOrFolderInfo*	theInfo = (NavFileOrFolderInfo*) info;

	// Accept non-file system thingies and folders.

	if (theItem->descriptorType != typeFSS)
		return true;

	if (theInfo->isFolder)
		return true;

	// Reject anything without additional data.

	if (theItem->dataHandle == NULL)
		return false;

	// If we are filtering on "All Files" and *only* "All Files",
	// accept it.  If there are other type-specific filters in the
	// list, use them and ignore the "All Files" filter.

	if (gFileTypes.size () == 1 && gFileTypes.front () == kFileTypeAll)
		return true;

	// Check the fast way, first.  That is, check by mere file type
	// before creating an EmFileRef and calling its IsType method.
	// (Actually, including this test is not really noticable.  But
	//  perhaps it would be if we were using a network drive.)

	if (1)
	{
		HostFileTypeList::iterator	iter = gHostFileTypes.begin ();
		while (iter != gHostFileTypes.end ())
		{
			if (theInfo->fileAndFolder.fileInfo.finderInfo.fdType == (*iter))
				return true;

			++iter;
		}
	}

	// OK, now check the slower way.  This way involves creating
	// an alias, so it's kind of time intensive.

	{
		FSSpec		spec = **(FSSpec**) (theItem->dataHandle);
		EmFileRef	ref (spec);
		
		EmFileTypeList::iterator	iter = gFileTypes.begin ();
		while (iter != gFileTypes.end ())
		{
			if (*iter != kFileTypeAll)
			{
				if (ref.IsType (*iter))
					return true;

				if (*iter == kFileTypePalmAll)
				{
					if (ref.IsType (kFileTypePalmApp))
						return true;

					if (ref.IsType (kFileTypePalmDB))
						return true;

					if (ref.IsType (kFileTypePalmQA))
						return true;
				}
			}

			++iter;
		}
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:	PrvConvertTypeList
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

HostFileTypeList PrvConvertTypeList (const EmFileTypeList& typeList)
{
	HostFileTypeList	hostList;

	EmFileTypeList::const_iterator	iter = typeList.begin ();

	while (iter != typeList.end ())
	{
		switch (*iter)
		{
			default:
				hostList.push_back (::PrvTypeToOSType (*iter));
				break;

			case kFileTypePalmAll:
				break;

			case kFileTypeAll:
				break;
		}

		++iter;
	}

	return hostList;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetROMScript
 *
 * DESCRIPTION:	Map from the ROM's character encoding to a Mac OS script id.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	MacOS script id that corresponds to the Palm OS ROM's
 *				character encoding, or smMacRoman if unknown.
 *
 ***********************************************************************/

static short PrvGetROMScript (void)
{
	switch (EmPatchState::GetROMCharEncoding ())
	{
		case charEncodingPalmSJIS:
			return smJapanese;

		case charEncodingBig5:
		case charEncodingBig5_HKSCS:
		case charEncodingBig5Plus:
		case charEncodingPalmBig5:
			return smTradChinese;

		case charEncodingGB2312:
		case charEncodingGBK:
		case charEncodingPalmGB:
			return smSimpChinese;
	}

	return smRoman;
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetScript
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Mac OS script id to use for drawing names of ROM components.
 *
 ***********************************************************************/

static short PrvGetScript (void)
{
	short targetScript = PrvGetROMScript ();
	
	if (::GetScriptVariable (targetScript, smScriptEnabled) != 0)
	{
		return targetScript;
	}
	
	return smSystemScript;
}


/***********************************************************************
 *
 * FUNCTION:	GetListFontAndSize
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void GetListFontAndSize (short* font, short* size)
{
	short	script	= ::PrvGetScript ();

	long	fontAndSize;
	fontAndSize = ::GetScriptVariable (script, smScriptSmallFondSize);

	// Keith says that using the small font doesn't work for Japanese
	// systems - the text shows up as Jinglish. To protect against that,
	// make sure the font we got maps to the script we asked for. If not,
	// then switch to the system font.
	if (::FontToScript(HiWord (fontAndSize)) != script)
	{
		fontAndSize = ::GetScriptVariable (script, smScriptSysFondSize);
	}
	
	*font = HiWord (fontAndSize);
	*size = LoWord (fontAndSize);
}


/***********************************************************************
 *
 * FUNCTION:	GetDialogFontAndSize
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void GetDialogFontAndSize (short* font, short* size)
{
	long	fontAndSize;

	/*
		On Mac OS 8.1, standard US version:

			Verb					 Font	  Size
			=====================	 ====	  ====
			smScriptMonoFondSize		4		 9		(Monaco 9)
			smScriptPrefFondSize		3		12		(Geneva 12)
			smScriptSmallFondSize		3		 9		(Geneva 9)
			smScriptSysFondSize		  7D2		12		(Charcoal 12)
			smScriptAppFondSize			3		12		(Geneva 12)
			smScriptHelpFondSize		3		 9		(Geneva 9)
	*/

	fontAndSize = ::GetScriptVariable (::PrvGetScript(), smScriptSysFondSize);

	*font = HiWord (fontAndSize);
	*size = LoWord (fontAndSize);
}


/***********************************************************************
 *
 * FUNCTION:	GetListCellHeight
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

short GetListCellHeight (void)
{
	GrafPtr	thePort;
	::GetPort (&thePort);

	short	oldFont = ::GetPortTextFont (thePort);
	short	oldSize = ::GetPortTextSize (thePort);

	short	font;
	short	size;

	::GetListFontAndSize (&font, &size);

	::TextFont (font);
	::TextSize (size);

	FontInfo	fontInfo;
	::GetFontInfo (&fontInfo);

	::TextFont (oldFont);
	::TextSize (oldSize);

	// LTextColumnView insets by 2 when drawing.  Technically, we should
	// add 4 so that we don't draw in the lower inset area, but that leads
	// to list views that seem too spaced out.

	return fontInfo.ascent + fontInfo.descent + fontInfo.leading + 2;
}


#pragma mark -

// ===========================================================================
//  EmDialogHandler										EmDialogHandler 
// ===========================================================================

EmDialogHandler::EmDialogHandler(ResIDT inDialogResID, LCommander* inSuper) :
	StDialogHandler (inDialogResID, inSuper)
{
}


EmDialogHandler::EmDialogHandler(LWindow* inWindow, LCommander* inSuper) :
	StDialogHandler (inWindow, inSuper)
{
}


EmDialogHandler::~EmDialogHandler()
{
}

MessageT EmDialogHandler::DoDialog()
{
	EventRecord macEvent;

	if (IsOnDuty()) {
		UEventMgr::GetMouseAndModifiers(macEvent);
		AdjustCursor(macEvent);
	}

	SetUpdateCommandStatus(false);
	mMessage = msg_Nothing;

	// ===== Only this line was changed from the standard DoDialog
	// ===== method.  High-level events are masked out.

	Boolean gotEvent = ::WaitNextEvent(everyEvent - highLevelEventMask,
								&macEvent, mSleepTime, mMouseRgn);

		// Let Attachments process the event. Continue with normal
		// event dispatching unless suppressed by an Attachment.

	if (LEventDispatcher::ExecuteAttachments(msg_Event, &macEvent)) {
		if (gotEvent) {
			DispatchEvent(macEvent);
		} else {
			UseIdleTime(macEvent);
		}
	}

									// Repeaters get time after every event
	LPeriodical::DevoteTimeToRepeaters(macEvent);

									// Update status of menu items
	if (IsOnDuty() && GetUpdateCommandStatus()) {
		UpdateMenus();
	}

	return mMessage;
}


#pragma mark -

// ===========================================================================
//  LCmdCAttachment										LCmdCAttachment 
// ===========================================================================

LCmdCAttachment::LCmdCAttachment (	LListener* listener,
									EmDlgItemID paneID,
									Boolean inExecuteHost) :
	LAttachment (msg_KeyPress, inExecuteHost),
	fListener (listener),
	fPaneID (paneID)
{
}


// ---------------------------------------------------------------------------
//	 ExecuteSelf												   [protected]
// ---------------------------------------------------------------------------

void
LCmdCAttachment::ExecuteSelf(
	MessageT	/* inMessage */,
	void		*ioParam)
{
	mExecuteHost = true;

	EventRecord*	event = static_cast<EventRecord*>(ioParam);
	if ((event->modifiers & cmdKey) != 0)
	{
		char	ch = event->message & charCodeMask;
		if (ch == 'c' || ch == 'C')
		{
			EmAssert (fListener);
			fListener->ListenToMessage (::PrvFromDlgItemID (fPaneID), NULL);

			mExecuteHost = false;
		}
	}
}


#pragma mark -

// ---------------------------------------------------------------------------
//	 EmDialogBox					Default Constructor		  [public]
// ---------------------------------------------------------------------------

EmDialogBox::EmDialogBox() :
	LDialogBox (),
	LPeriodical (),
	fHandler (NULL)
{
}


// ---------------------------------------------------------------------------
//	 EmDialogBox					Stream Constructor		  [public]
// ---------------------------------------------------------------------------

EmDialogBox::EmDialogBox(LStream* inStream) :
	LDialogBox (inStream),
	LPeriodical (),
	fHandler (NULL)
{
}


// ---------------------------------------------------------------------------
//	 ~EmDialogBox					Destructor				  [public]
// ---------------------------------------------------------------------------

EmDialogBox::~EmDialogBox()
{
	StStopListening	stopper (this);

	fContext.Destroy ();
}


// ---------------------------------------------------------------------------
//	 ListenToMessage										  [public]
// ---------------------------------------------------------------------------

void EmDialogBox::ListenToMessage (MessageT inMessage, void* /*ioParam*/)
{
	StStopListening	stopper (this);

	if (inMessage >= kDialogItemIDBase & inMessage < kDialogItemIDDblClickBase)
	{
		EmDlgItemID	itemID = ::PrvToDlgItemID (inMessage);

		if (fContext.Event (itemID) == kDlgResultClose)
		{
			this->AttemptClose ();
		}
	}
}


// ---------------------------------------------------------------------------
//	 SpendTime												  [protected]
// ---------------------------------------------------------------------------

void EmDialogBox::SpendTime (const EventRecord& /*inMacEvent*/)
{
	StStopListening	stopper (this);

	if (fContext.Idle () == kDlgResultClose)
	{
		this->AttemptClose ();
	}
}


// ---------------------------------------------------------------------------
//	 AttemptClose											  [public]
// ---------------------------------------------------------------------------

void EmDialogBox::AttemptClose (void)
{
	// See if this window is managed by an EmDialogHandler.
	// If so, tell the dialog handler we want to close.
	
	if (fHandler)
	{
		fHandler->ListenToMessage (msg_Cancel, NULL);
	}

	// Otherwise, close ourself.

	else
	{
		LDialogBox::AttemptClose ();
	}
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 CAboutDialog::DoAboutBox	[ static ]
// ---------------------------------------------------------------------------

void CAboutDialog::DoAboutBox (LCommander* inSuper)
{
	CAboutDialog	handler (wnd_About, inSuper);

	// Find the tab panel so that we can get setup to listen to it.
	LWindow*	dlg = handler.GetDialog ();
	EmAssert (dlg);

	LTabsControl*	tabs =  dynamic_cast<LTabsControl*>
		(dlg->FindPaneByID (item_aboutTabsController));
	EmAssert (tabs);

	tabs->StartBroadcasting ();
	tabs->AddListener (&handler);

	// Install the initial sample panel
	SInt32	param = 1;
	handler.ListenToMessage (item_aboutTabsController, &param);

	MessageT theMessage;

	do {
		theMessage = handler.DoDialog();
	} while (theMessage != msg_OK);
}


// ---------------------------------------------------------------------------
//		 CAboutDialog::CAboutDialog
// ---------------------------------------------------------------------------

CAboutDialog::CAboutDialog(ResIDT inDialogResID, LCommander* inSuper) :
	StDialogHandler (inDialogResID, inSuper),
	fCurrentPanel (NULL)
{
}


// ---------------------------------------------------------------------------
//		 CAboutDialog::~CAboutDialog
// ---------------------------------------------------------------------------

CAboutDialog::~CAboutDialog (void)
{
}


// ---------------------------------------------------------------------------
//		 CAboutDialog::ListenToMessage
// ---------------------------------------------------------------------------

void CAboutDialog::ListenToMessage (MessageT inMessage, void* ioParam)
{
#pragma unused (ioParam)

	mMessage = inMessage;  // store message for DoDialog to return

	if (mMessage >= item_aboutFirstURL && mMessage <= item_aboutLastURL)
	{
		Str255		theURL;

		LWindow*	dlg = this->GetDialog ();
		EmAssert (dlg);

		LPane*		button = dlg->FindPaneByID (mMessage);
		EmAssert (button);

		CURLPushButton*	urlButton = dynamic_cast<CURLPushButton*>(button);
		EmAssert (urlButton);

		urlButton->GetURL (theURL);

		long	startSel = 0;
		long	endSel = theURL[0];

		UInternetConfig::PP_ICLaunchURL ("\p", (char*) &theURL[1], theURL[0],
										&startSel, &endSel);
	}
	else if (mMessage == item_aboutTabsController)
	{
		SInt32	panelIndex = *(SInt32*) ioParam;
		switch (panelIndex)
		{
			case 1:
			{
				this->InstallPanel (about_panelMain);

				// Set the application name as a parameter.

				Errors::SetParameter ("%application", Platform::GetString (kStr_AppName));

				// Set the version string as a parameter.  Note that the returned
				// string may be empty if we can't get the version.

				Errors::SetParameter ("%version", Platform::GetShortVersionString ());

				// Format the string.

				string	appAndVersion = Errors::ReplaceParameters (kStr_AppAndVers);

				// Get a reference to the text pane that
				// will hold the string.

				LPane*		titlePane = fCurrentPanel->FindPaneByID (item_aboutTitle);
				EmAssert (titlePane);

				LStr255	appAndVers (appAndVersion.c_str ());
				titlePane->SetDescriptor (appAndVers);

				break;
			}

			case 2:	this->InstallPanel (about_panelWindows);	break;
			case 3:	this->InstallPanel (about_panelMac);		break;
			case 4:	this->InstallPanel (about_panelUAE);		break;
		}
	}
}


//-------------------------------------------------------------------------------------
//		 CAboutDialog::InstallPanel
//-------------------------------------------------------------------------------------

void			
CAboutDialog::InstallPanel (SInt16 inPanelID)
{
	LWindow*	dlg = this->GetDialog ();
	EmAssert (dlg);

	if (fCurrentPanel)
	{
		fCurrentPanel->Hide ();
		fCurrentPanel = NULL;
	}

	LView*	newPanel = dynamic_cast<LView*> (dlg->FindPaneByID (inPanelID));

	if (newPanel)
	{
		fCurrentPanel = newPanel;
		fCurrentPanel->Show ();
	}
}
