/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmDocumentMac.h"

#include "EmApplicationMac.h"	// gHostApplication
#include "EmPixMapMac.h"		// ConvertPixMapToHost
#include "EmScreen.h"			// EmScreenUpdateInfo
#include "EmSession.h"			// EmKeyEvent
#include "EmStreamFile.h"		// EmStreamFile
#include "EmWindow.h"			// gWindow
#include "ErrorHandling.h"		// Errors::ThrowIfNULL
#include "Platform.h"			// GetString
#include "ROMStubs.h"			// FtrGet
#include "Strings.r.h"			// kStr_Untitled

#if __profile__
#include <Profiler.h>
#endif

static void		PrvGetModifiers				(EmKeyEvent&, EventModifiers);
static uint16	PrvRemapMacCharacter		(uint8 macChar);

static Boolean	PrvHasPowerManager			(void);
static Boolean	PrvHasPowerBookPowerManager	(void);
static Boolean	PrvIsChargerConnected		(void);

EmDocumentMac*	gHostDocument;

// ---------------------------------------------------------------------------
//		 EmDocument::HostCreateDocument
// ---------------------------------------------------------------------------

EmDocument* EmDocument::HostCreateDocument (void)
{
	return new EmDocumentMac;
}


#pragma mark -

// ===========================================================================
//		 EmDocumentMac
// ===========================================================================

// ---------------------------------------------------------------------------
//		 EmDocumentMac::EmDocumentMac
// ---------------------------------------------------------------------------

EmDocumentMac::EmDocumentMac (void) :
	LSingleDoc (gHostApplication),
	EmDocument (),
	fHavePendingException (false),
	fPendingException (kResetSoft)
{
	EmAssert (gHostDocument == NULL);
	gHostDocument = this;
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::~EmDocumentMac
// ---------------------------------------------------------------------------

EmDocumentMac::~EmDocumentMac (void)
{
	EmAssert (mWindow == NULL);

	EmAssert (gHostDocument == this);
	gHostDocument = NULL;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmDocumentMac::UsesFileSpec
// ---------------------------------------------------------------------------
//	Returns whether the Document's File has the given FSSpec

Boolean EmDocumentMac::UsesFileSpec (const FSSpec& spec) const
{
	EmFileRef	file = this->GetFileRef ();

	if (file.IsSpecified ())
	{
		EmFileRef	other (spec);
		return file == other;
	}

	return false;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmDocumentMac::AskSaveAs
// ---------------------------------------------------------------------------
//	Ask the user to save a Document and give it a name
//
//	Returns false if the user cancels the operation
//
// Overridden to make route the operation to the common implementation. 
// This method is called from the following:
//
//		ObeyCommand for cmd_SaveAs (or unspecified cmd_Save)
//		AttemptClose
//		AttemptQuitSelf
//		DoAEClose
//		HandleAESave

Boolean EmDocumentMac::AskSaveAs (FSSpec& outFSSpec, Boolean inRecordIt)
{
	UNUSED_PARAM (inRecordIt);

	if (this->HandleSaveAs ())
	{
		EmFileRef	file = this->GetFileRef ();
		outFSSpec = file.GetFSSpec ();
		return true;
	}
	
	return false;
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::IsModified
// ---------------------------------------------------------------------------
//	Return whether a Document has been modified since it was last saved.
//	Emulator docs are always considered modified.

Boolean EmDocumentMac::IsModified(void)
{
	return true;
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::GetDescriptor
// ---------------------------------------------------------------------------
//	Pass back the name of a Document.

StringPtr EmDocumentMac::GetDescriptor (Str255 outDescriptor) const
{
	EmFileRef	file = this->GetFileRef ();
	string		name;

	if (file.IsSpecified ())
	{
		name = file.GetName ();
	}
	else
	{
		name = Platform::GetString (kStr_Untitled);
	}

	LStr255 nameStr (name.c_str ());
	LString::CopyPStr (nameStr, outDescriptor);

	return outDescriptor;
}


// ---------------------------------------------------------------------------
//	 AskSaveChanges												  [public]
// ---------------------------------------------------------------------------
//	Ask user whether to save changes before closing the Document or
//	quitting the Application
//
// Overridden in order to call EmDlg::DoSessionSave for two reasons:
//
//	(a) for consistancy with other platforms and
//	(b) so that we can implement a dialog that recognizes Cmd-D in the
//		non-NavServices case.

SInt16 EmDocumentMac::AskSaveChanges (bool inQuitting)
{
	this->MakeCurrent ();

	EmDlgItemID	item = this->EmDocument::AskSaveChanges (inQuitting);

	if (item == kDlgItemOK)
		return answer_Save;		// kNavAskSaveChangesSave;

	if (item == kDlgItemCancel)
		return answer_Cancel;	// kNavAskSaveChangesCancel;

	return answer_DontSave;		// kNavAskSaveChangesDontSave;
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::DoAESave
// ---------------------------------------------------------------------------
//	Save a Document in a particular file
//
//	This is a "Save As" operation
//
//	Subclasses should override this function to save a file and
//	should set mIsSpecified to true

void EmDocumentMac::DoAESave (FSSpec& inFileSpec, OSType)
{
	EmFileRef	file (inFileSpec);
	this->HandleSaveTo (file);
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::DoSave
// ---------------------------------------------------------------------------
//	Save a Document, which must already be "specified"
//
//	Subclasses should override to save a Document to an existing file

void EmDocumentMac::DoSave (void)
{
	EmAssert (this->GetFileRef ().IsSpecified ());

	this->HandleSave ();
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmDocumentMac::ObeyCommand
// ---------------------------------------------------------------------------

Boolean EmDocumentMac::ObeyCommand (CommandT	inCommand,
									void*		ioParam)
{
	Boolean wasHandled	= this->HandleCommand ((EmCommandID) inCommand);

	if (!wasHandled)
	{
		wasHandled = LSingleDoc::ObeyCommand (inCommand, ioParam);
	}

	return wasHandled;
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::FindCommandStatus
// ---------------------------------------------------------------------------

void EmDocumentMac::FindCommandStatus (	CommandT	inCommand,
										Boolean&	outEnabled,
										Boolean&	outUsesMark,
										UInt16& 	outMark,
										Str255		outName)
{
	// Pass the message on to the LCommander, skipping LSingleDoc and
	// LDocument.  Those classes will try to set the document-related
	// commands *their* way.  But we've already set up the menus the
	// way we want them, and EmApplicationMac::FindCommandStatus will
	// make sure they stay that way.  So don't let the document classes
	// mess things up.

	LCommander::FindCommandStatus (
		inCommand, outEnabled, outUsesMark, outMark, outName);
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::HandleKeyPress
// ---------------------------------------------------------------------------
// Notifies the emulated CPU when a key has been pressed on the desktop
// keyboard.

Boolean EmDocumentMac::HandleKeyPress (const EventRecord& iKeyEvent)
{
	// Ignore command keys.

	if (iKeyEvent.modifiers & cmdKey)
		return false;

	char			chr		= iKeyEvent.message & charCodeMask;
	uint16			key		= 0;
	SkinElementType	button	= kElement_None;

	switch (chr)
	{
		case char_LeftArrow:	key = chrLeftArrow;					break;
		case char_RightArrow:	key = chrRightArrow;				break;
		case char_UpArrow:		key = chrUpArrow;					break;
		case char_DownArrow:	key = chrDownArrow;					break;

		case char_Escape:		button = kElement_PowerButton;		break;
		case char_PageUp:		button = kElement_UpButton;			break;
		case char_PageDown:		button = kElement_DownButton;		break;

		case char_Function:
			switch (iKeyEvent.message & keyCodeMask)
			{
				case vkey_F1 << 8:	button = kElement_App1Button;	break;
				case vkey_F2 << 8:	button = kElement_App2Button;	break;
				case vkey_F3 << 8:	button = kElement_App3Button;	break;
				case vkey_F4 << 8:	button = kElement_App4Button;	break;
			}
			break;

		default:				key = PrvRemapMacCharacter(chr);	break;
	}

	if (key != 0)
	{
		EmKeyEvent	event (key);
		::PrvGetModifiers (event, iKeyEvent.modifiers);
		this->HandleKey (event);
	}

	if (button != kElement_None)
	{
		this->HandleButton (button, true);
		this->HandleButton (button, false);
	}

	return true;
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::AttemptClose
// ---------------------------------------------------------------------------
//	Try to close a Document.
//
//	The Document might not close if it is modified and the user cancels
//	the operation when asked whether to save the changes.

void EmDocumentMac::AttemptClose (Boolean inRecordIt)
{
	UNUSED_PARAM (inRecordIt);

	if ((mSuperCommander == NULL) || mSuperCommander->AllowSubRemoval (this))
	{

		FSSpec	fileSpec = this->GetFileRef ().GetFSSpec ();
		Bool	closed = this->HandleClose (false);

#if 0
		if (closed)
		{
			if (inRecordIt)
			{
				try
				{
					this->SendAEClose (kAEYes, fileSpec, ExecuteAE_No);
				}
				catch (...)
				{
				}
			}
		}
#endif
	}
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::AttemptQuitSelf
// ---------------------------------------------------------------------------
//	Try to close a Document when quitting the program.

Boolean EmDocumentMac::AttemptQuitSelf (SInt32 inSaveOption)
{
	UNUSED_PARAM (inSaveOption);

	Bool	closed = this->HandleClose (true);

	return closed;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmDocumentMac::HandleIdle
// ---------------------------------------------------------------------------

void EmDocumentMac::HandleIdle (void)
{
	// If the PPCSocket system recorded an exception to display and
	// handle, display and handle it.

	if (fHavePendingException)
	{
		fHavePendingException = false;

		fPendingException.Display ();
		fPendingException.DoAction ();
	}

	// Adjust our sleep time depending on the state of the processor.

	if (!gCPU || gCPU->Stopped ())
	{
		gHostApplication->SetSleepTime (2);
	}
	else
	{
		gHostApplication->SetSleepTime (0);
	}

	// Keep the screen from being dimmed and auto-sleep from occurring if the
	// charger is connected.

	if ((gHostApplication->GetSleepTime () == 0) && PrvIsChargerConnected ())
	{
		UpdateSystemActivity (UsrActivity);
	}

	EmDocument::HandleIdle ();

	// Idle the session.

#if __profile__
	short	oldStatus = ProfilerGetStatus ();
	ProfilerSetStatus (true);
#endif

	EmAssert (fSession);
	fSession->ExecuteIncremental ();

#if __profile__
	ProfilerSetStatus (oldStatus);
#endif
}


// ---------------------------------------------------------------------------
//		 EmDocumentMac::HostSaveScreen
// ---------------------------------------------------------------------------

void EmDocumentMac::HostSaveScreen (const EmFileRef& destRef)
{
	PicHandle	picHandle = NULL;

	// !!! This method should not rely on a window being open.

	if (gWindow)
	try
	{
		// Get information about the screen.

		EmScreenUpdateInfo	info;
		gWindow->GetLCDContents (info);

		// Fill in a GWorld for blitting from.  This GWorld will
		// also serve as the destination for the purpose of
		// recording the bitmap in a PICT.

		GWorldPtr	gWorld;
		EmPoint		imageSize (info.fImage.GetSize ());
		::ConvertPixMapToHost (info.fImage, gWorld, 0, imageSize.fY, false);

		{
			StGDeviceSaver	saved;
			::SetGWorld (gWorld, NULL);

			// Create the PicHandle that will contain the picture.

			OpenCPicParams	picHeader =
			{
				0, 0, imageSize.fY, imageSize.fX,	// srcRect
				0x00480000, 0x00480000,				// resolution
				-2, 0, 0							// version, reserved
			};

			picHandle = ::OpenCPicture (&picHeader);
			Errors::ThrowIfNULL (picHandle);

			// Draw the GWorld into itself, just to get the bits recorded
			// into the picture.

			::RGBBackColor (&Color_White);
			::RGBForeColor (&Color_Black);

			PixMapHandle	pixMap		= ::GetGWorldPixMap (gWorld);
			Rect			destRect	= (**pixMap).bounds;

			::LockPixels (pixMap);

			::CopyBits ((BitMap*) *pixMap, (BitMap*) *pixMap,
						&destRect, &destRect, srcCopy, NULL);

			// We're done recording.

			::ClosePicture ();
		}

		// Clean up (dispose of) the GWorld.

		::DisposeGWorld (gWorld);

		// Create the file we'll be writing to.  Create it as
		// a TeachText/SimpleText picture file.  If the file already
		// exists, write over it.

		EmStreamFile	stream (destRef, kCreateOrEraseForUpdate,
								kFileCreatorTeachText, kFileTypePicture);

		// Set the file to be at least 512 bytes long so that we
		// can set the write mark to that location.

		stream.SetLength (512);

		// Position ourselves to be 512 bytes into the file.  PICT
		// files start with a 512 byte header (which can be anything;
		// but we don't use it here).

		stream.SetMarker (512, kStreamFromStart);

		// Write out the PICT.

		long	picSize = ::GetHandleSize ((Handle) picHandle);
		stream.PutBytes (*picHandle, picSize);
	}
	catch (...)
	{
		if (picHandle)
			::KillPicture (picHandle);

		throw;
	}

	if (picHandle)
		::KillPicture (picHandle);
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmDocumentMac::SetPendingException
// ---------------------------------------------------------------------------

void EmDocumentMac::SetPendingException (const EmExceptionReset& e)
{
	fHavePendingException = true;
	fPendingException = e;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 PrvGetModifiers
// ---------------------------------------------------------------------------

void	PrvGetModifiers (EmKeyEvent& info, EventModifiers mods)
{
//	if ((mods & cmdKey) != 0)
//		info.fCommandDown = true;

	if ((mods & (shiftKey | rightShiftKey)) != 0)
		info.fShiftDown = true;

	if ((mods & alphaLock) != 0)
		info.fCapsLockDown = true;

	if ((mods & (optionKey | rightOptionKey)) != 0)
		info.fOptionDown = true;

	if ((mods & (controlKey | rightControlKey)) != 0)
		info.fControlDown = true;
}


/***********************************************************************
 *
 * FUNCTION:	PrvRemapMacCharacter
 *
 * DESCRIPTION:	Remap the character <macChar> from the script of the Mac
 *	keyboard to the device's character encoding, and return the result.
 *
 * PARAMETERS:
 *	macChar	 ->	Mac character code.
 *
 * RETURNED:	Device character, or 0 if it wasn't remappable.
 *
 ***********************************************************************/

uint16 PrvRemapMacCharacter (uint8 macChar)
{
	// We'll stash high-bytes here.
	static uint16 bufferedChar = 0;
	
	uint16 result = macChar;
	uint16 keyScript = ::GetScriptManagerVariable (smKeyScript);
	
	if (bufferedChar != 0)
	{
		// We have a buffered high-byte, so build the complete character.
		result = bufferedChar | macChar;
		bufferedChar = 0;
	}
	else if (::CharacterByteType ((char*) &macChar, 0, keyScript) == smFirstByte)
	{
		bufferedChar = macChar << 8;
		return(0);
	}

	uint32	deviceEncoding;
	if (::FtrGet (sysFtrCreator, sysFtrNumEncoding, &deviceEncoding) != errNone)
	{
		deviceEncoding = charEncodingPalmLatin;
	}

	return Platform::RemapHostChar (keyScript, deviceEncoding, result);
}


#pragma mark -

//=====================================================================================================
//
// Power Manager Utilities
//
//=====================================================================================================	

static long gPowerMgrAttr = 0;

Boolean PrvHasPowerManager (void)
{
	Boolean hasPowerManager = false;
	
	// If we haven't already ascertained the Power Manager attributes,
	// then do so now.

	if (gPowerMgrAttr == 0)
	{
		if (::Gestalt (gestaltPowerMgrAttr, &gPowerMgrAttr) != noErr)
		{
			gPowerMgrAttr = 0;
		}
	}
		
	// If the Power Manager attributes are non-zero and the
	// Power Manager dispatch-table bit is set, then we have
	// a Power Manager.

	if (gPowerMgrAttr & (1 << gestaltPMgrDispatchExists))
	{
		hasPowerManager = true;
	}
		
	return hasPowerManager;
}

Boolean PrvHasPowerBookPowerManager (void)
{
	Boolean hasPowerBookPowerManager = false;
	
	// If the Power Manager exists and the PowerBook Power Manager bit is set,
	// then we have a PowerBook Power Power Manager.
	//
	if (::PrvHasPowerManager ())
	{
		hasPowerBookPowerManager = (gPowerMgrAttr & (1 << gestaltPMgrExists)) != 0;
	}
		
	return hasPowerBookPowerManager;
}

Boolean PrvIsChargerConnected (void)
{
	// If we've got a Power Manager, then just say the "charger" is connected.
	// This takes care of the Power Manager on desktop CPUs where, in essence,
	// the charger is always connected (since there are never any batteries).
	//
	Boolean isChargerConnected = PrvHasPowerManager ();
	
	// If we've got a PowerBook Power Manager, then we really need to check
	// to see if there's a charger (aka wall adapter) connected or not by
	// asking the Power Manger itself.
	//
	if (::PrvHasPowerBookPowerManager ())
	{
		BatteryInfo batteryInfo = { 0 };
	
		// There's always at least one battery and we just need to know if the charger
		// is connected to it or not.  So, we simply ask the Power Manager about the
		// "first" battery.
		//
		::GetScaledBatteryInfo (1, &batteryInfo);
		isChargerConnected = (batteryInfo.flags & (1 << chargerConnected)) != 0;
	}

	return isChargerConnected;
}
