/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 2000-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmFileRefMac.h"

#include "EmDirRefMac.h"		// PrvUnspecify, et al.
#include "ErrorHandling.h"		// Errors::ThrowIfError

#include <errno.h>				// ENOENT, errno
#include <sys/types.h>			
#include <sys/stat.h>			// chmod, stat

static const OSType	kCreators[] =
{
	0,			// kFileCreatorNone,
	'LOT',		// kFileCreatorEmulator,
	'PPia',		// kFileCreatorInstaller,
	'ttxt',		// kFileCreatorTeachText,
	'CWIE',		// kFileCreatorCodeWarrior,
	'proF'		// kFileCreatorCodeWarriorProfiler
};


static const OSType	kTypes[] =
{
	0,			// kFileTypeNone,
	'APPL',		// kFileTypeApplication,
	'ROM',		// kFileTypeROM,
	'PSF',		// kFileTypeSession,
	'PEV',		// kFileTypeEvents,
	'TEXT',		// kFileTypePreference,
	'PRC ',		// kFileTypePalmApp,
	'PRC ',		// kFileTypePalmDB,
	'PRC ',		// kFileTypePalmQA,
	'TEXT',		// kFileTypeText,
	'PICT',		// kFileTypePicture,
	0,			// kFileTypeSkin,
	'proF',		// kFileTypeProfile,
	0,			// kFileTypePalmAll,
	0			// kFileTypeAll
};


static const char*	kExtension[] =
{
	NULL,		// kFileTypeNone,
	NULL,		// kFileTypeApplication,
	".rom",		// kFileTypeROM,
	".psf",		// kFileTypeSession,
	".pev",		// kFileTypeEvents,
	NULL,		// kFileTypePreference,
	".prc",		// kFileTypePalmApp,
	".pdb",		// kFileTypePalmDB,
	".pqa",		// kFileTypePalmQA,
	".txt",		// kFileTypeText,
	NULL,		// kFileTypePicture,
	".skin",	// kFileTypeSkin,
	".prof",	// kFileTypeProfile,
	NULL,		// kFileTypePalmAll,
	NULL		// kFileTypeAll
};


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::EmFileRef
 *
 * DESCRIPTION:	Various ways to make a file reference.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

EmFileRef::EmFileRef (void) :
	fFileAlias (NULL),
	fFileSpec ()
{
	::PrvUnspecify (&fFileSpec);

	COMPILE_TIME_ASSERT (countof (kCreators) == kFileCreatorLast);
	COMPILE_TIME_ASSERT (countof (kTypes) == kFileTypeLast);
	COMPILE_TIME_ASSERT (countof (kExtension) == kFileTypeLast);
}


EmFileRef::EmFileRef (const EmFileRef& other) :
	fFileAlias (NULL),
	fFileSpec (other.fFileSpec)
{
	Handle	newHandle = (Handle) other.fFileAlias;

	if (newHandle)
	{
		OSErr	err = ::HandToHand (&newHandle);
		Errors::ThrowIfError (err);
	}

	fFileAlias = (AliasHandle) newHandle;
}


EmFileRef::EmFileRef (const char* s) :
	fFileAlias (NULL)
{
	LStr255 name (s);
	::PrvMakeSpec (name, &fFileSpec);
}


EmFileRef::EmFileRef (const string& s) :
	fFileAlias (NULL)
{
	LStr255 name (s.c_str ());
	::PrvMakeSpec (name, &fFileSpec);
}


EmFileRef::EmFileRef (const EmDirRef& parent, const char* s) :
	fFileAlias (NULL)
{
	LStr255 name (s);
	::PrvMakeSpec (parent, name, &fFileSpec);
}


EmFileRef::EmFileRef (const EmDirRef& parent, const string& s) :
	fFileAlias (NULL)
{
	LStr255 name (s.c_str ());
	::PrvMakeSpec (parent, name, &fFileSpec);
}


EmFileRef::EmFileRef (const unsigned char* s) :
	fFileAlias (NULL)
{
	LStr255 name (s);
	::PrvMakeSpec (name, &fFileSpec);
}


EmFileRef::EmFileRef (const FSSpec& spec) :
	fFileAlias (NULL),
	fFileSpec (spec)
{
	// Get an alias now, before the spec can be invalidated.

	this->UpdateAlias ();
}


EmFileRef::EmFileRef (AliasHandle alias) :
	fFileAlias (alias),
	fFileSpec ()
{
	::PrvUnspecify (&fFileSpec);
	this->UpdateSpec ();
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::~EmFileRef
 *
 * DESCRIPTION: EmFileRef destructor.  Nothing special to do...
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

EmFileRef::~EmFileRef (void)
{
	::DisposeHandle ((Handle) fFileAlias);
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::operator=
 *
 * DESCRIPTION: Assignment operator.  If "other" is not the same as
 *				the controlled object, copy the contents.
 *
 * PARAMETERS:	other - object to copy.
 *
 * RETURNED:	reference to self.
 *
 ***********************************************************************/

EmFileRef&
EmFileRef::operator= (const EmFileRef& other)
{
	if (&other != this)
	{
		Handle	newHandle = (Handle) other.fFileAlias;

		if (newHandle)
		{
			OSErr	err = ::HandToHand (&newHandle);
			Errors::ThrowIfError (err);
		}

		::DisposeHandle ((Handle) fFileAlias);
		fFileAlias = (AliasHandle) newHandle;
		fFileSpec = other.fFileSpec;
	}

	return *this;
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::IsSpecified
 *
 * DESCRIPTION: Returns whether or not the controlled object has been
 *				pointed to a (possibly non-existant) file, or if it's
 *				empty (that it, it was created with the default ctor).
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	True if the object points to a file.
 *
 ***********************************************************************/

Bool
EmFileRef::IsSpecified (void) const
{
	return (fFileAlias != NULL) || ::PrvIsSpecified (fFileSpec);
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::Exists
 *
 * DESCRIPTION: Returns whether or not the controlled object points to
 *				an existing file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	True if the referenced file exists.
 *
 ***********************************************************************/

Bool
EmFileRef::Exists (void) const
{
	if (this->IsSpecified ())
	{
		FSSpec	spec = this->GetFSSpec ();

		return ::PrvSpecExists (spec);
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:    EmFileRef::IsType
 *
 * DESCRIPTION: DESCRIPTION
 *
 * PARAMETERS:  None
 *
 * RETURNED:    Nothing
 *
 ***********************************************************************/

Bool
EmFileRef::IsType (EmFileType type) const
{
	FSSpec	mySpec = this->GetFSSpec ();

	FInfo	fndrInfo;
	OSErr	err = ::FSpGetFInfo (&mySpec, &fndrInfo);
	if (err != noErr)
		return false;

	if (kTypes[type] != 0)
	{
		if (fndrInfo.fdType == kTypes[type])
			return true;
	}

	// Some PRC files have incorrect file types.
	// Special-case those by looking at their file extensions

	if (kExtension[type] != NULL)
	{
		LStr255 fileName (mySpec.name);
		fileName.SetCompareFunc (LString::CompareIgnoringCase);

		if (fileName.EndsWith (LStr255(kExtension[type])))
			return true;

		// Add special hacks for ROM files.
		if (type == kFileTypeROM)
		{
			if (fileName.EndsWith (LStr255(" rom")))
				return true;

			if (fileName.BeginsWith (LStr255("rom.")))
				return true;

			// Special case some ROM files we have around here that
			// don't follow the above rules.

			if (fndrInfo.fdCreator == 'mdos' ||
				fndrInfo.fdCreator == '??\??')
			{
				fileName.SetCompareFunc (LString::CompareBytes);

				if (fileName.BeginsWith (LStr255("ROM5")))
					return true;

				if (fileName.BeginsWith (LStr255("ROMf")))
					return true;
			}
		}
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::SetCreatorAndType
 *
 * DESCRIPTION: Set the Finder type and creator information of the
 *				managed file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void
EmFileRef::SetCreatorAndType (EmFileCreator creator, EmFileType fileType) const
{
	FInfo	fInfo;
	FSSpec	mySpec = this->GetFSSpec ();
	OSErr	err = ::FSpGetFInfo (&mySpec, &fInfo);

	if (err == noErr)
	{
		if (::PrvTypeToOSType (fileType))
			fInfo.fdType	= ::PrvTypeToOSType (fileType);

		if (::PrvCreatorToOSType (creator))
			fInfo.fdCreator	= ::PrvCreatorToOSType (creator);

		::FSpSetFInfo (&mySpec, &fInfo);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::GetAttr
 *
 * DESCRIPTION: Get basic file attributes of the managed file.
 *
 * PARAMETERS:	A pointer to an integer where the mode bits will be stored.
 *
 * RETURNED:	An integer containing an errno style error result, 0 for no error.
 *
 ***********************************************************************/

int
EmFileRef::GetAttr (int * mode) const
{
	EmAssert (mode);
	
	*mode = 0;
	
	if (!IsSpecified())
		return ENOENT;
	
	GetFSSpec();
	
	FInfo	theFInfo;

	if (FSpGetFInfo(&fFileSpec, &theFInfo) != noErr)
		return ENOENT;
	
	if (theFInfo.fdFlags & kIsInvisible)
		*mode |= kFileAttrHidden;
	
	struct stat stat_buf;
	if (stat(GetFullPath().c_str(), &stat_buf))
		return errno;
	
	if ((stat_buf.st_mode & S_IWUSR) == 0)
		*mode |= kFileAttrReadOnly;
	
	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::SetAttr
 *
 * DESCRIPTION: Set basic file attributes of the managed file.
 *
 * PARAMETERS:	An integer containing bits from the EmFileAttr enum.
 *
 * RETURNED:	An integer containing an errno style error result, 0 for no error.
 *
 ***********************************************************************/

int
EmFileRef::SetAttr (int mode) const
{
	if (!IsSpecified())
		return ENOENT;
		
	GetFSSpec();

	FInfo	theFInfo;
	
	if (FSpGetFInfo(&fFileSpec, &theFInfo) != noErr)
		return ENOENT;
	
	theFInfo.fdFlags &= ~kIsInvisible;
	theFInfo.fdFlags |= (mode & kFileAttrHidden) ? kIsInvisible : 0;

	if (FSpSetFInfo(&fFileSpec, &theFInfo) != noErr)
		return ENOENT;
		
	if (chmod(GetFullPath().c_str(), (mode & kFileAttrReadOnly) ? 0 : S_IWUSR))
		return errno;
	
	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::GetName
 *
 * DESCRIPTION: Returns the name of the referenced file.  Only the file
 *				*name* is returned, not the full path.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	A string containing the name.  If the file is not
 *				specified, an empty string is returned.  No checks are
 *				made to see if the file actually exists.
 *
 ***********************************************************************/

string
EmFileRef::GetName (void) const
{
	FSSpec	mySpec	= this->GetFSSpec ();
	return string ((char*) &mySpec.name[1], mySpec.name[0]);
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::GetParent
 *
 * DESCRIPTION:	Returns an object representing the parent (or container)
 *				of the managed file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	An object representing the file's parent.
 *
 ***********************************************************************/

EmDirRef
EmFileRef::GetParent (void) const
{
	EmDirRef	result;

	if (this->IsSpecified ())
	{
		FSSpec	mySpec	= this->GetFSSpec ();
		FSSpec	parSpec	= ::PrvGetParentSpec (mySpec);

		if (parSpec.vRefNum != 0)
		{
			result = EmDirRef (parSpec);
		}
	}

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::GetFullPath
 *
 * DESCRIPTION:	Get a full (platform-specific) path to the object.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	An string representing the file's path.
 *
 ***********************************************************************/

string
EmFileRef::GetFullPath (void) const
{
	FSSpec	mySpec	= this->GetFSSpec ();
	return ::PrvGetFullPath (mySpec);
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::operator==
 * FUNCTION:	EmFileRef::operator!=
 * FUNCTION:	EmFileRef::operator>
 * FUNCTION:	EmFileRef::operator<
 *
 * DESCRIPTION: Bogus operators for wiggy VC++ compiler which won't let
 *				us instantiate STL containers without them.
 *
 * PARAMETERS:	other - object to compare ourself to.
 *
 * RETURNED:	True if the requested condition is true.  Comparisons
 *				are based on the file's full path.
 *
 ***********************************************************************/

bool
EmFileRef::operator== (const EmFileRef& other) const
{
	FSSpec	mySpec = this->GetFSSpec ();
	FSSpec	hisSpec = other.GetFSSpec ();

	return ::PrvEqualFSSpec (mySpec, hisSpec);
}


bool
EmFileRef::operator!= (const EmFileRef& other) const
{
	return !(*this == other);
}


/***********************************************************************
 *
 * FUNCTION:	
 *
 * DESCRIPTION: 
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

bool
EmFileRef::FromPrefString (const string& s)
{
	if (fFileAlias)
	{
		::DisposeHandle ((Handle) fFileAlias);
		fFileAlias = NULL;
	}

	::PrvUnspecify (&fFileSpec);

	if (s.size() == 0)
	{
		return true;
	}

	short	vol;
	long	dir;
	::PrvGetAppVolDir (&vol, &dir);

	LStr255 name (s.c_str ());
	OSErr	err = ::FSMakeFSSpec (vol, dir, name, &fFileSpec);

	return (err == noErr) || (err == fnfErr) || (err == dirNFErr) || (err == nsvErr);
}


/***********************************************************************
 *
 * FUNCTION:	
 *
 * DESCRIPTION: 
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

string
EmFileRef::ToPrefString (void) const
{
	return this->GetFullPath ();
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::GetAlias
 *
 * DESCRIPTION: Returns the alias of the referenced file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	Alias for the referenced file.	If the EmFileRef
 *				was created with a file spec, we attempt to create an
 *				alias and return it.  If an alias cannot be created,
 *				this function could return NULL.
 *
 ***********************************************************************/

AliasHandle
EmFileRef::GetAlias (void) const
{
	if (!fFileAlias && ::PrvIsSpecified (fFileSpec))
	{
		const_cast<EmFileRef*>(this)->UpdateAlias ();
	}

	return fFileAlias;
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::GetFSSpec
 *
 * DESCRIPTION: Returns the filespec of the referenced file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	Filespec for the referenced file.  If the EmFileRef
 *				was created with an alias, we attempt to resolve it and
 *				return the resulting filespec.	If an alias cannot be
 *				resolved, this function could return an empty spec.
 *
 ***********************************************************************/

FSSpec
EmFileRef::GetFSSpec (void) const
{
	if (!::PrvIsSpecified (fFileSpec) && fFileAlias)
	{
		const_cast<EmFileRef*>(this)->UpdateSpec ();
	}

	return fFileSpec;
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::UpdateAlias
 *
 * DESCRIPTION: Creates an alias from the filespec for the controlled
 *				file.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void
EmFileRef::UpdateAlias (void)
{
	OSErr	err = ::NewAlias (NULL, &fFileSpec, &fFileAlias);

// I think there are too many "soft errors" (fnfErr, userCanceledErr)
// that can occur here that we don't want to throw an exception.
//
//	Errors::ThrowIfError (err);
}


/***********************************************************************
 *
 * FUNCTION:	EmFileRef::UpdateSpec
 *
 * DESCRIPTION: Resolves an alias and saves the resulting filespec.
 *
 * PARAMETERS:	none.
 *
 * RETURNED:	nothing.
 *
 ***********************************************************************/

void
EmFileRef::UpdateSpec (void)
{
	Boolean wasChanged;
	short	matchCount = 1;

	// Use MatchAlias instead of ResolveAlias to avoid any UI.

//	OSErr	err = ::ResolveAlias (NULL, fFileAlias, &fFileSpec, &wasChanged);
	OSErr	err = ::MatchAlias (NULL, kARMSearch | kARMNoUI, fFileAlias,
									&matchCount, &fFileSpec, &wasChanged, NULL, NULL);

// I think there are too many "soft errors" (fnfErr, userCanceledErr)
// that can occur here that we don't want to throw an exception.
//
//	Errors::ThrowIfError (err);
}

OSType	PrvCreatorToOSType (EmFileCreator creator)
{
	return kCreators[creator];
}

OSType	PrvTypeToOSType (EmFileType type)
{
	return kTypes[type];
}
