/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmMenusMac.h"


const SInt16	InstallMenuItem_AtEnd = 999;


// ---------------------------------------------------------------------------
//		 PrvGetTitle
// ---------------------------------------------------------------------------
// Convert the cross-platform menu title into something we can use on the
// Mac.  It's converted into a Pascal string, and the ampersands are removed.

static LStr255 PrvGetTitle (const string& title)
{
	// Create a Pascal version of the title.

	LStr255	result (title.c_str ());

	// Look for any '&', indicating a mnemonic character.  We
	// don't use those on the Mac, so they must be removed before
	// passing the title on to the Mac Menu Manager.

	UInt8	ampPos = result.Find ((unsigned char) '&');
	if (ampPos > 0)
		result.Remove (ampPos, 1);

	return result;
}


// ---------------------------------------------------------------------------
//		 PrvGetStandardMenuID
// ---------------------------------------------------------------------------
// Convert the cross-platform menu ID into something that can be used on the
// Mac.  Convert it into a value in the range 1..16.  This keeps it in the
// general range of 1..234 allowed for hierarchical and popup menus, while
// leaving a separate number space for hierarchical and popup menus (in case
// we install the same cross-platform menu more than once: as, for example,
// a standard menu and a popup menu.  Note that this range also avoids the
// IDs 71..74, which are usurped by "Instant Palm Desktop".

static ResIDT PrvGetStandardMenuID (EmCommandID commandID)
{
	ResIDT	result = (commandID % 16) + 1;
	return result;
}


// ---------------------------------------------------------------------------
//		 PrvGetHierarchicalMenuID
// ---------------------------------------------------------------------------
// Convert the cross-platform menu ID into something that can be used on the
// Mac.  Convert it into a value in the range 17..32.  This keeps it in the
// general range of 1..234 allowed for hierarchical and popup menus, while
// leaving a separate number space for standard and popup menus (in case
// we install the same cross-platform menu more than once: as, for example,
// a standard menu and a popup menu.  Note that this range also avoids the
// IDs 71..74, which are usurped by "Instant Palm Desktop".

static ResIDT PrvGetHierarchicalMenuID (EmCommandID commandID)
{
	ResIDT	result = (commandID % 16) + 17;
	return result;
}


// ---------------------------------------------------------------------------
//		 PrvGetPopupMenuID
// ---------------------------------------------------------------------------
// Convert the cross-platform menu ID into something that can be used on the
// Mac.  Convert it into a value in the range 33..48.  This keeps it in the
// general range of 1..234 allowed for hierarchical and popup menus, while
// leaving a separate number space for standard and hierarchical menus (in case
// we install the same cross-platform menu more than once: as, for example,
// a standard menu and a popup menu.  Note that this range also avoids the
// IDs 71..74, which are usurped by "Instant Palm Desktop".

static ResIDT PrvGetPopupMenuID (EmCommandID commandID)
{
	ResIDT	result = (commandID % 16) + 33;
	return result;
}


// ---------------------------------------------------------------------------
//		 PrvFindMenuItemByMenuChoice
// ---------------------------------------------------------------------------

static const EmMenuItem*
PrvFindMenuItemByMenuChoice (	const EmMenuItemList& menu,
								EmCommandID parentID,
								long menuChoice,
								Bool recurse)
{
	const EmMenuItem*	result = NULL;

	EmMenuItemList::const_iterator	iter = menu.begin ();
	while (iter != menu.end ())
	{
		if (::PrvGetPopupMenuID (parentID) == HiWord (menuChoice) &&
			iter - menu.begin () + 1 == LoWord (menuChoice))
		{
			return &(*iter);
		}

		if (recurse)
		{
			const EmMenuItemList& children = iter->GetChildren ();

			if (children.size () > 0)
			{
				result = ::PrvFindMenuItemByMenuChoice (children,
							iter->GetCommand (), menuChoice, recurse);

				if (result)
					return result;
			}
		}

		++iter;
	}

	return result;
}


// ---------------------------------------------------------------------------
//		 PrvAddMenuItems
// ---------------------------------------------------------------------------
// Add the given menu items to the given menu.  Normal menu items, divider
// lines, and hierarchical menu items are all handled (the latter by calling
// this function recursively).

static void PrvAddMenuItems (LMenuBar* menubar, LMenu* menu,
								const EmMenuItemList& menuItems,
								Bool isHierarchical)
{
	// Clear out the old items.

	SInt16	numItems = ::CountMenuItems (menu->GetMacMenuH ());
	while (numItems--)
	{
		menu->RemoveItem (1);
	}

	// Add the desired items.

	EmMenuItemList::const_iterator	iter = menuItems.begin ();
	while (iter != menuItems.end ())
	{
		if (iter->GetIsDivider ())
		{
			LStr255		title	("\p(-");
			CommandT	command	= kCommandNone;
			menu->InsertCommand (title, command, InstallMenuItem_AtEnd);
		}
		else
		{
			// Insert a dummy menu item.  The Mac Menu Manager parses the
			// text of the title passed to the InsertCommand method for
			// meta-formatting characters.  We don't want that parsing, so
			// we initially set a dummy title, and the replace the title by
			// calling SetMenuItemText (which does not scan for meta-characters).

			CommandT	command	= iter->GetCommand ();
			menu->InsertCommand (Str_Dummy, command, InstallMenuItem_AtEnd);

			// Get the Mac Menu Manager menu handle and the index to the
			// item we just  added so that we can make raw Menu Mgr calls.

			SInt16		index	= menu->IndexFromCommand (command);
			MenuRef		macMenu = menu->GetMacMenuH ();

			// Set the menu text to what we want.

			LStr255		title	= ::PrvGetTitle (iter->GetTitle ());
			::SetMenuItemText (macMenu, index, title);

			// Set the Cmd-key to what we want (if anything).
			// Only do this for hierarchical menus; popup menus
			// don't need them.

			if (isHierarchical && iter->GetShortcut ())
			{
				::SetItemCmd (macMenu, index, iter->GetShortcut ());
			}

			// Enable or disable the item.

			if (iter->GetIsActive ())
			{
				menu->EnableItem (index);
			}
			else
			{
				menu->DisableItem (index);
			}

			// Check or uncheck the item.

			if (iter->GetIsChecked ())
			{
				::SetItemMark (macMenu, index, checkMark);
			}
			else
			{
				::SetItemMark (macMenu, index, noMark);
			}

			// If there are any child menu items, make this a hierarchical menu.

			const EmMenuItemList&	subMenu = iter->GetChildren ();
			if (subMenu.size () > 0)
			{
				// Indicate that this menu is a hierarchical menu.  It's flagged
				// as a hierarchical meenu by setting its Cmd-character to
				// hMenuCmd, and setting its Mark-character to the ID of the
				// sub-menu.

				ResIDT	menuID	= isHierarchical
					? ::PrvGetHierarchicalMenuID (iter->GetCommand ())
					: ::PrvGetPopupMenuID (iter->GetCommand ());
				::SetItemCmd (macMenu, index, hMenuCmd);
				::SetItemMark (macMenu, index, menuID);

				// Now create the sub-menu, giving it the right ID, inserting
				// it into the menubar in the space reserved for hierarchicals
				// and popups, and populating it.

				LMenu*		menu	= menubar->FetchMenu (menuID);

				if (menu == NULL)
				{
					menu = new LMenu (menuID, Str_Empty);
					menubar->InstallMenu (menu, hierMenu);
				}

				::PrvAddMenuItems (menubar, menu, subMenu, isHierarchical);
			}
		}

		++iter;
	}
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 HostCreatePopupMenu
// ---------------------------------------------------------------------------

LMenu* HostCreatePopupMenu (const EmMenuItemList& menuList)
{
	// Create or find the top-level popup menu.

	LMenuBar*	menubar = LMenuBar::GetCurrentMenuBar ();
	ResIDT		menuID	= ::PrvGetPopupMenuID (kCommandRoot);
	LMenu*		menu	= menubar->FetchMenu (menuID);

	if (menu == NULL)
	{
		menu = new LMenu (menuID, Str_Empty);
		menubar->InstallMenu (menu, hierMenu);
	}

	// Add menu items.

	::PrvAddMenuItems (menubar, menu, menuList, false);

	return menu;
}


// ---------------------------------------------------------------------------
//		 HostCreateMenubar
// ---------------------------------------------------------------------------
// Create or update the Mac menubar.  Attempts to update menus if they're
// already there (handy for updating the MRU menus).  But will create the
// menus if they don't already exist.

void HostCreateMenubar (const EmMenuItemList& menu)
{
	LMenuBar*	menubar = LMenuBar::GetCurrentMenuBar ();

	// Add top-level menus

	EmMenuItemList::const_iterator	iter = menu.begin ();
	while (iter != menu.end ())
	{
		ResIDT		menuID	= ::PrvGetStandardMenuID (iter->GetCommand ());
		LStr255		title	= ::PrvGetTitle (iter->GetTitle ());
		LMenu*		menu	= menubar->FetchMenu (menuID);

		if (menu == NULL)
		{
			menu = new LMenu (menuID, title);
			menubar->InstallMenu (menu, InstallMenu_AtEnd);
		}

		// Add menu items.

		::PrvAddMenuItems (menubar, menu, iter->GetChildren (), true);

		// Move to next top-level menu.

		++iter;
	}
}


// ---------------------------------------------------------------------------
//		 HostFindMenuItemByMenuChoice
// ---------------------------------------------------------------------------

const EmMenuItem*	HostFindMenuItemByMenuChoice (	const EmMenuItemList& menu,
											long menuChoice,
											Bool recurse)
{
	const EmMenuItem*	result = ::PrvFindMenuItemByMenuChoice (menu, kCommandRoot,
													menuChoice, recurse);

	return result;
}
