/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 2000-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmPixMapMac.h"

#include "ErrorHandling.h"	// Errors::ThrowIfNULL
#include "Platform.h"		// Platform::AllocateMemory

#define RGB8To16(c)	((((uint16) c) << 8) + ((uint16) c))

static EmPixMapFormat PrvGetFormat	(const PixMap& pixMap);

// ---------------------------------------------------------------------------
//		 ConvertPixMapToHost
// ---------------------------------------------------------------------------

void ConvertPixMapToHost (	const EmPixMap& src, GWorldPtr& dest,
							int firstLine, int lastLine, Bool scale)
{
	// Determine a lot of the values we'll need.

	EmPoint				srcSize			= src.GetSize ();
	EmPixMapDepth		srcDepth		= src.GetDepth ();
	EmPixMapRowBytes	srcRowBytes		= src.GetRowBytes ();
	EmCoord				srcWidth		= srcSize.fX;
	EmCoord				srcHeight		= srcSize.fY;

	EmCoord				destWidth		= srcWidth * (scale ? 2 : 1);
	EmCoord				destHeight		= srcHeight * (scale ? 2 : 1);
	long				destDepth		= srcDepth <= 8 ? 8 : 32;
	Rect				destRect		= { 0, 0, destHeight, destWidth };

	// Set up the pixmap's color table.

	CTabHandle	cTable	= (CTabHandle) ::NewHandle (0);
	::ConvertFromRGBList (cTable, src.GetColorTable ());

	// Create the offscreen GWorld.

	OSErr		err		= ::NewGWorld (&dest, destDepth, &destRect,
								cTable, NULL, useTempMem | keepLocal);

	::DisposeCTable (cTable);

	if (err)
	{
		dest = NULL;
		return;
	}

	// Copy the pixels from the EmPixMap into the GWorld.

	PixMapHandle	pixMap = ::GetGWorldPixMap (dest);

	::LockPixels (pixMap);

	EmPixMap	wrapper;

	wrapper.SetSize (EmPoint (destWidth, destHeight));
	wrapper.SetFormat (destDepth == 8 ? kPixMapFormat8 : kPixMapFormat32ARGB);
	wrapper.SetRowBytes ((**pixMap).rowBytes & 0x3FFF);
	wrapper.SetColorTable (src.GetColorTable ());
	wrapper.SetBits ((**pixMap).baseAddr);

	EmRect	srcBounds (0, firstLine, srcSize.fX, lastLine);
	EmRect	destBounds (srcBounds);

	if (scale)
	{
		destBounds *= EmPoint (2, 2);
	}

	EmPixMap::CopyRect (wrapper, src, destBounds, srcBounds);

	::UnlockPixels (pixMap);
}


// ---------------------------------------------------------------------------
//		 CreatePixMapFromHost
// ---------------------------------------------------------------------------

void CreatePixMapFromHost (GWorldPtr src, EmPixMap& dest)
{
	PixMapHandle	pixMap = ::GetGWorldPixMap (src);

	if (::LockPixels (pixMap))
	{
		StHandleLocker	lock ((Handle) pixMap);
		::CreatePixMapFromHost (**pixMap, dest);
		::UnlockPixels (pixMap);
	}
}


void CreatePixMapFromHost (const PixMap& src, EmPixMap& dest)
{
	// Wrap up the PixMap in an EmPixMap.

	EmPixMap	wrapper;

	long	height = src.bounds.bottom - src.bounds.top;
	long	width = src.bounds.right - src.bounds.left;

	wrapper.SetSize (EmPoint (width, height));
	wrapper.SetFormat (::PrvGetFormat (src));
	wrapper.SetRowBytes (src.rowBytes & 0x3FFF);

	int	numColors = (**src.pmTable).ctSize + 1;
	if (numColors)
	{
		RGBList	colors;
		::ConvertToRGBList (colors, src.pmTable);
		wrapper.SetColorTable (colors);
	}

	wrapper.SetBits (src.baseAddr);

	// Copy the data to the destination.

	dest = wrapper;
}


// ---------------------------------------------------------------------------
//		 ConvertRGBTypeToHost
// ---------------------------------------------------------------------------

RGBColor	ConvertRGBTypeToHost	(const RGBType& inRGB)
{
	RGBColor	outRGB;

	outRGB.red		= (inRGB.fRed << 8) | inRGB.fRed;
	outRGB.green	= (inRGB.fGreen << 8) | inRGB.fGreen;
	outRGB.blue		= (inRGB.fBlue << 8) | inRGB.fBlue;

	return outRGB;
}


// ---------------------------------------------------------------------------
//		 ConvertRGBTypeFromHost
// ---------------------------------------------------------------------------

RGBType		ConvertRGBTypeFromHost	(const RGBColor&  inRGB)
{
	RGBType	outRGB;

	outRGB.fRed		= inRGB.red >> 8;
	outRGB.fGreen	= inRGB.green >> 8;
	outRGB.fBlue	= inRGB.blue >> 8;

	return outRGB;
}


// ---------------------------------------------------------------------------
//		 ConvertFromRGBList
// ---------------------------------------------------------------------------

void ConvertFromRGBList (CTabHandle outColors, const RGBList& inColors)
{
	long	numColors	= inColors.size ();
	long	newSize		= offsetof(ColorTable, ctTable) +
							numColors * sizeof(ColorSpec);

	::SetHandleSize ((Handle) outColors, newSize);
	EmAssert (::MemError () == noErr);
	EmAssert (::GetHandleSize ((Handle) outColors) == newSize);

	if (numColors)
	{
		(**outColors).ctSeed	= GetCTSeed ();
		(**outColors).ctFlags	= 0;
		(**outColors).ctSize	= numColors - 1;

		for (int color = 0; color < numColors; ++color)
		{
			const RGBType&	srcColor = inColors[color];
			ColorSpec&		destColor = (**outColors).ctTable[color];

			destColor.value		= color;
			destColor.rgb.red	= RGB8To16 (srcColor.fRed);
			destColor.rgb.green	= RGB8To16 (srcColor.fGreen);
			destColor.rgb.blue	= RGB8To16 (srcColor.fBlue);
		}
	}
	else
	{
		(**outColors).ctSeed	= 0;
		(**outColors).ctFlags	= 0;
		(**outColors).ctSize	= 0;
	}
}


// ---------------------------------------------------------------------------
//		 ConvertToRGBList
// ---------------------------------------------------------------------------

void ConvertToRGBList (RGBList& outColors, CTabHandle inColors)
{
	outColors.clear ();

	long	numColors = (**inColors).ctSize + 1;

	for (long ii = 0; ii < numColors; ++ii)
	{
		const RGBColor&	srcColor = (**inColors).ctTable[ii].rgb;
		RGBType			rgb (srcColor.red, srcColor.green, srcColor.blue);

		outColors.push_back (rgb);
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvGetFormat
 *
 * DESCRIPTION:	Determine the format used by the PixMap, in terms of the
 *				values used by EmPixMap.
 *
 * PARAMETERS:	pixMap - PixMap describing the bitmap whose format
 *					we want to determine.
 *
 * RETURNED:	An EmPixMapFormat with the format value.
 *
 ***********************************************************************/

EmPixMapFormat PrvGetFormat (const PixMap& pixMap)
{
	EmPixMapFormat	format;

	if (pixMap.pixelSize == 1)
		format = kPixMapFormat1;

	else if (pixMap.pixelSize == 2)
		format = kPixMapFormat2;

	else if (pixMap.pixelSize == 4)
		format = kPixMapFormat4;

	else if (pixMap.pixelSize == 8)
		format = kPixMapFormat8;

	else if (pixMap.pixelSize == 16)
		format = kPixMapFormat16RGB555;

	else if (pixMap.pixelSize == 32)
		format = kPixMapFormat32ARGB;

	else
		EmAssert (false);

	return format;
}
