/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1999-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmTransportSerialMac.h"

#include "EmClassic.h"
#include "EmErrCodes.h"			// kError_CantDownloadROM_BadBaudRate
#include "ErrorHandling.h"		// Errors::ThrowIfError
#include "Logging.h"			// LogSerial

#include <CRMSerialDevices.h>	// crmSerialDevice
#include <CommResources.h>		// CRMRec
#include <Serial.h>				// SerSetBuf


#define PRINTF	if (!LogSerial ()) ; else LogAppendMsg


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostConstruct
 *
 * DESCRIPTION:	Construct platform-specific objects/data.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	The platform-specific serial object.
 *
 ***********************************************************************/

void EmTransportSerial::HostConstruct (void)
{
	fHost = new EmHostTransportSerial;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostDestruct
 *
 * DESCRIPTION:	Destroy platform-specific objects/data.
 *
 * PARAMETERS:	hostData - The platform-specific serial object.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmTransportSerial::HostDestruct (void)
{
	delete fHost;
	fHost = NULL;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostOpen
 *
 * DESCRIPTION:	Open the serial port in a platform-specific fashion.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	0 if no error.
 *
 ***********************************************************************/

ErrCode EmTransportSerial::HostOpen (void)
{
	PRINTF ("EmTransportSerial::HostOpen: attempting to open port \"%s\"",
			fConfig.fPort.c_str());

	ErrCode	result = noErr;

	if (0)
	{
		CRMRec		crmRec = {0};
		CRMRecPtr	crm = &crmRec;

		crmRec.crmDeviceType	= crmSerialDevice;
		crmRec.crmDeviceID		= 0;

		Str255	name, inPortName, outPortName;

		do {
			crm = CRMSearch (crm);
			if (crm)
			{
				CRMSerialPtr	portInfo = (CRMSerialPtr) crm->crmAttributes;

				memcpy (&name[0], &(*portInfo->name)[0], StrLength (*portInfo->name) + 1);
				memcpy (&inPortName[0], &(*portInfo->inputDriverName)[0], StrLength (*portInfo->inputDriverName) + 1);
				memcpy (&outPortName[0], &(*portInfo->outputDriverName)[0], StrLength (*portInfo->outputDriverName) + 1);

				LogAppendMsg ("name = %.*s, in = %.*s, out = %.*s",
					name[0], &name[1],
					inPortName[0], &inPortName[1],
					outPortName[0], &outPortName[1]);
			}
		} while (crm);
	}

	LStr255	portName (fConfig.fPort.c_str());
	LStr255	inDriverName;
	LStr255	outDriverName;

	// Get the port driver names.

	result = fHost->GetPortDriverNames (portName, inDriverName, outDriverName);
	if (result != noErr)
		goto Error;

	// Try opening drivers with those names.

	result = fHost->OpenDrivers (inDriverName, outDriverName);
	if (result != noErr)
		goto Error;

	return noErr;

Error:
	PRINTF ("EmTransportSerial::HostOpen: Can't open serial port (%ld).", result);

	this->HostClose ();

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostClose
 *
 * DESCRIPTION:	Close the serial port in a platform-specific fashion.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	0 if no error.
 *
 ***********************************************************************/

ErrCode EmTransportSerial::HostClose (void)
{
	// Close the serial port.

	if (fHost->fInPortRefNum)
	{
		PRINTF ("EmTransportSerial::HostClose: Closing input driver...");

		char	fakeBuffer;
		::SerSetBuf (fHost->fInPortRefNum, &fakeBuffer, 0);
		::CloseDriver (fHost->fInPortRefNum);
		fHost->fInPortRefNum = 0;
	}

	if (fHost->fOutPortRefNum)
	{
		PRINTF ("EmTransportSerial::HostClose: Closing output driver...");

		char	fakeBuffer;
		::SerSetBuf (fHost->fOutPortRefNum, &fakeBuffer, 0);
		::CloseDriver (fHost->fOutPortRefNum);
		fHost->fOutPortRefNum = 0;
	}

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostRead
 *
 * DESCRIPTION:	Read bytes from the port in a platform-specific fashion.
 *
 * PARAMETERS:	len - maximum number of bytes to read.
 *				data - buffer to receive the bytes.
 *
 * RETURNED:	0 if no error.  The number of bytes actually read is
 *				returned in len if there was no error.
 *
 ***********************************************************************/

ErrCode EmTransportSerial::HostRead (long& len, void* data)
{
	ErrCode	err = ::FSRead (fHost->fInPortRefNum, &len, (Ptr) data);

	return err;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostWrite
 *
 * DESCRIPTION:	Write bytes to the port in a platform-specific fashion.
 *
 * PARAMETERS:	len - number of bytes in the buffer.
 *				data - buffer containing the bytes.
 *
 * RETURNED:	0 if no error.  The number of bytes actually written is
 *				returned in len if there was no error.
 *
 ***********************************************************************/

ErrCode EmTransportSerial::HostWrite (long& len, const void* data)
{
	ErrCode	err = ::FSWrite (fHost->fOutPortRefNum, &len, (Ptr) data);

	return err;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostBytesInBuffer
 *
 * DESCRIPTION:	Returns the number of bytes that can be read with the
 *				Read method.  Note that bytes may be received in
 *				between the time BytesInBuffer is called and the time
 *				Read is called, so calling the latter with the result
 *				of the former is not guaranteed to fetch all received
 *				and buffered bytes.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Number of bytes that can be read.
 *
 ***********************************************************************/

long EmTransportSerial::HostBytesInBuffer (long /*minBytes*/)
{
	long	bytesRead;
	ErrCode	err = ::SerGetBuf (fHost->fInPortRefNum, &bytesRead);
	Errors::ThrowIfError (err);
	return bytesRead;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostSetConfig
 *
 * DESCRIPTION:	Configure the serial port in a platform-specific
 *				fasion.  The port is assumed to be open.
 *
 * PARAMETERS:	config - configuration information.
 *
 * RETURNED:	0 if no error.
 *
 ***********************************************************************/

ErrCode EmTransportSerial::HostSetConfig (const ConfigSerial& config)
{
	ErrCode	result = noErr;

	// Initialize the handshake settings.
	//
	// Note that I used to have:
	//
	//	Byte	v = config.fHwrHandshake ? 1 : 0;
	//
	// However, on the Windows side, Olivier Naudan argues:
	//
	//	But I don't agree with: dcb.fRtsControl  = config.fHwrHandshake ?
	//		RTS_CONTROL_HANDSHAKE : RTS_CONTROL_ENABLE;
	//	As hardware overrun can't be emulated, because of Poser timing that
	//	can't be accurate, I believe that fRtsControl should always be set to
	//	RTS_CONTROL_HANDSHAKE. So bytes will always arrive succesfully to host,
	//	and only software overrun could happen if PalmOS does not control the
	//	emulated RTS.

	SerShk	handshake;
	Byte	v = 1;	// config.fHwrHandshake ? 1 : 0;

	handshake.fXOn	= 0;		/* XOn/XOff output flow control flag */
	handshake.fCTS	= v;		/* hardware output flow control flags */
	handshake.xOn	= 0x11;		/* XOn character */
	handshake.xOff	= 0x13;		/* XOff character */
	handshake.errs	= 0;		/* errors mask bits */
	handshake.evts	= 0;		/* event enable mask bits */
	handshake.fInX	= 0;		/* XOn/XOff input flow control flag */
	handshake.fDTR	= v;		/* hardware input flow control flags */

	PRINTF ("	handshake.fXOn = %d", handshake.fXOn);
	PRINTF ("	handshake.fCTS = %d", handshake.fCTS);
	PRINTF ("	handshake.xOn  = %d", handshake.xOn);
	PRINTF ("	handshake.xOff = %d", handshake.xOff);
	PRINTF ("	handshake.errs = %d", handshake.errs);
	PRINTF ("	handshake.evts = %d", handshake.evts);
	PRINTF ("	handshake.fInX = %d", handshake.fInX);
	PRINTF ("	handshake.fDTR = %d", handshake.fDTR);

	// Technote DV 555 says that hardware handshaking is not fully
	// supported when using the SerHShake call.  Instead, we must
	// make a direct Control call.

//	result = ::SerHShake (gInReference, &handshake);
//	result = ::Control (gInReference, kSERDHandshake, &handshake);
//	if (result != noErr)
//		goto Error;

//	result = ::SerHShake (gOutReference, &handshake);
	result = ::Control (fHost->fOutPortRefNum, kSERDHandshake, &handshake);
	if (result != noErr)
		goto Error;

	// Initialize the port settings.

	short	settings = 0;

	settings |= fHost->GetBaud (config.fBaud);
	settings |= fHost->GetParity (config.fParity);
	settings |= fHost->GetStopBits (config.fStopBits);
	settings |= fHost->GetDataBits (config.fDataBits);

//	err = SerReset (fInReference, settings);
//	if (err != noErr)
//		goto Error;

	PRINTF ("	settings = %04X", (long) settings);
	result = ::SerReset (fHost->fOutPortRefNum, settings);
	if (result != noErr)
		goto Error;

	// Set up for very high baud rates (not available with all drivers)

	Byte serialVersion[2] = {};
	result = ::Status (fHost->fOutPortRefNum, kSERDVersion, serialVersion);
	if (result != noErr)
		goto Error;

	// First byte has version number
	// >= 9 indicates SerialDMA (or greater)

	if (serialVersion[0] >= 9)
	{
		if (config.fBaud == 115200)
		{
			PRINTF ("	Making 115K baud Control call.");
			result = ::Control (fHost->fOutPortRefNum, kSERD115KBaud, 0);
		}
		else if (config.fBaud == 230400)
		{
			PRINTF ("	Making 230K baud Control call.");
			result = ::Control (fHost->fOutPortRefNum, kSERD230KBaud, 0);
		}
	}
	else
	{
		if (config.fBaud == 115200 || config.fBaud == 230400)
		{
			result = kError_CantDownloadROM_BadBaudRate;	// !!! Wrong error number
			goto Error;
		}
	}

Error:
	if (result != noErr)
		PRINTF ("EmTransportSerial::HostSetConfig: Can't set settings (%ld).", (long) result);

	return result;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostSetRTS
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmTransportSerial::HostSetRTS (RTSControl /*state*/)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostSetDTR
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmTransportSerial::HostSetDTR (Bool /*state*/)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostSetBreak
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmTransportSerial::HostSetBreak (Bool /*state*/)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostGetCTS
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

Bool EmTransportSerial::HostGetCTS (void)
{
	return false;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostGetDSR
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

Bool EmTransportSerial::HostGetDSR (void)
{
	return false;
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostGetPortNameList
 *
 * DESCRIPTION:	Return the list of serial ports on this computer.  Used
 *				to prepare a menu of serial port choices.
 *
 * PARAMETERS:	nameList - port names are added to this list.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmTransportSerial::HostGetPortNameList (PortNameList& results)
{
	if (::EmClassicHasSerialManager ())
	{
		CRMRec		crmRec = {0};
		CRMRecPtr	crm = &crmRec;

		crmRec.crmDeviceType	= crmSerialDevice;
		crmRec.crmDeviceID		= 0;

		while ((crm = ::CRMSearch (crm)) != NULL)
		{
			CRMSerialPtr	portInfo = (CRMSerialPtr) crm->crmAttributes;
			StringPtr		namePtr = *portInfo->name;

			// Copy the name into a temp buffer on the stack.  That
			// way, we don't have to worry about the handle moving
			// when we create the string object.

			char			portName[256];
			portName[namePtr[0]] = 0;
			::BlockMove (&namePtr[1], portName, namePtr[0]);

			results.push_back (EmTransportSerial::PortName (portName));
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmTransportSerial::HostGetSerialBaudList
 *
 * DESCRIPTION:	Return the list of baud rates support by this computer.
 *				Used to prepare a menu of baud rate choices.
 *
 * PARAMETERS:	baudList - baud rates are added to this list.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmTransportSerial::HostGetSerialBaudList (BaudList& results)
{
	long	maxBaud = 115200;	// ::PrvGetMaxBaudRate ()? But that
								// requires the port to be open (in
								// order to make the control call that
								// returns the information).

	switch (maxBaud)
	{
		case 115200:	results.push_back (115200);
		case 57600:		results.push_back (57600);
		case 38400:		results.push_back (38400);
		case 28800:		results.push_back (28800);
		case 19200:		results.push_back (19200);
		case 14400:		results.push_back (14400);
		case 9600:		results.push_back (9600);
	}
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	EmHostTransportSerial c'tor
 *
 * DESCRIPTION:	Constructor.  Initialize our data members.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmHostTransportSerial::EmHostTransportSerial (void) :
	fInPortRefNum (0),
	fOutPortRefNum (0)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmHostTransportSerial d'tor
 *
 * DESCRIPTION:	Destructor.  Delete our data members.
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmHostTransportSerial::~EmHostTransportSerial (void)
{
}


/***********************************************************************
 *
 * FUNCTION:	EmHostTransportSerial::GetPortDriverNames
 *
 * DESCRIPTION:	Get the names of the input and output drivers to be
 *				used with the given serial port.
 *
 * PARAMETERS:	portName - name of the serial port for which we want
 *					the drivers.
 *				inDriverName - name of input driver.
 *				outDriverName - name of output driver.
 *
 * RETURNED:	0 if no error.
 *
 ***********************************************************************/

ErrCode EmHostTransportSerial::GetPortDriverNames (const LStr255& portName,
													LStr255& inDriverName,
													LStr255& outDriverName)
{
	// Get the port driver names.

	CRMRec		crmRec = {0};
	CRMRecPtr	crm = &crmRec;

	crmRec.crmDeviceType	= crmSerialDevice;
	crmRec.crmDeviceID		= 0;

	while (1)
	{
		crm = ::CRMSearch (crm);

		if (crm == NULL)
		{
			PRINTF ("EmTransportSerial::GetPortDriverNames: can't find a serial port.");
			return kError_CantFindSerialPort;
		}

		CRMSerialPtr	portInfo = (CRMSerialPtr) crm->crmAttributes;
		LStr255			thisPortName (*portInfo->name);

//		PRINTF ("EmTransportSerial::GetPortDriverNames: found port \"%.*s\"", thisPortName[0], &thisPortName[1]);

		if (thisPortName == portName)
		{
			inDriverName = *portInfo->inputDriverName;
			outDriverName = *portInfo->outputDriverName;
			return errNone;
		}
	}

	return -1;
}


/***********************************************************************
 *
 * FUNCTION:	EmHostTransportSerial::OpenDrivers
 *
 * DESCRIPTION:	Open the given input and output serial drivers and set
 *				larger buffers for them.  If an error occurs, any
 *				partial work is NOT cleaned up.
 *
 * PARAMETERS:	outDriverName - name of the output driver.
 *
 *				inDriverName - name of the input driver
 *
 * RETURNED:	Any errors from OpenDriver or SerSetBuf.
 *
 ***********************************************************************/

ErrCode EmHostTransportSerial::OpenDrivers	(const LStr255& inDriverName,
											 const LStr255& outDriverName)
{
	OSErr	result;

	// Open the drivers for the in/out ports.

	PRINTF ("EmHostTransportSerial::OpenDrivers: Opening output driver (%#s)...",
			(ConstStringPtr) outDriverName);

	result = ::OpenDriver (outDriverName, &fOutPortRefNum);
	if (result != noErr)
		return result;

	PRINTF ("EmHostTransportSerial::OpenDrivers: Opening input driver (%#s)...",
			(ConstStringPtr) inDriverName);

	result = ::OpenDriver (inDriverName, &fInPortRefNum);
	if (result != noErr)
		return result;

	// Give us some bigger buffers.  We have a hard time handling
	// > 256 byte packets with the defaults.

	PRINTF ("EmHostTransportSerial::OpenDrivers: Setting buffers...");

	result = ::SerSetBuf (fInPortRefNum, fInBuffer, sizeof (fInBuffer));
	if (result != noErr)
		return result;

	result = ::SerSetBuf (fOutPortRefNum, fOutBuffer, sizeof (fOutBuffer));
	if (result != noErr)
		return result;

	return noErr;
}


/***********************************************************************
 *
 * FUNCTION:	EmHostTransportSerial::GetBaud
 *
 * DESCRIPTION:	Map a baud rate into the Mac OS constant that represents
 *				that rate in a SerSettings call.
 *
 * PARAMETERS:	baud - raw baud rate.
 *
 * RETURNED:	Mac OS constant that represents that rate.
 *
 ***********************************************************************/

short EmHostTransportSerial::GetBaud (EmTransportSerial::Baud baud)
{
	switch (baud)
	{
		case    150:		PRINTF ("	Baud = 150");		return baud150;
		case    300:		PRINTF ("	Baud = 300");		return baud300;
		case    600:		PRINTF ("	Baud = 600");		return baud600;
		case   1200:		PRINTF ("	Baud = 1200");		return baud1200;
		case   1800:		PRINTF ("	Baud = 1800");		return baud1800;
		case   2400:		PRINTF ("	Baud = 2400");		return baud2400;
		case   3600:		PRINTF ("	Baud = 3600");		return baud3600;
		case   4800:		PRINTF ("	Baud = 4800");		return baud4800;
		case   7200:		PRINTF ("	Baud = 7200");		return baud7200;
		case   9600:		PRINTF ("	Baud = 9600");		return baud9600;
		case  14400:		PRINTF ("	Baud = 14400");		return baud14400;
		case  19200:		PRINTF ("	Baud = 19200");		return baud19200;
		case  28800:		PRINTF ("	Baud = 28800");		return baud28800;
		case  38400:		PRINTF ("	Baud = 38400");		return baud38400;
		case  57600:		PRINTF ("	Baud = 57600");		return baud57600;
		case 115200:		PRINTF ("	Baud = 115200");	return baud57600;
		case 230400:		PRINTF ("	Baud = 230400");	return baud57600;
	}

	PRINTF ("	Unknown Baud value.");
	EmAssert (false);
	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	EmHostTransportSerial::GetParity
 *
 * DESCRIPTION:	Map a parity value into the Mac OS constant that
 *				represents that parity in a SerSettings call.
 *
 * PARAMETERS:	parity - raw parity value.
 *
 * RETURNED:	Mac OS constant that represents that parity value.
 *
 ***********************************************************************/

short EmHostTransportSerial::GetParity (EmTransportSerial::Parity parity)
{
	switch (parity)
	{
		case EmTransportSerial::kNoParity:		PRINTF ("	parity = none");	return noParity;
		case EmTransportSerial::kOddParity:		PRINTF ("	parity = odd");		return oddParity;
		case EmTransportSerial::kEvenParity:	PRINTF ("	parity = even");	return evenParity;
	}

	PRINTF ("	Unknown Parity value.");
	EmAssert (false);
	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	EmHostTransportSerial::GetStopBits
 *
 * DESCRIPTION:	Map a stopBits value into the Mac OS constant that
 *				represents that value in a SerSettings call.
 *
 * PARAMETERS:	stopBits - raw stop bits value.
 *
 * RETURNED:	Mac OS constant that represents that stopBits value.
 *
 ***********************************************************************/

short EmHostTransportSerial::GetStopBits (EmTransportSerial::StopBits stopBits)
{
	switch (stopBits)
	{
		case 1:				PRINTF ("	stopBits = 1");		return stop10;
		case 2:				PRINTF ("	stopBits = 2");		return stop20;
	}

	PRINTF ("	Unknown StopBits value.");
	EmAssert (false);
	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	EmHostTransportSerial::GetDataBits
 *
 * DESCRIPTION:	Map a dataBits value into the Mac OS constant that
 *				represents that value in a SerSettings call.
 *
 * PARAMETERS:	dataBits - raw data bits value.
 *
 * RETURNED:	Mac OS constant that represents that dataBits value.
 *
 ***********************************************************************/

short EmHostTransportSerial::GetDataBits (EmTransportSerial::DataBits dataBits)
{
	switch (dataBits)
	{
		case 7:				PRINTF ("	dataBits = 7");		return data7;
		case 8:				PRINTF ("	dataBits = 8");		return data8;
	}

	PRINTF ("	Unknown DataBits value.");
	EmAssert (false);
	return 0;
}

