/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmWindowMac.h"

#include "EmDocumentMac.h"		// gHostDocument
#include "EmMenusMac.h"			// HostCreatePopupMenu
#include "EmPixMapMac.h"		// ConvertRGBTypeToHost
#include "EmRegionMac.h"		// ConvertRegionToHost
#include "EmScreen.h"			// EmScreenUpdateInfo
#include "EmWindow.h"			// EmWindow
#include "ErrorHandling.h"		// Errors::ThrowIfError
#include "Platform.h"			// Platform::DisposeMemory

#include "EmApplicationMac.h"	// gHostApplication->GetEmulatorDoc
#include "EmDocumentMac.h"		// To convert GetEmulatorDoc to LCommander
#include "Palm OS Emulator.r.h"	// pict_DefaultSmall, pict_DefaultLarge


EmWindowMac*	gHostWindow;


// ---------------------------------------------------------------------------
//		 EmWindow::NewWindow
// ---------------------------------------------------------------------------

EmWindow* EmWindow::NewWindow (void)
{
	// Create the new PowerPlant window.

	StDeleter<EmWindowMac>	theWindow (new EmWindowMac (gHostDocument));

	// Platform-specific post-creation initialization.

	LView::SetDefaultView (theWindow);
	LCommander::SetDefaultCommander (theWindow);
	LAttachable::SetDefaultAttachable (NULL);

	theWindow->FinishCreate ();

	// Cross-platform post-creation initialization.

	theWindow->WindowInit ();

	// Release the window from the StDeleter and return it.

	return theWindow.Release ();
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmWindowMac::EmWindowMac
// ---------------------------------------------------------------------------

EmWindowMac::EmWindowMac (LCommander* inSuperCommander) :
	LWindow (EmWindowMac::CreateMacWindow (), inSuperCommander),
	EmWindow (),
	fDragAndDrop (mMacWindowP, this),
	fLastEvent (),
	fCachedSkin (NULL)
{
	EmAssert (gHostWindow == NULL);
	gHostWindow = this;
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::~EmWindowMac
// ---------------------------------------------------------------------------

EmWindowMac::~EmWindowMac (void)
{
	this->PreDestroy ();
	this->CacheFlush ();

	EmAssert (gHostWindow == this);
	gHostWindow = NULL;
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::DrawSelf
// ---------------------------------------------------------------------------

void EmWindowMac::DrawSelf (void)
{
	this->HandleUpdate ();
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::ClickSelf
// ---------------------------------------------------------------------------
// Notifies the emulated CPU of the mouse down event.

void EmWindowMac::ClickSelf (const SMouseDownEvent& evt)
{
	// Do any base-class processing (currently none).

	LPane::ClickSelf (evt);

	// See if we need to show and track the popup menu.
	
	if (LCommander::GetTarget () && (evt.macEvent.modifiers & controlKey))
	{
		this->PopupMenu (evt.macEvent.where);
		return;
	}
	
	// Perform common processing.

	this->MouseCommon (evt.macEvent, true);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::EventMouseUp
// ---------------------------------------------------------------------------
// Notifies the emulated CPU of the mouse up event.

void EmWindowMac::EventMouseUp (const EventRecord& macEvent)
{
	// Do any base-class processing (record sWhenLastMouseUp).

	LPane::EventMouseUp (macEvent);
	
	// Perform common processing.

	this->MouseCommon (macEvent, false);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::ActivateSelf
// ---------------------------------------------------------------------------
//	Pane is being Activated

void EmWindowMac::ActivateSelf (void)
{
	this->HandleActivate (true);
	LWindow::ActivateSelf ();
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::DeactivateSelf
// ---------------------------------------------------------------------------
//	Pane is being Deactivated

void EmWindowMac::DeactivateSelf (void)
{
	this->HandleActivate (false);
	LWindow::DeactivateSelf ();
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::MouseCommon
// ---------------------------------------------------------------------------

void EmWindowMac::MouseCommon (const EventRecord& event, Bool down)
{
	// Save the incoming event for later, if needed.

	this->SetLastEvent (event);

	// Convert from global/screen to local/window coordinates.

	Point	where = event.where;
	this->FocusDraw ();
	::GlobalToLocal (&where);

	// Perform common processing.

	this->HandlePenEvent (EmPoint (where), down);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::PopupMenu
// ---------------------------------------------------------------------------

void EmWindowMac::PopupMenu (const Point& where)
{
	EmMenu*	menu		= ::MenuFindMenu (kMenuPopupMenuPreferred);
	EmAssert (menu);

	::MenuUpdateMruMenus (*menu);
	::MenuUpdateMenuItemStatus (*menu);

	LMenu*	hostMenu	= ::HostCreatePopupMenu (*menu);
	EmAssert (hostMenu);

	long	menuChoice	= ::PopUpMenuSelect (hostMenu->GetMacMenuH (),
												where.v, where.h, 1);

	if (HiWord (menuChoice) != 0)
	{
		const EmMenuItem*	selected = ::HostFindMenuItemByMenuChoice (
											*menu, menuChoice, true);
		EmAssert (selected);

		if (selected)
		{
			EmCommandID	command = selected->GetCommand ();
			LCommander::SetUpdateCommandStatus (true);
			LCommander::GetTarget ()->ProcessCommand (command, &menuChoice);
		}
	}
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::SetLastEvent
// ---------------------------------------------------------------------------

void EmWindowMac::SetLastEvent (const EventRecord& event)
{
	fLastEvent = event;
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::GetLastEvent
// ---------------------------------------------------------------------------

EventRecord EmWindowMac::GetLastEvent (void)
{
	return fLastEvent;
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::CacheFlush
// ---------------------------------------------------------------------------

void EmWindowMac::CacheFlush (void)
{
	if (fCachedSkin)
	{
		::DisposeGWorld (fCachedSkin);
		fCachedSkin = NULL;
	}
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::GetSkin
// ---------------------------------------------------------------------------

GWorldPtr EmWindowMac::GetSkin (void)
{
	if (!fCachedSkin)
	{
		EmPoint	size = this->GetCurrentSkin ().GetSize ();

		::ConvertPixMapToHost (	this->GetCurrentSkin (), fCachedSkin,
								0, size.fY, false);
	}

	EmAssert (fCachedSkin);

	return fCachedSkin;
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::GetSystemPalette
// ---------------------------------------------------------------------------

PaletteHandle EmWindowMac::GetSystemPalette (const EmScreenUpdateInfo& info)
{
	// Skip if the screen is not an indexed device.
	
	if (this->DirectDevices ())
	{
		return NULL;
	}

	// Check to see if the colors have changed.  If not, we don't need a
	// new palette.

	const RGBList& newColorTable = info.fImage.GetColorTable ();
	if (!this->PrevLCDColorsChanged (newColorTable))
	{
		return NULL;
	}

	this->SaveLCDColors (newColorTable);

	// Create the palette.

	RGBList	colors;
	this->GetSystemColors (info, colors);

	CTabHandle	cTable = (CTabHandle) ::NewHandle (0);
	::ConvertFromRGBList (cTable, colors);

	PaletteHandle	result = ::NewPalette (colors.size (),
								cTable, pmTolerant, 0);

	::DisposeCTable (cTable);

	// Return the palette.

	return result;
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::DirectDevices
// ---------------------------------------------------------------------------

Bool EmWindowMac::DirectDevices (void)
{
	GDHandle	dev = ::GetDeviceList ();
	while (dev)
	{
		if ((**dev).gdType != directType)
			return false;

		dev = ::GetNextDevice (dev);
	}

	return true;
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::CreateMacWindow
// ---------------------------------------------------------------------------

WindowRef EmWindowMac::CreateMacWindow (void)
{
	// Create the new Mac window.

	WindowDefSpec		wdefSpec;
	wdefSpec.defType	= kWindowDefProcPtr;
	wdefSpec.u.defProc	= NewWindowDefUPP (&EmWindowMac::WDEF);

	WindowClass			windClass	= kDocumentWindowClass;
	WindowAttributes	windAttr	= kWindowNoAttributes;
	Rect				windBounds	= { 40, 20, 440, 270 };

	WindowRef			newWindow;
	ThrowIfOSStatus_ (::CreateCustomWindow (&wdefSpec, windClass,
								windAttr, &windBounds, &newWindow));
	ThrowIfNil_ (newWindow);

	return newWindow;
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::WDEF
// ---------------------------------------------------------------------------

pascal long EmWindowMac::WDEF (	short		varCode,
										WindowRef	window,
										short		message,
										long		param)
{
	#pragma unused (varCode)

	// Initialize the return value.

	long	result = 0;

	// Handle the message.

	switch (message)
	{
		case kWindowMsgDraw:
			// Nothing to draw.
			break;

		case kWindowMsgHitTest:

			// Return the part of the window that is currently under the mouse.

			result = wNoHit;

			if (::IsWindowVisible (window))
			{
				result = wInContent;
			}

			break;

		case kWindowMsgCalculateShape:
			// We don't do this one...we do kWindowMsgGetRegion instead.
			// I think that's the only way under Carbon, since we can't
			// get at the window's content and structure region handles
			// in order to change them like we did under pre-Carbon.
			break;

		case kWindowMsgInitialize:
			// Nothing to initialize.
			break;

		case kWindowMsgCleanUp:
			// Nothing to clean up.
			break;

		case kWindowMsgDrawGrowOutline:
			// Nothing to draw.
			break;

		case kWindowMsgDrawGrowBox:
			// Nothing to draw.
			break;

		case kWindowMsgGetFeatures:
			*(OptionBits*) param = kWindowCanGetWindowRegion |
									kWindowDefSupportsColorGrafPort;
			result = 1;
			break;

		case kWindowMsgGetRegion:
		{
			// Calculate the content and structure regions of the window.

			GetWindowRegionRec* rgnRec = (GetWindowRegionRec*) param;

			if (window &&
				((rgnRec->regionCode == kWindowContentRgn) ||
				 (rgnRec->regionCode == kWindowStructureRgn)))
			{
				// Get the region handle from the cross-platform region.

				LWindow*	windObject = LWindow::FetchWindowObject (window);

				// windObject may be NULL if the Mac window has been created,
				// the the EmWindowMac or LWindow hasn't been fully constructed yet.

				if (!windObject)
					break;

				EmWindowMac*	This = dynamic_cast<EmWindowMac*>(windObject);
				EmAssert (This);

				if (!This)
					break;

				RgnHandle	rgn	= ::ConvertRegionToHost (This->GetCurrentSkinRegion ());

				// Move the outline region to where the window is.  First,
				// zero-justify it.  Then move it to the global window bounds.

				Rect		bounds;
				::GetRegionBounds (rgn, &bounds);
				::OffsetRgn (rgn, -bounds.left, -bounds.top);

				::GetWindowBounds (window, kWindowGlobalPortRgn, &bounds);
				::OffsetRgn (rgn, bounds.left, bounds.top);

				// Copy the region to the window region.

				::MacCopyRgn (rgn, rgnRec->winRgn);

				::DisposeRgn (rgn);
			}

			break;
		}
	}

	return result;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmWindowMac::HostWindowReset
// ---------------------------------------------------------------------------
// Update the window's appearance due to a skin change.

void EmWindowMac::HostWindowReset (void)
{
	// Delete the old GWorld.

	this->CacheFlush ();

	// Resize our window according to the size of the skin.

	EmRect		newBounds	= this->GetCurrentSkinRegion ().Bounds ();
	this->ResizeWindowTo (newBounds.Width (), newBounds.Height ());

	// Redraw the contents (new skin, LCD in new position, etc.)

	this->Refresh ();
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostDrawingBegin
// ---------------------------------------------------------------------------
// Prepare the host window object for drawing outside of an update event.

void EmWindowMac::HostDrawingBegin (void)
{
	this->FocusDraw ();
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostPaletteSet
// ---------------------------------------------------------------------------
// Establish the palette to be used for drawing.

void EmWindowMac::HostPaletteSet (const EmScreenUpdateInfo& info)
{
	PaletteHandle	pal = this->GetSystemPalette (info);

	if (pal)
	{
		::NSetPalette (this->GetMacWindow (), pal, pmAllUpdates);
	}
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostWindowMoveBy
// ---------------------------------------------------------------------------
// Move the host window object by the given offset.

void EmWindowMac::HostWindowMoveBy (const EmPoint& offset)
{
	this->MoveWindowBy (offset.fX, offset.fY);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostWindowMoveTo
// ---------------------------------------------------------------------------
// Move the host window object to the given location.

void EmWindowMac::HostWindowMoveTo (const EmPoint& loc)
{
	this->MoveWindowTo (loc.fX, loc.fY);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostWindowBoundsGet
// ---------------------------------------------------------------------------
// Get the global bounds of the host window object.

EmRect EmWindowMac::HostWindowBoundsGet (void)
{
	Rect	bounds;
	this->GetGlobalBounds (bounds);

	return EmRect (bounds);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostWindowCenter
// ---------------------------------------------------------------------------
// Center the window to the main display.

void EmWindowMac::HostWindowCenter (void)
{
	BitMap	screenBits;
	::GetQDGlobalsScreenBits (&screenBits);
	Rect	screenBounds = screenBits.bounds;

	SDimension16	frameSize;
	this->GetFrameSize (frameSize);
	
	SPoint32		location;
	location.h = ((screenBounds.right - screenBounds.left) - frameSize.width) / 2;
	location.v = ((screenBounds.bottom - screenBounds.top) - frameSize.height) / 3;

	this->MoveWindowTo (location.h, location.v);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostWindowShow
// ---------------------------------------------------------------------------
// Make the host window object visible.

void EmWindowMac::HostWindowShow (void)
{
	this->Show ();

	// For some reason we have to select the window to activate the palette.

	WindowRef	window = this->GetMacWindow ();
	if (window != NULL)
	{
		::SelectWindow (window);
	}
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostWindowDrag
// ---------------------------------------------------------------------------
// The user has clicked in a region of the host window object that causes
// the window to be dragged.  Drag the window around.

void EmWindowMac::HostWindowDrag (void)
{
	this->ClickInDrag (this->GetLastEvent ());
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostRectFrame
// ---------------------------------------------------------------------------
// Draw a rectangle frame with the given width in the given color.

void EmWindowMac::HostRectFrame (const EmRect& r, const EmPoint& pen, const RGBType& color)
{
	// Save the old pen state.

	PenState	penState;
	::GetPenState (&penState);

	// Initialize the pen state to something clean, and then set the pen
	// size to what the caller specified.

	::PenNormal ();
	::PenSize (pen.fX, pen.fY);

	// Set the pen color to what the caller specified.

	RGBColor	rgb = ::ConvertRGBTypeToHost (color);
	::RGBForeColor (&rgb);

	// Draw the rectangle.

	Rect	macRect = r;
	::FrameRect (&macRect);

	// Completely restore the original pen state.

	::SetPenState (&penState);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostOvalPaint
// ---------------------------------------------------------------------------
// Fill an oval with the given color.

void EmWindowMac::HostOvalPaint (const EmRect& r, const RGBType& color)
{
	// Save the old pen state.

	PenState	penState;
	::GetPenState (&penState);

	// Initialize the pen state to something clean.

	::PenNormal ();

	// Set the pen color to what the caller specified.

	RGBColor	rgb = ::ConvertRGBTypeToHost (color);
	::RGBForeColor (&rgb);

	// Draw the rectangle.

	Rect	macRect = r;
	::PaintOval (&macRect);

	// Completely restore the original pen state.

	::SetPenState (&penState);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostPaintCase
// ---------------------------------------------------------------------------
// Draw the skin.

void EmWindowMac::HostPaintCase (const EmScreenUpdateInfo&)
{
	// Draw into it.

	GrafPtr			destPort	= this->GetMacPort ();
	GWorldPtr		gWorld		= this->GetSkin ();
	PixMapHandle	pixMap		= ::GetGWorldPixMap (gWorld);
	Rect			destRect	= (**pixMap).bounds;

	::LockPixels (pixMap);

	::RGBBackColor (&Color_White);
	::RGBForeColor (&Color_Black);

	::CopyBits (	(BitMap*) *pixMap, ::GetPortBitMapForCopyBits (destPort),
					&destRect, &destRect, srcCopy, NULL);

	::UnlockPixels (pixMap);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostPaintLCD
// ---------------------------------------------------------------------------
// Draw the LCD area.

void EmWindowMac::HostPaintLCD (const EmScreenUpdateInfo& info, const EmRect& srcRect,
						  const EmRect& destRect, Bool scaled)
{
	GWorldPtr	gWorld;
	::ConvertPixMapToHost (	info.fImage, gWorld,
							info.fFirstLine, info.fLastLine,
							scaled);

	if (!gWorld)
		return;

	PixMapHandle	pixMap = ::GetGWorldPixMap (gWorld);
	::LockPixels (pixMap);

	::RGBBackColor (&Color_White);
	::RGBForeColor (&Color_Black);

	GrafPtr	destPort	= this->GetMacPort ();
	Rect	macSrcRect	= srcRect;
	Rect	macDestRect	= destRect;

	::CopyBits (	(BitMap*) *pixMap, ::GetPortBitMapForCopyBits (destPort),
					&macSrcRect, &macDestRect, srcCopy, NULL);

	::UnlockPixels (pixMap);
	::DisposeGWorld (gWorld);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostGetDefaultSkin
// ---------------------------------------------------------------------------
// Get the default (built-in) skin image.

void EmWindowMac::HostGetDefaultSkin (EmPixMap& pixMap, int scale)
{
	// Get the default picture.

	short		bitmapID = scale == 1 ? pict_DefaultSmall : pict_DefaultLarge;
	StResource	bitmap (ResType_Picture, bitmapID);
	PicHandle	picture = (PicHandle) bitmap.Get ();

	EmAssert (picture);
	if (picture == NULL)
		return;

	// Create the output GWorld.

	GWorldPtr	gWorld	= NULL;
	Rect		bounds	= (**picture).picFrame;
	OSErr		err		= ::NewGWorld (&gWorld, 24, &bounds,
								/* colorTable */ NULL, /* gDevice */ NULL,
								useTempMem | keepLocal);

	EmAssert (!err);
	if (err)
		return;

	// Save the current device/port.

	CGrafPtr	oldPort;
	GDHandle	oldDevice;
	::GetGWorld (&oldPort, &oldDevice);

	// Switch to the offscreen GWorld and draw the skin into it.

	PixMapHandle	pixBase = ::GetGWorldPixMap (gWorld);

	::LockPixels (pixBase);
	::SetGWorld (gWorld, NULL);
	::DrawPicture (picture, &bounds);
	::SetGWorld (oldPort, oldDevice);
	::UnlockPixels (pixBase);

	// Convert the result into a platform-independent PixMap.

	::CreatePixMapFromHost (gWorld, pixMap);

	// Clean up.

	::DisposeGWorld (gWorld);
}


// ---------------------------------------------------------------------------
//		 EmWindowMac::HostGetCurrentMouse
// ---------------------------------------------------------------------------

EmPoint EmWindowMac::HostGetCurrentMouse (void)
{
	this->FocusDraw ();

	Point	mouse;
	::GetMouse (&mouse);	// Receives location in local coordinates.

	EmPoint	where (mouse);
	return where;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmDragAndDrop::EmDragAndDrop
// ---------------------------------------------------------------------------

EmDragAndDrop::EmDragAndDrop (WindowPtr	inMacWindow,
								LPane*		inPane) :
	LDragAndDrop (inMacWindow, inPane)
{
}


// ---------------------------------------------------------------------------
//		 EmDragAndDrop::~EmDragAndDrop
// ---------------------------------------------------------------------------

EmDragAndDrop::~EmDragAndDrop (void)
{
}


// ---------------------------------------------------------------------------
//		 EmDragAndDrop::ItemIsAcceptable
// ---------------------------------------------------------------------------
//	We'll accept anything that's a file system object for now.  We'll
//	filter based on other criteria after the user has dropped the files
//	on us.  That way, we can put up an error message if they drop the
//	wrong stuff on us.

Boolean EmDragAndDrop::ItemIsAcceptable (	DragReference	inDragRef,
											ItemReference	inItemRef)
{
	FlavorFlags		theFlags;
	OSErr			err = ::GetFlavorFlags (	inDragRef,
												inItemRef,
												flavorTypeHFS,
												&theFlags);

	return err == noErr;
}


// ---------------------------------------------------------------------------
//		 EmDragAndDrop::DoDragReceive
// ---------------------------------------------------------------------------
//	Package this sucker up into an Apple event so that we can open the files.

void EmDragAndDrop::DoDragReceive (DragReference	inDragRef)
{
	Try_
	{
		// Create the list of files to open.

		StAEDescriptor	fileList;
		OSErr	err = ::AECreateList (NULL, 0, false, fileList);

		// Get the number of dragged items so that we can iterate
		// over the DragRef.

		UInt16	itemCount;
		::CountDragItems (inDragRef, &itemCount);

		for (UInt16 ii = 1; ii <= itemCount; ++ii)
		{
			// Get the nth item.

			ItemReference	itemRef;
			::GetDragItemReferenceNumber(inDragRef, ii, &itemRef);

			// Get the FSSpec (and some other information) from it.

			FSSpec	droppedFile;
			if (this->GetSpecFromItem (inDragRef, itemRef, droppedFile))
			{
				OSErr	err = ::AEPutPtr (fileList, ii, typeFSS,
									&droppedFile, sizeof(FSSpec));
				Errors::ThrowIfError (err);
			}
		}

		// Now create the "Open" Apple event.
		//
		// Note that we _don't_ call UAppleEventsMgr::MakeAppleEvent to
		// create the Apple event.  That will create an event that is
		// "direct sent" to the application.  We don't want that; we want
		// the event to be posted to the event queue.  Doing the direct
		// send thing will cause dialogs to appear in the context of the
		// "drag dropped" event, which causes hanging problems.  This
		// context also appears to cause AEInteractWithUser to never
		// switch us to front.
		//
		// "Direct sends" are triggered by using an address descriptor
		// record of descriptor type typeProcessSerialNumber with the
		// lowLongOfPSN field set to kCurrentProcess and the highLongOfPSN
		// set to 0.  We avoid that condition by getting the real
		// process serial number with GetCurrentProcess.

		ProcessSerialNumber	currProcess;
		err = ::GetCurrentProcess (&currProcess);
		Errors::ThrowIfError (err);

		StAEDescriptor	selfTargetAddr;
		err = ::AECreateDesc (typeProcessSerialNumber, (Ptr) &currProcess,
								sizeof(currProcess), selfTargetAddr);
		Errors::ThrowIfError (err);

		AppleEvent	openEvent;
		err = ::AECreateAppleEvent (kCoreEventClass, kAEOpen,
											selfTargetAddr,
											kAutoGenerateReturnID,
											kAnyTransactionID,
											&openEvent);
		Errors::ThrowIfError (err);

		// Add our file list as the direct object.
		
		err = ::AEPutParamDesc (&openEvent, keyDirectObject, fileList);
		Errors::ThrowIfError (err);

		// Send it away...
		
		UAppleEventsMgr::SendAppleEvent (openEvent);
	}
	Catch_(inErr)
	{
	} EndCatch_
}


// ---------------------------------------------------------------------------
//		 EmDragAndDrop::GetSpecFromItem
// ---------------------------------------------------------------------------

Bool EmDragAndDrop::GetSpecFromItem (	DragReference	inDragRef,
										ItemReference	inItemRef,
										FSSpec&			outSpec)
{
	FlavorFlags		theFlags;
	OSErr			err = ::GetFlavorFlags (	inDragRef,
												inItemRef,
												flavorTypeHFS,
												&theFlags);
	if (err == noErr)
	{
		Size	dataSize;
		::GetFlavorDataSize (inDragRef, inItemRef, flavorTypeHFS, &dataSize);

		if (dataSize <= sizeof (HFSFlavor))
		{
			HFSFlavor	hfsFlavor;
			::GetFlavorData (inDragRef, inItemRef, flavorTypeHFS, &hfsFlavor,
							 &dataSize, 0);
			outSpec = hfsFlavor.fileSpec;
			return true;
		}
	}

	return false;
}
