/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "Platform.h"

#include "ChunkFile.h"			// ReadChunk
#include "EmClassic.h"			// EmClassicHasPPCToolbox
#include "EmDocumentMac.h"		// gHostDocument, SetPendingException
#include "EmSession.h"			// EmSessionStopper, gSession
#include "ErrorHandling.h"		//
#include "Hordes.h"				// Hordes::CanStop, etc.
#include "Logging.h"			// LogAppendMsg
#include "Miscellaneous.h"		// PinToScreen
#include "ROMStubs.h"			// FtrGet
#include "Palm OS Emulator.r.h" // ????, etc.
#include "PPCMessaging.h"		// PPCOpenPorts, etc
#include "SessionFile.h"		// SessionFile::kROMAliasTag
#include "SLP.h"				// SLP::EventCallback
#include "SocketMessaging.h"	// CSocket
#include "Strings.r.h"			// kStr_ values

#include <DriverServices.h>		// UpTime, AddDurationToAbsolute, MPDelayUntil
#include <Timer.h>				// Microseconds


// ===========================================================================
//		 fixed_size_deque
// ===========================================================================
//	A simple deque-like class.	We've created this deque because we need on
//	that doesn't allocate any memory.  This deque uses a fixed-size buffer.

const int kMaxElts = 4 * 1024;	// arbitrary size

// DEFTEMPARG is defined for and required by CodeWarrior 5 and 5.1.
// It appears to have disappeared in 5.2.

template <class T, class Allocator = allocator<T> >
class fixed_size_deque
{
	public:
								fixed_size_deque () : fNumElts (0) {}
								~fixed_size_deque () {}

		typedef Allocator							allocator_type;
		typedef allocator_type::size_type			size_type;
		typedef allocator_type::difference_type 	difference_type;
		typedef allocator_type::reference			reference;
		typedef allocator_type::const_reference 	const_reference;
		typedef allocator_type::pointer 			pointer;
		typedef allocator_type::const_pointer		const_pointer;
		typedef pointer 							iterator;
		typedef const_pointer						const_iterator;

		size_type				size () 	{ return fNumElts; }

		iterator				begin ()	{ return &fData[0]; }
		iterator				end ()		{ return &fData[fNumElts]; }

		void					insert (iterator position,
										const_iterator first,
										const_iterator last);
		void					erase (iterator first, iterator last);

	private:
		// I'd use a vector here instead, but that won't allocate the memory
		// for its buffer until it's constructed.  But by that time, we'll
		// probably be in the caller's context.

		allocator_type			fAllocator;
		T						fData[kMaxElts];
		size_type				fNumElts;
};


template <typename T, typename Allocator>
void fixed_size_deque<T, Allocator>::insert (	iterator position,
												const_iterator first,
												const_iterator last)
{
	if (first == last) return;

	long	numEltsToInsert = (long) distance (first, last);
	long	numTrailingEltsToMove = (long) distance (position, end ());
	
	EmAssert (position >= &fData[0]);
	EmAssert (position <= &fData[fNumElts]);
	EmAssert (fNumElts + numEltsToInsert < kMaxElts);

	/*
		Current array:
		
				xxxxxxxxxxxxxxxxxxxx............................
		
		After inserting yyyyy:

			Inserting near the beginning:
				xxxyyyyyxxxxxxxxxxxxxxxxx.......................

				# to move with copy = (end - position) - len(y) = (20 - 3) - 5 = 12
				# to move with uninitialized_copy = (end - position) - copy = 5

			Inserting close to the end:
				xxxxxxxxxxxxxyyyyyxxxxxxx.......................

				# to move with copy = (end - position) - len(y) = (20 - 13) - 5 = 2
				# to move with uninitialized_copy = (end - position) - copy = 5

			Inserting very close to the end:
				xxxxxxxxxxxxxxxxxxyyyyyxx.......................

				# to move with copy = (end - position) - len(y) = -3
				# to move with uninitialized_copy = (end - position) - copy = 5
				>>> clipping the first value to be non-negative gives us the desired results:
				# to move with copy = (end - position) - len(y) = 0
				# to move with uninitialized_copy = (end - position) - copy = 2
	*/

	long	numEltsToCopyIntoInitSpace = numTrailingEltsToMove - numEltsToInsert;
	if (numEltsToCopyIntoInitSpace < 0)
		numEltsToCopyIntoInitSpace = 0;

	long	numEltsToCopyIntoUninitSpace = numTrailingEltsToMove - numEltsToCopyIntoInitSpace;

	EmAssert (numEltsToCopyIntoUninitSpace >= 0);
	EmAssert (numEltsToCopyIntoUninitSpace <= fNumElts);

	EmAssert (numEltsToCopyIntoInitSpace >= 0);
	EmAssert (numEltsToCopyIntoInitSpace <= fNumElts);

	EmAssert (numEltsToCopyIntoInitSpace + numEltsToCopyIntoUninitSpace == numTrailingEltsToMove);

	// Copy n elements from their current location into No Man's land
	// past the last element.  Use uninitialized_copy () since the destination
	// doesn't contain any initialized elements.

	uninitialized_copy (end () - numEltsToCopyIntoUninitSpace, end (), end ());

	// Copy any additional elements down.  Use copy () since the destination
	// is full of element that we copied with the previous statement.

	copy_backward (position, position + numEltsToCopyIntoInitSpace, end ());

	// Copy in the elements from the user.

	copy (first, last, position);

	// Update our element count.

	fNumElts += numEltsToInsert;

	EmAssert (fNumElts <= kMaxElts);
	EmAssert (fNumElts >= 0);
}


template <typename T, typename Allocator>
void fixed_size_deque<T, Allocator>::erase (iterator first, iterator last)
{
	// Copy any trailing elements into their new
	// location over the elements being erased.
	iterator	i = copy (last, end (), first);
	
	// Destroy the elements now past the end of the
	// new end.
	while (i != end ())
		fAllocator.destroy (i++);

	// Update our element count.
	fNumElts -= (last - first);

	EmAssert (fNumElts <= kMaxElts);
	EmAssert (fNumElts >= 0);
}



class CPPCSocket : public CSocket
{
	public:
								CPPCSocket			(void);
		virtual 				~CPPCSocket 		(void);

		virtual ErrCode 		Open				(void);
		virtual ErrCode 		Close				(void);
		virtual ErrCode 		Write				(const void* buffer, long amountToWrite, long*);
		virtual ErrCode 		Read				(void* buffer, long sizeOfBuffer, long*);
		virtual Bool			HasUnreadData		(long timeout);
		virtual ErrCode 		Idle				(void);

		virtual Bool			ShortPacketHack 	(void);

	private:
		void					HandlePacket		(void);

	private:
		fixed_size_deque<unsigned char>  fBufferedData;
		Bool					fConnected;
		Bool					fShortPacketHack;
		deque<Handle>			fBufferedPackets;
};


// Non-constants.

#pragma mark Globals

static Bool 	gForceNewDirectory;
UInt32			gLastCycleTime;

#if _DEBUG
Bool			gExittedMain;
#endif

Boolean ShouldConvertHighAscii (void);
void ConvertHighAscii (char* d, const char* s);
void ConvertHighAscii (LStr255& str);
void ConvertHighAscii (string& str);

static void PrvFreeSoundChannel (void);

#include <ctype.h>

// Compare lexigraphically two strings

int _stricmp(const char *s1, const char *s2)
{
	char c1, c2;
	while (1)
	{
		c1 = tolower(*s1++);
		c2 = tolower(*s2++);
		if (c1 < c2) return -1;
		if (c1 > c2) return 1;
		if (c1 == 0) return 0;
	}
}


int _strnicmp(const char *str1, const char *str2, int n)
{
	const	unsigned char * p1 = (unsigned char *) str1 - 1;
	const	unsigned char * p2 = (unsigned char *) str2 - 1;
	unsigned long		c1, c2;

	n++;

	while (--n)
	{
		c1 = tolower (*++p1);
		c2 = tolower (*++p2);

		if (c1 != c2)
			return (c1 - c2);
		else if (!c1)
			break;
	}

	return 0;
}

char* _strdup (const char* s)
{
	char*	s2 = NULL;

	if (s)
	{
		s2 = (char*) malloc (strlen (s) + 1);
		if (s2)
		{
			strcpy (s2, s);
		}
	}

	return s2;
}


#pragma mark -

// DOLATER kwk - what you'd really like here is to have external mapping
// tables from various vendor character encodings to Unicode (e.g. PalmLatin,
// MacRoman, CP1252, PalmSJIS, etc). Then you'd build the conversion table
// from x -> y as needed, without bloating the size of Poser. Unfortunately
// the data is too big to just embed in Poser (several hundred K) and will
// continue to grow as new encodings are added to Palm devices (e.g. Korean).
// So for now we've got a limited set of tables that handle the common
// conversion cases, for a hard-coded set of encodings that we know about.
// If we wanted to be really cool, we'd use the ICU source, and define new
// tables for the Palm encodings - this would give us a nice, well-defined
// API for doing the encoding conversion.

// Or we could figure out how to provide Palm-specific mapping tables to the
// MacOS, and then use Unicode as the storage format on the scrap ('utxt'?).
// Though that might be MacOS 8.6 or later...would that be OK?

// This structure contains mapping information from one source character
// to one destination character. The source character is in the source
// encoding, and the destination character is in the destination encoding.
typedef struct
{
	uint16 srcChar;
	uint16 dstChar;
} CharMapType;

const uint16 kNoMap = 0xFFFF;

const ScrapFlavorType kScrapFlavorTypePalmText     = FOUR_CHAR_CODE('PLMT');       /* Palm text*/

// This structure contains information re how to map between two different
// character encodings. The entries in the .mappings array are assumed to
// be sorted by CharMapType.srcChar.
typedef struct
{
	uint8 srcEncoding;
	uint8 dstEncoding;
	int16 numMappings;			// Number of entries in .mappings array.
	const CharMapType* mappings;
} MapEncodingType;

// This table contains mappings from the latest Mac Roman encoding to
// Palm Latin.
static CharMapType kMacRomanToPalmLatinMapping[] =
{
	{ 0x000D, 0x000A }, // CARRIAGE RETURN (maps to LINEFEED)
	{ 0x0080, 0x00C4 },	// LATIN CAPITAL LETTER A WITH DIAERESIS
	{ 0x0081, 0x00C5 },	// LATIN CAPITAL LETTER A WITH RING ABOVE
	{ 0x0082, 0x00C7 },	// LATIN CAPITAL LETTER C WITH CEDILLA
	{ 0x0083, 0x00C9 },	// LATIN CAPITAL LETTER E WITH ACUTE
	{ 0x0084, 0x00D1 },	// LATIN CAPITAL LETTER N WITH TILDE
	{ 0x0085, 0x00D6 },	// LATIN CAPITAL LETTER O WITH DIAERESIS
	{ 0x0086, 0x00DC },	// LATIN CAPITAL LETTER U WITH DIAERESIS
	{ 0x0087, 0x00E1 },	// LATIN SMALL LETTER A WITH ACUTE
	{ 0x0088, 0x00E0 },	// LATIN SMALL LETTER A WITH GRAVE
	{ 0x0089, 0x00E2 },	// LATIN SMALL LETTER A WITH CIRCUMFLEX
	{ 0x008A, 0x00E4 },	// LATIN SMALL LETTER A WITH DIAERESIS
	{ 0x008B, 0x00E3 },	// LATIN SMALL LETTER A WITH TILDE
	{ 0x008C, 0x00E5 },	// LATIN SMALL LETTER A WITH RING ABOVE
	{ 0x008D, 0x00E7 },	// LATIN SMALL LETTER C WITH CEDILLA
	{ 0x008E, 0x00E9 },	// LATIN SMALL LETTER E WITH ACUTE
	{ 0x008F, 0x00E8 },	// LATIN SMALL LETTER E WITH GRAVE
	{ 0x0090, 0x00EA },	// LATIN SMALL LETTER E WITH CIRCUMFLEX
	{ 0x0091, 0x00EB },	// LATIN SMALL LETTER E WITH DIAERESIS
	{ 0x0092, 0x00ED },	// LATIN SMALL LETTER I WITH ACUTE
	{ 0x0093, 0x00EC },	// LATIN SMALL LETTER I WITH GRAVE
	{ 0x0094, 0x00EE },	// LATIN SMALL LETTER I WITH CIRCUMFLEX
	{ 0x0095, 0x00EF },	// LATIN SMALL LETTER I WITH DIAERESIS
	{ 0x0096, 0x00F1 },	// LATIN SMALL LETTER N WITH TILDE
	{ 0x0097, 0x00F3 },	// LATIN SMALL LETTER O WITH ACUTE
	{ 0x0098, 0x00F2 },	// LATIN SMALL LETTER O WITH GRAVE
	{ 0x0099, 0x00F4 },	// LATIN SMALL LETTER O WITH CIRCUMFLEX
	{ 0x009A, 0x00F6 },	// LATIN SMALL LETTER O WITH DIAERESIS
	{ 0x009B, 0x00F5 },	// LATIN SMALL LETTER O WITH TILDE
	{ 0x009C, 0x00FA },	// LATIN SMALL LETTER U WITH ACUTE
	{ 0x009D, 0x00F9 },	// LATIN SMALL LETTER U WITH GRAVE
	{ 0x009E, 0x00FB },	// LATIN SMALL LETTER U WITH CIRCUMFLEX
	{ 0x009F, 0x00FC },	// LATIN SMALL LETTER U WITH DIAERESIS
	{ 0x00A0, 0x0086 },	// DAGGER
	{ 0x00A1, 0x00B0 },	// DEGREE SIGN
	{ 0x00A4, 0x00A7 },	// SECTION SIGN
	{ 0x00A5, 0x0095 },	// BULLET
	{ 0x00A6, 0x00B6 },	// PILCROW SIGN
	{ 0x00A7, 0x00DF },	// LATIN SMALL LETTER SHARP S
	{ 0x00A8, 0x00AE },	// REGISTERED SIGN
	{ 0x00AA, 0x0099 },	// TRADE MARK SIGN
	{ 0x00AB, 0x00B4 },	// ACUTE ACCENT
	{ 0x00AC, 0x00A8 },	// DIAERESIS
	{ 0x00AD, kNoMap },	// NOT EQUAL TO
	{ 0x00AE, 0x00C6 },	// LATIN CAPITAL LETTER AE
	{ 0x00AF, 0x00D8 },	// LATIN CAPITAL LETTER O WITH STROKE
	{ 0x00B0, kNoMap },	// INFINITY
	{ 0x00B2, kNoMap },	// LESS-THAN OR EQUAL TO
	{ 0x00B3, kNoMap },	// GREATER-THAN OR EQUAL TO
	{ 0x00B4, 0x00A5 },	// YEN SIGN
	{ 0x00B6, kNoMap },	// PARTIAL DIFFERENTIAL
	{ 0x00B7, kNoMap },	// N-ARY SUMMATION
	{ 0x00B8, kNoMap },	// N-ARY PRODUCT
	{ 0x00B9, kNoMap },	// GREEK SMALL LETTER PI
	{ 0x00BA, kNoMap },	// INTEGRAL
	{ 0x00BB, 0x00AA },	// FEMININE ORDINAL INDICATOR
	{ 0x00BC, 0x00BA },	// MASCULINE ORDINAL INDICATOR
	{ 0x00BD, kNoMap },	// GREEK CAPITAL LETTER OMEGA
	{ 0x00BE, 0x00E6 },	// LATIN SMALL LETTER AE
	{ 0x00BF, 0x00F8 },	// LATIN SMALL LETTER O WITH STROKE
	{ 0x00C0, 0x00BF },	// INVERTED QUESTION MARK
	{ 0x00C1, 0x00A1 },	// INVERTED EXCLAMATION MARK
	{ 0x00C2, 0x00AC },	// NOT SIGN
	{ 0x00C3, kNoMap },	// SQUARE ROOT
	{ 0x00C4, 0x0083 },	// LATIN SMALL LETTER F WITH HOOK
	{ 0x00C5, kNoMap },	// ALMOST EQUAL TO
	{ 0x00C6, kNoMap },	// INCREMENT
	{ 0x00C7, 0x00AB },	// LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
	{ 0x00C8, 0x00BB },	// RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
	{ 0x00C9, 0x0085 },	// HORIZONTAL ELLIPSIS
	{ 0x00CA, 0x00A0 },	// NO-BREAK SPACE
	{ 0x00CB, 0x00C0 },	// LATIN CAPITAL LETTER A WITH GRAVE
	{ 0x00CC, 0x00C3 },	// LATIN CAPITAL LETTER A WITH TILDE
	{ 0x00CD, 0x00D5 },	// LATIN CAPITAL LETTER O WITH TILDE
	{ 0x00CE, 0x008C },	// LATIN CAPITAL LIGATURE OE
	{ 0x00CF, 0x009C },	// LATIN SMALL LIGATURE OE
	{ 0x00D0, 0x0096 },	// EN DASH
	{ 0x00D1, 0x0097 },	// EM DASH
	{ 0x00D2, 0x0093 },	// LEFT DOUBLE QUOTATION MARK
	{ 0x00D3, 0x0094 },	// RIGHT DOUBLE QUOTATION MARK
	{ 0x00D4, 0x0091 },	// LEFT SINGLE QUOTATION MARK
	{ 0x00D5, 0x0092 },	// RIGHT SINGLE QUOTATION MARK
	{ 0x00D6, 0x00F7 },	// DIVISION SIGN
	{ 0x00D7, kNoMap },	// LOZENGE
	{ 0x00D8, 0x00FF },	// LATIN SMALL LETTER Y WITH DIAERESIS
	{ 0x00D9, 0x009F },	// LATIN CAPITAL LETTER Y WITH DIAERESIS
	{ 0x00DA, kNoMap },	// FRACTION SLASH
	{ 0x00DB, 0x0080 },	// EURO SIGN
	{ 0x00DC, 0x008B },	// SINGLE LEFT-POINTING ANGLE QUOTATION MARK
	{ 0x00DD, 0x009B },	// SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
	{ 0x00DE, kNoMap },	// LATIN SMALL LIGATURE FI
	{ 0x00DF, kNoMap },	// LATIN SMALL LIGATURE FL
	{ 0x00E0, 0x0087 },	// DOUBLE DAGGER
	{ 0x00E1, 0x00B7 },	// MIDDLE DOT
	{ 0x00E2, 0x0082 },	// SINGLE LOW-9 QUOTATION MARK
	{ 0x00E3, 0x0084 },	// DOUBLE LOW-9 QUOTATION MARK
	{ 0x00E4, 0x0089 },	// PER MILLE SIGN
	{ 0x00E5, 0x00C2 },	// LATIN CAPITAL LETTER A WITH CIRCUMFLEX
	{ 0x00E6, 0x00CA },	// LATIN CAPITAL LETTER E WITH CIRCUMFLEX
	{ 0x00E7, 0x00C1 },	// LATIN CAPITAL LETTER A WITH ACUTE
	{ 0x00E8, 0x00CB },	// LATIN CAPITAL LETTER E WITH DIAERESIS
	{ 0x00E9, 0x00C8 },	// LATIN CAPITAL LETTER E WITH GRAVE
	{ 0x00EA, 0x00CD },	// LATIN CAPITAL LETTER I WITH ACUTE
	{ 0x00EB, 0x00CE },	// LATIN CAPITAL LETTER I WITH CIRCUMFLEX
	{ 0x00EC, 0x00CF },	// LATIN CAPITAL LETTER I WITH DIAERESIS
	{ 0x00ED, 0x00CC },	// LATIN CAPITAL LETTER I WITH GRAVE
	{ 0x00EE, 0x00D3 },	// LATIN CAPITAL LETTER O WITH ACUTE
	{ 0x00EF, 0x00D4 },	// LATIN CAPITAL LETTER O WITH CIRCUMFLEX
	{ 0x00F0, kNoMap },	// Apple logo
	{ 0x00F1, 0x00D2 },	// LATIN CAPITAL LETTER O WITH GRAVE
	{ 0x00F2, 0x00DA },	// LATIN CAPITAL LETTER U WITH ACUTE
	{ 0x00F3, 0x00DB },	// LATIN CAPITAL LETTER U WITH CIRCUMFLEX
	{ 0x00F4, 0x00D9 },	// LATIN CAPITAL LETTER U WITH GRAVE
	{ 0x00F5, kNoMap },	// LATIN SMALL LETTER DOTLESS I
	{ 0x00F6, 0x0088 },	// MODIFIER LETTER CIRCUMFLEX ACCENT
	{ 0x00F7, 0x0098 },	// SMALL TILDE
	{ 0x00F8, 0x00AF },	// MACRON
	{ 0x00F9, kNoMap },	// BREVE
	{ 0x00FA, kNoMap },	// DOT ABOVE
	{ 0x00FB, kNoMap },	// RING ABOVE
	{ 0x00FC, 0x00B8 },	// CEDILLA
	{ 0x00FD, kNoMap },	// DOUBLE ACUTE ACCENT
	{ 0x00FE, kNoMap },	// OGONEK
	{ 0x00FF, kNoMap }, // CARON
};

// This table contains mappings from the latest PalmLatin encoding to
// MacRoman (assumed 8.0 or later, with Euro support).
static CharMapType kPalmLatinToMacRomanMapping[] =
{
	{ 0x000A, 0x000D }, // LINEFEED (maps to CARRIAGE RETURN)
	{ 0x0080, 0x00DB },	// EURO SIGN
	{ 0x0081, kNoMap },	// UNDEFINED
	{ 0x0082, 0x00E2 },	// SINGLE LOW-9 QUOTATION MARK
	{ 0x0083, 0x00C4 },	// LATIN SMALL LETTER F WITH HOOK
	{ 0x0084, 0x00E3 },	// DOUBLE LOW-9 QUOTATION MARK
	{ 0x0085, 0x00C9 },	// HORIZONTAL ELLIPSIS
	{ 0x0086, 0x00A0 },	// DAGGER
	{ 0x0087, 0x00F6 },	// MODIFIER LETTER CIRCUMFLEX ACCENT
	{ 0x0089, 0x00E4 },	// PER MILLE SIGN
	{ 0x008A, kNoMap },	// LATIN CAPITAL LETTER S WITH CARON
	{ 0x008B, 0x00DC },	// SINGLE LEFT-POINTING ANGLE QUOTATION MARK
	{ 0x008C, 0x00CE },	// LATIN CAPITAL LIGATURE OE
	{ 0x008D, kNoMap },	// UNDEFINED
	{ 0x008E, kNoMap },	// LATIN CAPITAL LETTER Z WITH CARON
	{ 0x008F, kNoMap },	// UNDEFINED
	{ 0x0090, kNoMap },	// UNDEFINED
	{ 0x0091, 0x00D4 },	// LEFT SINGLE QUOTATION MARK
	{ 0x0092, 0x00D5 },	// RIGHT SINGLE QUOTATION MARK
	{ 0x0093, 0x00D2 },	// LEFT DOUBLE QUOTATION MARK
	{ 0x0094, 0x00D3 },	// RIGHT DOUBLE QUOTATION MARK
	{ 0x0095, 0x00A5 },	// BULLET
	{ 0x0096, 0x00D0 },	// EN DASH
	{ 0x0097, 0x00D1 },	// EM DASH
	{ 0x0098, 0x00F7 },	// SMALL TILDE
	{ 0x0099, 0x00AA },	// TRADE MARK SIGN
	{ 0x009A, kNoMap },	// LATIN SMALL LETTER S WITH CARON
	{ 0x009B, 0x00DD },	// SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
	{ 0x009C, 0x00CF },	// LATIN SMALL LIGATURE OE
	{ 0x009D, kNoMap },	// UNDEFINED
	{ 0x009E, 0x00D9 },	// LATIN CAPITAL LETTER Y WITH DIAERESIS
	{ 0x00A0, 0x00CA },	// NO-BREAK SPACE
	{ 0x00A1, 0x00C1 },	// INVERTED EXCLAMATION MARK
	{ 0x00A4, kNoMap },	// CURRENCY SIGN
	{ 0x00A5, 0x00B4 },	// YEN SIGN
	{ 0x00A6, kNoMap },	// BROKEN BAR
	{ 0x00A7, 0x00A4 },	// SECTION SIGN
	{ 0x00A8, 0x00AC },	// DIAERESIS
	{ 0x00AA, 0x00BB },	// FEMININE ORDINAL INDICATOR
	{ 0x00AB, 0x00C7 },	// LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
	{ 0x00AC, 0x00C2 },	// NOT SIGN
	{ 0x00AD, kNoMap },	// SOFT HYPHEN
	{ 0x00AE, 0x00A8 },	// REGISTERED SIGN
	{ 0x00AF, 0x00F8 },	// MACRON
	{ 0x00B0, 0x00A1 },	// DEGREE SIGN
	{ 0x00B2, kNoMap },	// SUPERSCRIPT TWO
	{ 0x00B3, kNoMap },	// SUPERSCRIPT THREE
	{ 0x00B4, 0x00AB },	// ACUTE ACCENT
	{ 0x00B6, 0x00A6 },	// PILCROW SIGN
	{ 0x00B7, 0x00E1 },	// MIDDLE DOT
	{ 0x00B8, 0x00FC },	// CEDILLA
	{ 0x00B9, kNoMap },	// SUPERSCRIPT ONE
	{ 0x00BA, 0x00BC },	// MASCULINE ORDINAL INDICATOR
	{ 0x00BB, 0x00C8 },	// RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
	{ 0x00BC, kNoMap },	// VULGAR FRACTION ONE QUARTER
	{ 0x00BD, kNoMap },	// VULGAR FRACTION ONE HALF
	{ 0x00BE, kNoMap },	// VULGAR FRACTION THREE QUARTERS
	{ 0x00BF, 0x00C0 },	// INVERTED QUESTION MARK
	{ 0x00C0, 0x00CB },	// LATIN CAPITAL LETTER A WITH GRAVE
	{ 0x00C1, 0x00E7 },	// LATIN CAPITAL LETTER A WITH ACUTE
	{ 0x00C2, 0x00E5 },	// LATIN CAPITAL LETTER A WITH CIRCUMFLEX
	{ 0x00C3, 0x00CC },	// LATIN CAPITAL LETTER A WITH TILDE
	{ 0x00C4, 0x0080 },	// LATIN CAPITAL LETTER A WITH DIAERESIS
	{ 0x00C5, 0x0081 },	// LATIN CAPITAL LETTER A WITH RING ABOVE
	{ 0x00C6, 0x00AE },	// LATIN CAPITAL LETTER AE
	{ 0x00C7, 0x0082 },	// LATIN CAPITAL LETTER C WITH CEDILLA
	{ 0x00C8, 0x00E9 },	// LATIN CAPITAL LETTER E WITH GRAVE
	{ 0x00C9, 0x0083 },	// LATIN CAPITAL LETTER E WITH ACUTE
	{ 0x00CA, 0x00E6 },	// LATIN CAPITAL LETTER E WITH CIRCUMFLEX
	{ 0x00CB, 0x00E8 },	// LATIN CAPITAL LETTER E WITH DIAERESIS
	{ 0x00CC, 0x00ED },	// LATIN CAPITAL LETTER I WITH GRAVE
	{ 0x00CD, 0x00EA },	// LATIN CAPITAL LETTER I WITH ACUTE
	{ 0x00CE, 0x00EB },	// LATIN CAPITAL LETTER I WITH CIRCUMFLEX
	{ 0x00CF, 0x00EC },	// LATIN CAPITAL LETTER I WITH DIAERESIS
	{ 0x00D0, kNoMap },	// LATIN CAPITAL LETTER ETH
	{ 0x00D1, 0x0084 },	// LATIN CAPITAL LETTER N WITH TILDE
	{ 0x00D2, 0x00F1 },	// LATIN CAPITAL LETTER O WITH GRAVE
	{ 0x00D3, 0x00EE },	// LATIN CAPITAL LETTER O WITH ACUTE
	{ 0x00D4, 0x00EF },	// LATIN CAPITAL LETTER O WITH CIRCUMFLEX
	{ 0x00D5, 0x00CD },	// LATIN CAPITAL LETTER O WITH TILDE
	{ 0x00D6, 0x0085 },	// LATIN CAPITAL LETTER O WITH DIAERESIS
	{ 0x00D7, kNoMap },	// MULTIPLICATION SIGN
	{ 0x00D8, 0x00AF },	// LATIN CAPITAL LETTER O WITH STROKE
	{ 0x00D9, 0x00F4 },	// LATIN CAPITAL LETTER U WITH GRAVE
	{ 0x00DA, 0x00F2 },	// LATIN CAPITAL LETTER U WITH ACUTE
	{ 0x00DB, 0x00F3 },	// LATIN CAPITAL LETTER U WITH CIRCUMFLEX
	{ 0x00DC, 0x0086 },	// LATIN CAPITAL LETTER U WITH DIAERESIS
	{ 0x00DD, kNoMap },	// LATIN CAPITAL LETTER Y WITH ACUTE
	{ 0x00DE, kNoMap },	// LATIN CAPITAL LETTER THORN
	{ 0x00DF, 0x00A7 },	// LATIN SMALL LETTER SHARP S
	{ 0x00E0, 0x0088 },	// LATIN SMALL LETTER A WITH GRAVE
	{ 0x00E1, 0x0087 },	// LATIN SMALL LETTER A WITH ACUTE
	{ 0x00E2, 0x0089 },	// LATIN SMALL LETTER A WITH CIRCUMFLEX
	{ 0x00E3, 0x008B },	// LATIN SMALL LETTER A WITH TILDE
	{ 0x00E4, 0x008A },	// LATIN SMALL LETTER A WITH DIAERESIS
	{ 0x00E5, 0x008C },	// LATIN SMALL LETTER A WITH RING ABOVE
	{ 0x00E6, 0x00BE },	// LATIN SMALL LETTER AE
	{ 0x00E7, 0x008D },	// LATIN SMALL LETTER C WITH CEDILLA
	{ 0x00E8, 0x008F },	// LATIN SMALL LETTER E WITH GRAVE
	{ 0x00E9, 0x008E },	// LATIN SMALL LETTER E WITH ACUTE
	{ 0x00EA, 0x0090 },	// LATIN SMALL LETTER E WITH CIRCUMFLEX
	{ 0x00EB, 0x0091 },	// LATIN SMALL LETTER E WITH DIAERESIS
	{ 0x00EC, 0x0093 },	// LATIN SMALL LETTER I WITH GRAVE
	{ 0x00ED, 0x0092 },	// LATIN SMALL LETTER I WITH ACUTE
	{ 0x00EE, 0x0094 },	// LATIN SMALL LETTER I WITH CIRCUMFLEX
	{ 0x00EF, 0x0095 },	// LATIN SMALL LETTER I WITH DIAERESIS
	{ 0x00F0, kNoMap },	// LATIN SMALL LETTER ETH
	{ 0x00F1, 0x0096 },	// LATIN SMALL LETTER N WITH TILDE
	{ 0x00F2, 0x0098 },	// LATIN SMALL LETTER O WITH GRAVE
	{ 0x00F3, 0x0097 },	// LATIN SMALL LETTER O WITH ACUTE
	{ 0x00F4, 0x0099 },	// LATIN SMALL LETTER O WITH CIRCUMFLEX
	{ 0x00F5, 0x009B },	// LATIN SMALL LETTER O WITH TILDE
	{ 0x00F6, 0x009A },	// LATIN SMALL LETTER O WITH DIAERESIS
	{ 0x00F7, 0x00D6 },	// DIVISION SIGN
	{ 0x00F8, 0x00BF },	// LATIN SMALL LETTER O WITH STROKE
	{ 0x00F9, 0x009D },	// LATIN SMALL LETTER U WITH GRAVE
	{ 0x00FA, 0x009C },	// LATIN SMALL LETTER U WITH ACUTE
	{ 0x00FB, 0x009E },	// LATIN SMALL LETTER U WITH CIRCUMFLEX
	{ 0x00FC, 0x009F },	// LATIN SMALL LETTER U WITH DIAERESIS
	{ 0x00FD, kNoMap },	// LATIN SMALL LETTER Y WITH ACUTE
	{ 0x00FE, kNoMap },	// LATIN SMALL LETTER THORN
	{ 0x00FF, 0x00D8 },	// LATIN SMALL LETTER Y WITH DIAERESIS
};

// This table contains mappings from the Mac Japanese encoding to
// Palm Shift-JIS.
static CharMapType kMacSJISToPalmSJISMapping[] =
{
	{ 0x000D, 0x000A }, // CARRIAGE RETURN (maps to LINEFEED)

	// Special single-byte characters.
	// { 0x0080, 0x815F }, // REVERSE SOLIDUS (map to full-width version)
	{ 0x0080, kNoMap }, // REVERSE SOLIDUS
	{ 0x00A0, kNoMap }, // NO-BREAK SPACE
	{ 0x00FD, kNoMap }, // COPYRIGHT SIGN
	{ 0x00FE, kNoMap }, // TRADE MARK SIGN
	{ 0x00FF, 0x0018 }, // HORIZONTAL ELLIPSIS

	// Here is where you'd like to put a bunch of the Apple vertical
	// form characters which get mapped to nothing. I'll put in a few
	// as examples.
	{ 0xEB41, kNoMap }, // vertical form for IDEOGRAPHIC COMMA
	{ 0xEB42, kNoMap }, // vertical form for IDEOGRAPHIC FULL STOP
};

// This table contains mappings from the Palm Shift-JIS encoding to
// Mac Japanese.
static CharMapType kPalmSJISToMacSJISMapping[] =
{
	{ 0x000A, 0x000D }, // LINEFEED (maps to CARRIAGE RETURN)

	// Special single-byte characters.
	{ 0x0014, kNoMap },	// chrOtaSecure
	{ 0x0015, kNoMap },	// chrOta
	{ 0x0016, kNoMap },	// chrCommandStroke
	{ 0x0017, kNoMap },	// chrShortcutStroke
	{ 0x0018, 0x00FF },	// HORIZONTAL ELLIPSIS
	{ 0x0019, kNoMap },	// NUMERIC SPACE
	
	// Here is where you'd like to put a bunch of the Microsoft extension
	// characters that get mapped to nothing. I'll put in a few
	// as examples. Some of these are also duplicates of Han at other code
	// points, so they might wind up getting mapped to another existing code
	// point.
	{ 0xED40, kNoMap }, // 
	{ 0xED41, kNoMap }, // 
};

// This table contains mappings from the Mac Traditional Chinese encoding to
// Palm Big-5.
static CharMapType kMacTradChineseToPalmBig5Mapping[] =
{
	{ 0x000D, 0x000A }, // CARRIAGE RETURN (maps to LINEFEED)

	// Special single-byte characters.
	// { 0x0080, 0x815F }, // REVERSE SOLIDUS (map to full-width version)
	{ 0x0080, kNoMap }, // REVERSE SOLIDUS
	{ 0x00A0, kNoMap }, // NO-BREAK SPACE
	{ 0x00FD, kNoMap }, // COPYRIGHT SIGN
	{ 0x00FE, kNoMap }, // TRADE MARK SIGN
	{ 0x00FF, 0x0018 }, // HORIZONTAL ELLIPSIS
};

// This table contains mappings from the Palm Big-5 encoding to
// Mac Traditional Chinese.
static CharMapType kPalmBig5ToMacTradChineseMapping[] =
{
	{ 0x000A, 0x000D }, // LINEFEED (maps to CARRIAGE RETURN)

	// Special single-byte characters.
	{ 0x0014, kNoMap },	// chrOtaSecure
	{ 0x0015, kNoMap },	// chrOta
	{ 0x0016, kNoMap },	// chrCommandStroke
	{ 0x0017, kNoMap },	// chrShortcutStroke
	{ 0x0018, 0x00FF },	// HORIZONTAL ELLIPSIS
	{ 0x0019, kNoMap },	// NUMERIC SPACE
};

// This table contains mappings from the Mac Simplified Chinese encoding to
// Palm GB.
static CharMapType kMacSimpChineseToPalmGBMapping[] =
{
	{ 0x000D, 0x000A }, // CARRIAGE RETURN (maps to LINEFEED)

	// Special single-byte characters.
	{ 0x0080, kNoMap }, // LATIN SMALL LETTER U WITH DIAERESIS
	{ 0x00A0, kNoMap }, // NO-BREAK SPACE
	{ 0x00FD, kNoMap }, // COPYRIGHT SIGN
	{ 0x00FE, kNoMap }, // TRADE MARK SIGN
	{ 0x00FF, 0x0018 }, // HORIZONTAL ELLIPSIS
};

// This table contains mappings from the Palm GB encoding to
// Mac Simplified Chinese.
static CharMapType kPalmGBToMacSimpChineseMapping[] =
{
	{ 0x000A, 0x000D }, // LINEFEED (maps to CARRIAGE RETURN)

	// Special single-byte characters.
	{ 0x0014, kNoMap },	// chrOtaSecure
	{ 0x0015, kNoMap },	// chrOta
	{ 0x0016, kNoMap },	// chrCommandStroke
	{ 0x0017, kNoMap },	// chrShortcutStroke
	{ 0x0018, 0x00FF },	// HORIZONTAL ELLIPSIS
	{ 0x0019, kNoMap },	// NUMERIC SPACE
};

// Table that contains mapping information between pairs of encodings.
// Note that there are different tables that go both ways (Palm to Mac
// and Mac to Palm)
const MapEncodingType kPalmMacEncodingMaps[] = 
{
	// Mac -> Palm tables
	{ smRoman, charEncodingPalmLatin, countof (kMacRomanToPalmLatinMapping), kMacRomanToPalmLatinMapping },
	{ smJapanese, charEncodingPalmSJIS, countof (kMacSJISToPalmSJISMapping), kMacSJISToPalmSJISMapping },
	
	{ smTradChinese, charEncodingBig5, countof (kMacTradChineseToPalmBig5Mapping), kMacTradChineseToPalmBig5Mapping },
	{ smTradChinese, charEncodingBig5_HKSCS, countof (kMacTradChineseToPalmBig5Mapping), kMacTradChineseToPalmBig5Mapping },
	{ smTradChinese, charEncodingBig5Plus, countof (kMacTradChineseToPalmBig5Mapping), kMacTradChineseToPalmBig5Mapping },
	{ smTradChinese, charEncodingPalmBig5, countof (kMacTradChineseToPalmBig5Mapping), kMacTradChineseToPalmBig5Mapping },
	
	{ smSimpChinese, charEncodingGB2312, countof (kMacSimpChineseToPalmGBMapping), kMacSimpChineseToPalmGBMapping },
	{ smSimpChinese, charEncodingGBK, countof (kMacSimpChineseToPalmGBMapping), kMacSimpChineseToPalmGBMapping },
	{ smSimpChinese, charEncodingPalmGB, countof (kMacSimpChineseToPalmGBMapping), kMacSimpChineseToPalmGBMapping },
	
	// Palm -> Mac tables
	{ charEncodingPalmLatin, smRoman, countof (kPalmLatinToMacRomanMapping), kPalmLatinToMacRomanMapping },
	{ charEncodingPalmSJIS, smJapanese, countof (kPalmSJISToMacSJISMapping), kPalmSJISToMacSJISMapping },
	
	// We're not sure which Traditional Chinese encoding will get used during development, so provide mappings
	// for all of the possibilities.
	{ charEncodingBig5, smTradChinese, countof (kPalmBig5ToMacTradChineseMapping), kPalmBig5ToMacTradChineseMapping },
	{ charEncodingBig5_HKSCS, smTradChinese, countof (kPalmBig5ToMacTradChineseMapping), kPalmBig5ToMacTradChineseMapping },
	{ charEncodingBig5Plus, smTradChinese, countof (kPalmBig5ToMacTradChineseMapping), kPalmBig5ToMacTradChineseMapping },
	{ charEncodingPalmBig5, smTradChinese, countof (kPalmBig5ToMacTradChineseMapping), kPalmBig5ToMacTradChineseMapping },

	// We're not sure which Simplified Chinese encoding will get used during development, so provide mappings
	// for all of the possibilities.
	{ charEncodingGB2312, smSimpChinese, countof (kPalmGBToMacSimpChineseMapping), kPalmGBToMacSimpChineseMapping },
	{ charEncodingGBK, smSimpChinese, countof (kPalmGBToMacSimpChineseMapping), kPalmGBToMacSimpChineseMapping },
	{ charEncodingPalmGB, smSimpChinese, countof (kPalmGBToMacSimpChineseMapping), kPalmGBToMacSimpChineseMapping }
};


/***********************************************************************
 * CompareCharToMap
 *
 * Compare a character to a character contained in a CharMapType record,
 * and return the results ala strcmp.
 ***********************************************************************/

static int
CompareCharToMap(const void *keyval, const void *datum)
{
	uint16 searchChar = *(uint16*)keyval;
	const CharMapType* mapEntry = (const CharMapType*)datum;
	
	if (searchChar < mapEntry->srcChar)
	{
		return(-1);
	}
	else if (searchChar > mapEntry->srcChar)
	{
		return(1);
	}
	else
	{
		return(0);
	}
} // CompareCharToMap


/***********************************************************************
 * GetCharMap
 *
 * Return back a table that's appropriate for converting text from
 * <iSrcEncoding> to <iDstEncoding>. Also pass back the # of mapping
 * entries from the table in <oNumMappings>.
 ***********************************************************************/

static const CharMapType*
GetCharMap(uint8 iSrcEncoding, uint8 iDstEncoding, int16& oNumMappings)
{
	int16 numTables = countof (kPalmMacEncodingMaps);
	for (int16 i = 0; i < numTables; i++)
	{
		if ((kPalmMacEncodingMaps[i].srcEncoding == iSrcEncoding)
		 && (kPalmMacEncodingMaps[i].dstEncoding == iDstEncoding))
		{
			oNumMappings = kPalmMacEncodingMaps[i].numMappings;
			return (kPalmMacEncodingMaps[i].mappings);
		}
	}
	
	oNumMappings = 0;
	return (NULL);
} // GetCharMap

static int	gIntlMgrExists = -1;

static Bool IntlMgrExists (void)
{
	if (gIntlMgrExists < 0)
	{
		// Note that we need to check by calling the feature manager rather than
		// checking to see if the trap is implemented. sysTrapIntlDispatch is
		// sysTrapPsrInit on 1.0 systems and sysUnused2 on intermediate systems.
		// That means that the trap IS implemented, but just not the one we want.

		UInt32	data;
		Err		err = FtrGet (sysFtrCreator, sysFtrNumIntlMgr, &data);

		gIntlMgrExists = !err && (data & intlMgrExists) != 0;
	}

	return gIntlMgrExists != 0;
}


static UInt16 _TxtGetNextChar (const Char *inText, UInt32 inOffset, WChar *outChar)
{
	if (IntlMgrExists ())
	{
		return TxtGetNextChar (inText, inOffset, outChar);
	}

	if (outChar)
		*outChar = (UInt8) inText[inOffset];

	return sizeof (Char);
}


const uint8	kUndefinedChar = '?';

// Table for mapping from MacRoman high ASCII (0x80..0xFF) to low ASCII.
static const char* kHighToLowAscii[128] = 
{
	"A" /*  */,	"A" /*  */,	"C" /*  */,	"E" /*  */,
	"N" /*  */,	"O" /*  */,	"U" /*  */,	"a" /*  */,
	"a" /*  */,	"a" /*  */,	"a" /*  */,	"a" /*  */,
	"a" /*  */,	"c" /*  */,	"e" /*  */,	"e" /*  */,

	"e" /*  */,	"e" /*  */,	"i" /*  */,	"i" /*  */,
	"i" /*  */,	"i" /*  */,	"n" /*  */,	"o" /*  */,
	"o" /*  */,	"o" /*  */,	"o" /*  */,	"o" /*  */,
	"u" /*  */,	"u" /*  */,	"u" /*  */,	"u" /*  */,

	"?" /*  */,	"?" /*  */,	"?" /*  */,	"?" /*  */,
	"?" /*  */,	"*" /*  */,	"?" /*  */,	"?" /*  */,
	"(r)" /*  */,	"(c)" /*  */,	"(tm)" /*  */,	"'" /*  */,
	"?" /*  */,	"<>" /*  */,	"AE" /*  */,	"?" /*  */,

	"?" /*  */,	"+/-" /*  */,	"<=" /*  */,	">=" /*  */,
	"?" /*  */,	"?" /*  */,	"?" /*  */,	"?" /*  */,
	"?" /*  */,	"?" /*  */,	"?" /*  */,	"?" /*  */,
	"?" /*  */,	"?" /*  */,	"ae" /*  */,	"?" /*  */,

	"?" /*  */,	"?" /*  */,	"?" /*  */,	"?" /*  */,
	"?" /*  */,	"?" /*  */,	"?" /*  */,	"<<" /*  */,
	">>" /*  */,	"..." /*  */,	" " /*  */,	"A" /*  */,
	"A" /*  */,	"O" /*  */,	"OE" /*  */,	"oe" /*  */,

	"-" /*  */,	"-" /*  */,	"\"" /*  */,	"\"" /*  */,
	"'" /*  */,	"'" /*  */,	"/" /*  */,	"?" /*  */,
	"y" /*  */,	"Y" /*  */,	"/" /*  */,	"?" /*  */,
	"<" /*  */,	">" /*  */,	"fi" /*  */,	"fl" /*  */,

	"?" /*  */,	"?" /*  */,	"?" /*  */,	"?" /*  */,
	"%" /*  */,	"A" /*  */,	"E" /*  */,	"A" /*  */,
	"E" /*  */,	"E" /*  */,	"I" /*  */,	"I" /*  */,
	"I" /*  */,	"I" /*  */,	"O" /*  */,	"O" /*  */,

	"?" /*  */,	"O" /*  */,	"U" /*  */,	"U" /*  */,
	"U" /*  */,	"?" /*  */,	"^" /*  */,	"~" /*  */,
	"?" /*  */,	"?" /*  */,	"?" /*  */,	"?" /*  */,
	"?" /*  */,	"?" /*  */,	"?" /*  */,	"?" /*  */
};

// Return true if Poser is running with a character encoding other
// than MacRoman. In that case we can really only trust low-ASCII
// text in alert messages.
Boolean ShouldConvertHighAscii (void)
{
	// When running w/a language kit, the system/app fonts are
	// dynamically changed when an app becomes active, so that
	// you can register an app as "Japanese" for example and it
	// gets Japanese menus and such even when the system is Roman.
	// So the right way to check is to determine the script of
	// the system font (which also gets swapped w/the active app).
	return (::FontToScript(::GetSysFont()) != smRoman);
}


void ConvertHighAscii (char* d, const char* s)
{
	unsigned char	ch;

	while ((ch = (unsigned char) *s++) != 0)
	{
		if (ch >= 0x080)
		{
			strcpy (d, kHighToLowAscii[ch - 0x080]);
			d += strlen(kHighToLowAscii[ch - 0x080]);
		}
		else
		{
			*d++ = (char) ch;
		}
	}

	*d = 0;
}


void ConvertHighAscii (LStr255& str)
{
	LStr255	result;

	// Terminate the incoming string.

	str[str[0] + 1] = 0;

	// Convert it using the version that takes C string.

	char*	s = (char*) &str[1];
	char*	d = (char*) &result[1];
	
	::ConvertHighAscii (d, s);

	// Set the length of the result.

	result[0] = strlen (d);

	// Copy the result back into the return parameter.

	str = result;
}

void ConvertHighAscii (string& str)
{
	string::iterator	iter = str.begin();
	while (iter != str.end())
	{
		unsigned char	ch = *iter;
		if (ch >= 0x080)
		{
			string::size_type	offset = iter - str.begin();
			str.replace (offset, 1, kHighToLowAscii[ch - 0x080]);
			iter = str.begin() + offset;
		}
		else
		{
			++iter;
		}
	}
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 Platform::Initialize
// ---------------------------------------------------------------------------
// Initializes platform-dependent stuff.

void Platform::Initialize (void)
{
}


// ---------------------------------------------------------------------------
//		 Platform::Reset
// ---------------------------------------------------------------------------
// Resets platform-dependent stuff.

void Platform::Reset (void)
{
	PrvFreeSoundChannel ();
}


// ---------------------------------------------------------------------------
//		 Platform::Save
// ---------------------------------------------------------------------------

void Platform::Save (SessionFile&)
{
}


// ---------------------------------------------------------------------------
//		 Platform::Load
// ---------------------------------------------------------------------------

void Platform::Load (SessionFile&)
{
}


// ---------------------------------------------------------------------------
//		 Platform::Dispose
// ---------------------------------------------------------------------------

void Platform::Dispose (void)
{
	PrvFreeSoundChannel ();
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 Platform::GetString
// ---------------------------------------------------------------------------

string Platform::GetString (StrCode id)
{
#if 1
	// Now use the same method on the Mac as we do on Unix.  That is, get the
	// text from a string table compiled into the application instead of from
	// resources.  The problem with a resource-based approach is that we can't
	// use them to generate error messages when in the context of handling
	// a PPC Toolbox message from a debugger.  In that context, the caller's
	// resource chain is in effect, not our own, and so we don't have access
	// to our error messages in the resource.

	const char* _ResGetString(int idx);

	const char* str = _ResGetString (id);
	if (str)
	{
		string result (str);
		if (::ShouldConvertHighAscii ())
		{
			::ConvertHighAscii (result);
		}

		return result;
	}

	char	buffer[20];
	sprintf (buffer, "%ld", (long) id);
	return string ("<missing string ") + buffer + ">";

#else
	Handle			strHandle = GetResource ('STRA', 1000);
	StHandleLocker	locker (strHandle);
	Ptr 			strPtr = *strHandle;

	short	numStrings = *(short*) strPtr;
	strPtr += sizeof (short);

	for (short ii = 0; ii < numStrings; ++ii)
	{
		short	strID = *(short*) strPtr;
		strPtr += sizeof (short);

		short	strLen = *(short*) strPtr;
		strPtr += sizeof (short);

		if (strID == id)
		{
			string	str ((const char*) strPtr, strLen);
			if (::ShouldConvertHighAscii())
				::ConvertHighAscii(str);
			return str;
		}

		strPtr += strLen;
	}

	char	buffer[20];
	sprintf (buffer, "%d", id);
	return string ("<missing string ") + buffer + ">";
#endif
}


// ---------------------------------------------------------------------------
//		 Platform::GetIDForError
// ---------------------------------------------------------------------------

int Platform::GetIDForError (ErrCode error)
{
	switch (error)
	{
		case userCanceledErr:		return kStr_UserCancel;
		case kernelTimeoutErr:		return kStr_TimeoutError;
		case dirFulErr: 			return kStr_DirectoryFull;
		case dskFulErr: 			return kStr_DiskFull;
		case nsvErr:				return kStr_DiskMissing;
		case volOffLinErr:			return kStr_DiskMissing;
		case volGoneErr:			return kStr_DiskMissing;
		case ioErr: 				return kStr_IOError;
		case bdNamErr:				return kStr_BadFileName;
		case tmfoErr:				return kStr_TooManyFilesOpen;
		case fnfErr:				return kStr_FileNotFound;
		case wPrErr:				return kStr_DiskWriteProtected;
		case fLckdErr:				return kStr_FileLocked;
		case permErr:				return kStr_FileLocked;
		case vLckdErr:				return kStr_DiskLocked;
		case fBsyErr:				return kStr_FileBusy;
		case dupFNErr:				return kStr_DuplicateFileName;
		case memFullErr:			return kStr_MemFull;
	}

	return kStr_GenericError;
}


// ---------------------------------------------------------------------------
//		 Platform::GetIDForRecovery
// ---------------------------------------------------------------------------

int Platform::GetIDForRecovery (ErrCode error)
{
	switch (error)
	{
		case dirFulErr: 			return kStr_DeleteSomeFiles;
		case dskFulErr: 			return kStr_DeleteSomeFiles;
		case wPrErr:				return kStr_UnlockTheDisk;
		case fLckdErr:				return kStr_UnlockTheFile;
		case permErr:				return kStr_UnlockTheFile;
		case vLckdErr:				return kStr_UnlockTheDisk;
		case memFullErr:			return kStr_NeedMoreRAM;
	}

	return 0;
}


// ---------------------------------------------------------------------------
//		 Platform::GetShortVersionString
// ---------------------------------------------------------------------------
// Returns a short version string.	The format of the string is:
//
//		#.# (.#) ([dab]#)
//
//		# = one or more numeric digits
//		. = literal '.'
//		Patterns in parentheses are optional
//		Patterns in brackets mean "one of these characters"
//		Spaces are shown above for clarity; they do not appear in the string
//
// Valid strings would be: 2.1d7, 2.1.1b14, 2.99, 2.1.1

string Platform::GetShortVersionString (void)
{
	// Get the version resource.

	StResource		versionResource ('vers', 1);		
	VersRecHndl 	versHandle = (VersRecHndl) (Handle) versionResource;
	StHandleLocker	lock ((Handle) versHandle);

	// If we've got one, get the long version string and cut it
	// down to size.  I'd use the short version string, but it
	// doesn't allow for internal release numbers (e.g., 3.0a1e8).
	// So I get the long version string and take everything up
	// to the first comma.

	string	version;
	if (versHandle)
	{
		VersRecPtr	versptr			= *versHandle;

		uint8*		shortVersPtr	= &versptr->shortVersion[0];
		uint8		shortVersLength = shortVersPtr[0];

		uint8*		longVersPtr		= &shortVersPtr[shortVersLength + 1];
		uint8		longVersLength	= longVersPtr[0];

		version = string ((char*) (longVersPtr + 1), longVersLength);

		string::size_type	commaPos = version.find (',');
		if (commaPos != string::npos)
		{
			version = version.substr (0, commaPos);
		}
	}

	return version;
}

#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	Platform::CopyToClipboard
 *
 * DESCRIPTION:	Copy data from Palm clipboard to Mac clipboard,
 *	transliterating text data as needed.
 *
 * PARAMETERS:
 *	data	 ->	Pointer to the Palm clipboard data.
 *	length	 ->	Bytes of data in <data>
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void Platform::CopyToClipboard (const ByteList& palmChars,
								const ByteList& hostChars)
{
	ByteList	palmChars2 (palmChars);
	ByteList	hostChars2 (hostChars);

	// See if any mapping needs to be done.

	if (hostChars2.size () > 0 && palmChars2.size () == 0)
	{
		Platform::RemapHostToPalmChars (hostChars2, palmChars2);
	}
	else if (palmChars2.size () > 0 && hostChars2.size () == 0)
	{
		Platform::RemapPalmToHostChars (palmChars2, hostChars2);
	}

	// Add both kinds of data to the scrap.  Note that the order appears
	// to be important here.  Reversing the order seems to confuse some
	// applications, and they fail to see that text is on the clipboard.

	try
	{
		UScrap::SetData (kScrapFlavorTypeText, &hostChars2[0], hostChars2.size (), true);
		UScrap::SetData (kScrapFlavorTypePalmText, &palmChars2[0], palmChars2.size (), false);
	}
	catch (...)
	{
	}
}


/***********************************************************************
 *
 * FUNCTION:	Platform::CopyFromClipboard
 *
 * DESCRIPTION:	Load data from the Mac clipboard, transliterate it into
 *	the device's character encoding, and return the resulting text in
 *	<data>.
 *
 * PARAMETERS:	Ref to where to put converted data.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void Platform::CopyFromClipboard (	ByteList& palmChars,
									ByteList& hostChars)
{
	// Assume failure.

	palmChars.clear ();
	hostChars.clear ();

	StHandleBlock	ioData (::NewHandle (0));
	SInt32			byteCount;

	// Get the clipboard data. First we'll try for the special Palm text.

	try
	{
		byteCount = UScrap::GetData (kScrapFlavorTypePalmText, ioData);

		if (byteCount > 0)
		{
			StHandleLocker	lock (ioData);
			Ptr				p = *ioData.Get ();

			copy (p, p + byteCount, back_inserter (palmChars));
		}
	}
	catch (...)
	{
	}

	// Next, try for the host-specific text.

	try
	{
		byteCount = UScrap::GetData (kScrapFlavorTypeText, ioData);

		if (byteCount > 0)
		{
			StHandleLocker	lock (ioData);
			Ptr				p = *ioData.Get ();

			copy (p, p + byteCount, back_inserter (hostChars));
		}
	}
	catch (...)
	{
	}

	// See if any mapping needs to be done.

	if (hostChars.size () > 0 && palmChars.size () == 0)
	{
		Platform::RemapHostToPalmChars (hostChars, palmChars);
	}
	else if (palmChars.size () > 0 && hostChars.size () == 0)
	{
		Platform::RemapPalmToHostChars (palmChars, hostChars);
	}
}


/***********************************************************************
 *
 * FUNCTION:	Platform::RemapHostChar
 *
 * DESCRIPTION:	Remap character <hostChar> from <hostEncoding> to
 *				<deviceEncoding>
 *
 * PARAMETERS:
 *	hostEncoding	 ->	Host platform character encoding.
 *	deviceEncoding	 ->	Palm encoding for ROM being emulated.
 *	hostChar		 ->	Character code.
 *
 * RETURNED:	Remapped character, or 0 if it couldn't be remapped.
 *
 ***********************************************************************/

uint16 Platform::RemapHostChar (uint8 hostEncoding,
								uint8 deviceEncoding,
								uint16 hostChar)
{
	int16 numEntries;
	const CharMapType* charMapP = ::GetCharMap (hostEncoding,
												deviceEncoding,
												numEntries);

	// Map the character. 

	if (charMapP != NULL)
	{
		void* result = bsearch (&hostChar, charMapP, numEntries,
								sizeof(CharMapType), ::CompareCharToMap);

		if (result != NULL)
		{
			const CharMapType* theMap = (const CharMapType*) result;
			hostChar = theMap->dstChar;
		}
		else
		{
			// No remapping found, so assume it's OK.
		}
	}
	else
	{
		if (hostChar == 0x000D)
		{
			hostChar = chrLineFeed;
		}
		else if (hostChar >= 0x0080)
		{
			hostChar = 0x0000;
		}
	}

	return hostChar;
}


/***********************************************************************
 *
 * FUNCTION:	Platform::RemapPalmToHostChars
 *
 * DESCRIPTION:	Load data from the Mac clipboard, transliterate it into
 *	the device's character encoding, and return the resulting text in
 *	<data>.
 *
 * PARAMETERS:	Ref to where to put converted data.
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void Platform::RemapPalmToHostChars (	const ByteList& palmChars,
										ByteList& hostChars)
{
	// Clear the output vector.

	hostChars.clear ();

	// Get the source and dest encoding.

	uint16	macEncoding = ::FontToScript (::GetSysFont ());
	uint32	deviceEncoding;

	if (::FtrGet (sysFtrCreator, sysFtrNumEncoding, &deviceEncoding) != errNone)
	{
		deviceEncoding = charEncodingPalmLatin;
	}

	// Get a table that will map from the source to dest character set.

	int16 numEntries;
	const CharMapType* charMapP = ::GetCharMap (deviceEncoding,
												macEncoding,
												numEntries);

	// Iterate over all the input characters.

	ByteList::const_iterator	iter = palmChars.begin ();

	while (iter < palmChars.end ())
	{
		// Get a character.

		WChar ch;
		iter += _TxtGetNextChar ((const Char*) iter, 0, &ch);

		// Map the character.  If a mapping exists but it maps to
		// 0, then use the undefined character (currently '?').

		if (charMapP != NULL)
		{
			void* result = bsearch (&ch, charMapP, numEntries,
									sizeof (CharMapType), ::CompareCharToMap);

			if (result != NULL)
			{
				const CharMapType* theMap = (const CharMapType*) result;
				ch = theMap->dstChar;
			}
		}
		else
		{
			if (ch == chrLineFeed)
			{
				ch = 0x000D;
			}
			else if (ch >= 0x0080)
			{
				ch = kNoMap;
			}
		}

		// Add the mapped character to our output vector.

		if (ch == kNoMap)
		{
			hostChars.push_back (kUndefinedChar);
		}
		else if (ch < 0x0100)
		{
			hostChars.push_back (ch);
		}
		else
		{
			hostChars.push_back (ch >> 8);
			hostChars.push_back (ch);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	Platform::RemapHostToPalmChars
 *
 * DESCRIPTION:	Copy data from Palm clipboard to Mac clipboard,
 *	transliterating text data as needed.
 *
 * PARAMETERS:
 *	data	 ->	Pointer to the Palm clipboard data.
 *	length	 ->	Bytes of data in <data>
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void Platform::RemapHostToPalmChars (const ByteList& hostChars,
									ByteList& palmChars)
{
	// Clear the output vector.

	palmChars.clear ();

	// Get the source and dest encoding.

	uint16	macEncoding = ::FontToScript (::GetSysFont ());
	uint32	deviceEncoding;

	if (::FtrGet (sysFtrCreator, sysFtrNumEncoding, &deviceEncoding) != errNone)
	{
		deviceEncoding = charEncodingPalmLatin;
	}

	// Get a table that will map from the source to dest character set.

	int16 numEntries;
	const CharMapType* charMapP = ::GetCharMap (macEncoding,
												deviceEncoding,
												numEntries);

	// Iterate over all the input characters.

	ByteList::const_iterator	iter = hostChars.begin ();

	while (iter < hostChars.end ())
	{
		// Get a character, testing to see if it's a double-byte character
		// and fetching the second byte if so.  NOTE: that this code fails
		// for > 2 byte characters!

		short			byteType	= ::CharacterByteType ((Ptr) iter, 0, macEncoding);
		unsigned short	ch			= *iter++;

		if ((byteType == smFirstByte) && (iter < hostChars.end ()))
		{
			ch = (ch << 8) | *iter++;
		}

		// Map the character.  If a mapping exists but it maps to
		// 0, then use the undefined character (currently '?').

		if (charMapP != NULL)
		{
			void* result = bsearch (&ch, charMapP, numEntries,
									sizeof (CharMapType), ::CompareCharToMap);

			if (result != NULL)
			{
				const CharMapType* theMap = (const CharMapType*) result;
				ch = theMap->dstChar;
			}
		}
		else
		{
			if (ch == 0x000D)
			{
				ch = chrLineFeed;
			}
			else if (ch >= 0x0080)
			{
				ch = kNoMap;
			}
		}

		// Add the mapped character to our output vector.

		if (ch == kNoMap)
		{
			palmChars.push_back (kUndefinedChar);
		}
		else if (ch < 0x0100)
		{
			palmChars.push_back (ch);
		}
		else
		{
			palmChars.push_back (ch >> 8);
			palmChars.push_back (ch);
		}
	}
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	Platform::PinToScreen
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	True if the rectangle was changed.
 *
 ***********************************************************************/

Bool Platform::PinToScreen (EmRect& r)
{
	// Get the visible desktop area.

	RgnHandle	desktopRgn = ::GetGrayRgn ();

	// If we're in it, just return now.  We're not changing
	// the rectangle, so return false.

	Point	topLeft = { r.fTop, r.fLeft };
	Point	topRight = { r.fTop, r.fRight - 1 };
	Point	bottomLeft = { r.fBottom - 1, r.fLeft };
	Point	bottomRight = { r.fBottom - 1, r.fRight - 1 };

	if (::PtInRgn (topLeft, desktopRgn) &&
		::PtInRgn (topRight, desktopRgn) &&
		::PtInRgn (bottomLeft, desktopRgn) &&
		::PtInRgn (bottomRight, desktopRgn))
	{
		return false;
	}

	// We need to adjust the rectangle.  Walk the device list,
	// looking for the device we intersect most with.

	long		maxArea = 0;
	GDHandle	currentDevice = ::GetDeviceList ();
	GDHandle	maxDevice = ::GetMainDevice ();	// Default to main screen

	while (currentDevice)
	{
		Rect	gdRect = (**currentDevice).gdRect;
		Rect	intersection;
		Rect	macRect = r;
		::SectRect (&macRect, &gdRect, &intersection);

		long	area = (intersection.right - intersection.left) * 
						(intersection.bottom - intersection.top);

		if (maxArea < area)
		{
			maxArea = area;
			maxDevice = currentDevice;
		}

		currentDevice = ::GetNextDevice (currentDevice);
	}

	// We have the device we're going to move the window onto.
	// Get it's bounds, intersect it with the desktop region
	// (in order to account for the menubar), and then work with
	// that rectangular bounds.

	EmAssert (maxDevice);

	Rect		gdRect = (**maxDevice).gdRect;
	RgnHandle	gdRgn = ::NewRgn ();

	::RectRgn (gdRgn, &gdRect);
	::SectRgn (desktopRgn, gdRgn, gdRgn);

	::GetRegionBounds (gdRgn, &gdRect);

	::DisposeRgn (gdRgn);

	// Finally, pin the incoming rectangle to gdRect.

	EmRect	absRect (gdRect);
	::PinRectInRect (r, absRect);

	// We modified the original rect, so return true.

	return true;
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	Platform::ToHostEOL
 *
 * DESCRIPTION: Converts a string of characters into another string
 *				where the EOL sequence is consistant for files on the
 *				current platform.
 *
 * PARAMETERS:	dest - receives the result.  The buffer is sized by
 *					this function, so the caller doesn't have to
 *					allocate any space itself.
 *
 *				destLen - receives the length of the resulting string.
 *
 *				src - pointer to input characters.
 *
 *				srcLen - number of characters pointed to by src.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void Platform::ToHostEOL (	StMemory& dest, long& destLen,
							const char* src, long srcLen)
{
	char*	d = (char*) Platform::AllocateMemory (srcLen);
	char*	p = d;
	Bool	previousWas0x0D = false;

	for (long ii = 0; ii < srcLen; ++ii)
	{
		char	ch = src[ii];

		// Convert 0x0A to 0x0D, but not if it's
		// part of a 0x0D/0x0A sequence

		if (ch == 0x0A)
		{
			if (!previousWas0x0D)
				*p++ = 0x0D;
		}

		// Copy all other characters straight through.

		else
		{
			*p++ = ch;
		}

		previousWas0x0D = ch == 0x0D;
	}

	destLen = p - d;
	d = (char*) Platform::ReallocMemory (d, destLen);
	dest.Adopt (d);
}


// ---------------------------------------------------------------------------
//		 Platform::ReadROMFileReference
// ---------------------------------------------------------------------------

Bool Platform::ReadROMFileReference (ChunkFile& docFile, EmFileRef& f)
{
	// First look for a ROM file alias.

	Chunk	chunk;
	if (docFile.ReadChunk (SessionFile::kROMAliasTag, chunk))
	{
		Handle	romFileAlias = ::NewHandle (chunk.GetLength ());
		Errors::ThrowIfNULL (romFileAlias);

		StHandleLocker	locker (romFileAlias);
		memcpy (*romFileAlias, chunk.GetPointer (), chunk.GetLength ());

		f = EmFileRef ((AliasHandle) romFileAlias);

		if (f.Exists ())
			return true;
	}

	return false;
}


// ---------------------------------------------------------------------------
//		 Platform::WriteROMFileReference
// ---------------------------------------------------------------------------

void Platform::WriteROMFileReference (ChunkFile& docFile, const EmFileRef& f)
{
	// Save the ROM alias.

	Handle			romFileAlias = (Handle) f.GetAlias ();
	StHandleLocker	locker (romFileAlias);
	docFile.WriteChunk (SessionFile::kROMAliasTag,
					::GetHandleSize (romFileAlias), *romFileAlias);
}


// ---------------------------------------------------------------------------
//		 Platform::Delay
// ---------------------------------------------------------------------------
// Delay for a time period appropriate for a sleeping 68328.

void Platform::Delay (void)
{
	// Delay 9 ms (8.8 msecs was actually required). This value was
	// empirically determined by seeing if the Palm device would go to
	// sleep after the right amount of time.

	AbsoluteTime	theExpirationTime = ::UpTime ();
	theExpirationTime = ::AddDurationToAbsolute (9, theExpirationTime);
	::MPDelayUntil (&theExpirationTime);
}


// ---------------------------------------------------------------------------
//		 Platform::CycleSlowly
// ---------------------------------------------------------------------------

void Platform::CycleSlowly (void)
{
	if (TickCount () - gLastCycleTime > EVENT_THRESHHOLD)
	{
		gLastCycleTime = TickCount ();

		// Break from the CPU loop so that we can process events, refresh
		// the screen, track the mouse, etc.

		EmAssert (gSession);
		gSession->ScheduleSuspendTimeout ();
	}
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 Platform::RealAllocateMemory
// ---------------------------------------------------------------------------
// Allocate a block of memory from Tempoary memory. Save the handle, lock it
// down, and return the pointer to the caller.

typedef map<Ptr, Handle>	HandleMap;
HandleMap&	PrvGetHandleMap (void);

// Private function to create and return a reference to the HandleMap.	We
// have this as a private function instead of a naked global because
// AllocateMemory can be called by the contructors of other global objects.
// If we have gHandleMap as a naked global, we wouldn't know if it had
// been initialized when AllocateMemory was called.

HandleMap&	PrvGetHandleMap (void)
{
	static HandleMap	gHandleMap;

	// Make sure that we're not trying to access this after exitting
	// main().  After main(), the runtime cleans up global objects by
	// calling their destructors.  Any destructors calling DisposeMemory
	// may fail, since gHandleMap may have already been destroyed itself!

	EmAssert (!gExittedMain);

	return gHandleMap;
}

void* Platform::RealAllocateMemory (size_t size, Bool clear, const char* file, int line)
{
#pragma unused (file, line)

	// Allocate the new block from temporary memory.

	OSErr	err = noErr;
	Handle	hdl = ::TempNewHandle (size, &err);
	Errors::ThrowIfError (err);

	// If the user wants it cleared, clear it (there doesn't appear to
	// be a TempNewHandleClear call).

	if (clear)
	{
		memset (*hdl, 0, size);
	}

	// Lock the memory; we're returning just a pointer, so it can't move around.
	// I'm assuming that calling MoveHHi here is OK, and that the Memory Manager
	// is smart about where to really move the handle.

	::HLockHi (hdl);

	// Remember the pointer and it's associated handle. That we, we can
	// do our own version of RecoverHandle later.

	PrvGetHandleMap().insert (HandleMap::value_type (*hdl, hdl));

	// Return the pointer to the caller.

	return *hdl;
}


// ---------------------------------------------------------------------------
//		 Platform::RealReallocMemory
// ---------------------------------------------------------------------------
// Resize the relocatable memory block indicated by "p".

void* Platform::RealReallocMemory (void* p, size_t newSize, const char* file, int line)
{
#pragma unused (file, line)

	// If "p" is NULL, treat this like an AllocateMemory call.

	if (p == NULL)
	{
		return Platform::AllocateMemory (newSize);
	}

	// If "newSize" is zero, treat this like a DisposeMemory call.

	if (newSize == 0)
	{
		Platform::DisposeMemory (p);
		return NULL;
	}

	// Look for the handle for this pointer.  If we can't
	// find it (probably due to an error by the caller), return NULL.

	HandleMap::iterator iter = PrvGetHandleMap().find ((Ptr) p);

	if (iter != PrvGetHandleMap().end ())
	{
		// Get handy references to our data.

		Ptr 	ptr = (*iter).first;
		Handle	hdl = (*iter).second;

		// Make sure we got what we expect.

		EmAssert (ptr == (Ptr) p);
		EmAssert (ptr == *hdl);

		// Unlock and resize the memory block.

		::HUnlock (hdl);
		::SetHandleSize (hdl, newSize);

		// Remove the old entry from the map.

		PrvGetHandleMap().erase (iter);

		// Lock down the handle and re-add it to the map.
		// I'm assuming that calling MoveHHi here is OK, and that the Memory Manager
		// is smart about where to really move the handle.

		::HLockHi (hdl);
		PrvGetHandleMap().insert (HandleMap::value_type (*hdl, hdl));

		// Return the pointer to the caller.

		return *hdl;
	}

	EmAssert (NULL);

	return NULL;
}


// ---------------------------------------------------------------------------
//		 Platform::RealDisposeMemory
// ---------------------------------------------------------------------------

void Platform::RealDisposeMemory (void* p)
{
	// If p is NULL, just return.

	if (!p)
		return;

	// Look for the handle for this pointer.  If we can't
	// find it (probably due to an error by the caller), do nothing.

	HandleMap::iterator iter = PrvGetHandleMap().find ((Ptr) p);

	if (iter != PrvGetHandleMap().end ())
	{
		// Get handy references to our data.

		Ptr 	ptr = (*iter).first;
		Handle	hdl = (*iter).second;

		// Make sure we got what we expect.

		EmAssert (ptr == (Ptr) p);
		EmAssert (ptr == *hdl);

		// Dispose of the memory block. If we get an error, do nothing.

		OSErr	err;
		::TempDisposeHandle (hdl, &err);

		// Remove the entry from the map.

		PrvGetHandleMap().erase (iter);
	}

	else
	{
		EmAssert (NULL);
	}
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	Platform::ForceStartupScreen
 *
 * DESCRIPTION:	See if the user has requested that the Startup dialog
 *				be presented instead of attempting to use the latest
 *				session file or creation settings.
 *
 *				The current signal is to toggle the CAPSLOCK key.
 *
 * PARAMETERS:	none
 *
 * RETURNED:	True if the user has signalled that the dialog should
 *				be presented.
 *
 ***********************************************************************/

Bool Platform::ForceStartupScreen (void)
{
	UInt32	mods = ::GetCurrentKeyModifiers ();
	return (mods & alphaLock) != 0;
}


// ---------------------------------------------------------------------------
//		 Platform::StopOnResetKeyDown
// ---------------------------------------------------------------------------

Bool Platform::StopOnResetKeyDown (void)
{
#if 0
	const int		kKey = 0x3B;	// Ctrl key
	KeyMapByteArray keyMap;

	::GetKeys (*(KeyMap*) &keyMap);
	return (keyMap[kKey / 8] & (1 << (kKey & 7))) != 0;
#endif

	UInt32	mods = ::GetCurrentKeyModifiers ();
	return (mods & controlKey) != 0;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 Platform::CollectOptions
// ---------------------------------------------------------------------------

Bool Platform::CollectOptions (int argc, char** argv, int& errorArg,
								int (*cb)(int, char**, int&))
{
	errorArg = 1;	// Skip argv[0]
	while (errorArg < argc)
	{
		if (cb (argc, argv, errorArg) == 0)
			return false;
	}

	return true;
}


// ---------------------------------------------------------------------------
//		 Platform::PrintHelp
// ---------------------------------------------------------------------------

void Platform::PrintHelp (void)
{
}



#pragma mark -

// ---------------------------------------------------------------------------
//		 Platform::GetMilliseconds
// ---------------------------------------------------------------------------

uint32 Platform::GetMilliseconds (void)
{
	UnsignedWide	now;
	Microseconds (&now);

	// The number of microseconds was stored in a 64-bit field.  We want to
	// return the number of _milliseconds_ in a _32-bit_ field.  To perform
	// the conversion, we extract bits 10 through 42 (giving us our 32 bits),
	// shifting them down by 10 (effectively dividing by 1024 to convert from
	// microseconds to milliseconds).


//	64-bit microseconds:
//
//	+----------- now.hi ------------+----------- now.lo ------------+
//
//	64		56		48		40		32		24		16		8		0
//	|		|		|		|		|		|		|		|		|
//	 ****************************************************************
//						  42							  10
//						  +-------------------------------+
//						  | 							  |
//						  V 							  V
//
//	32-bit milliseconds:
//						  32							  0
//						  | 							  |
//						   ********************************
//

	return (now.hi << 22) | (now.lo >> 10); // return milliseconds
}


#pragma mark -

/***********************************************************************
 *
 * FUNCTION:	Platform::CreateDebuggerSocket
 *
 * DESCRIPTION: .
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

CSocket* Platform::CreateDebuggerSocket (void)
{
	if (!::EmClassicHasPPCToolbox ())
		return NULL;

	return new CPPCSocket;
}


// ---------------------------------------------------------------------------
//		 Platform::ExitDebugger
// ---------------------------------------------------------------------------

void Platform::ExitDebugger (void)
{
	// Oops...back this one out.  The idea here was to bring Poser to
	// front when we exited the source-level debugger.	Unfortunately,
	// this appears to mess up single-stepping.  Single-stepping exits
	// the debugger in the same way that executing "Run" does.	But in
	// the case of the former, we don't want to bring Poser to front.
	// If we did that, the debugger would be in the background, and it
	// doesn't bring itself back intothe foreground when it's re-entered.

	// Perhaps later we could implement some sort of heuristic that says
	// to bring Poser foreward if some small amount of time has elapsed
	// since we exited debug mode.

//	gApplication->BringToFront ();
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 Platform::ViewDrawLine
// ---------------------------------------------------------------------------

void Platform::ViewDrawLine (int xStart, int yStart, int xEnd, int yEnd)
{
#pragma unused (xStart, yStart, xEnd, yEnd)

#if 0
	if (gScreenPane != NULL)
	{
		gScreenPane->FocusDraw();
		
		::PenNormal();
		::MoveTo(xStart, yStart);
		::Line(xEnd - xStart, yEnd - yStart);
	}
#endif
}


// ---------------------------------------------------------------------------
//		 Platform::ViewDrawPixel
// ---------------------------------------------------------------------------

void Platform::ViewDrawPixel (int xPos, int yPos)
{
#pragma unused (xPos, yPos)

#if 0
	if (gScreenPane != NULL)
	{
		gScreenPane->FocusDraw();
		
		::PenNormal();
		::MoveTo(xPos, yPos);
		::Line(0, 0);
	}
#endif
}

#pragma mark -

#define midiMaxKey				(127)				// maximum note key #
#define midiMaxVel				(127)				// maximum key velocity

#define prvTwelfthRootOfTwo		twelfthRootTwo

#define prvFrqToMidiKey(frq)	\
	((long)(log10(frq) / log10(prvTwelfthRootOfTwo) - 36.37))

static SndChannelPtr GMacSndChanP	 = NULL;	// pointer to current Macintosh sound channel

static void PrvPlaySound (int amplitude, int midiKey, int duration)
{
	// If no amplitude or duration, do nothing.
	if (amplitude == 0 || duration == 0)
		return;

	SndCommand	sndCmd;

	// Set the channel volume
	sndCmd.cmd		= volumeCmd;	// ampCmd;
	sndCmd.param1	= min (254, 254 * amplitude / sndMaxAmp);
	sndCmd.param2	= 0;
	SndDoCommand (GMacSndChanP, &sndCmd, false);

	// Play the note
	sndCmd.cmd		= rateMultiplierCmd;	// freqDurationCmd;
	sndCmd.param1	= duration * 2;
	sndCmd.param2	= min (midiKey, 127);
	SndDoCommand (GMacSndChanP, &sndCmd, false);

	// Turn off the note
	sndCmd.cmd		= quietCmd;
	sndCmd.param1	= 0;
	sndCmd.param2	= 0;
	SndDoCommand (GMacSndChanP, &sndCmd, false);
}

static void PrvCmdFreqDurationAmp (SndCommandType& cmd)
{
	int	amplitude	= cmd.param3;
	int	midiKey		= prvFrqToMidiKey (cmd.param1);
	int	duration	= cmd.param2;

	PrvPlaySound (amplitude, midiKey, duration);

	// !!! Wait for it to finish?  If so, then we need to
	// queue up a callBackCmd.
}

static void PrvCmdNoteOn (SndCommandType& cmd)
{
	int	amplitude	= (sndMaxAmp * cmd.param3) / midiMaxVel;
	int	midiKey		= cmd.param1;
	int	duration	= cmd.param2;

	PrvPlaySound (amplitude, midiKey, duration);
}

static void PrvCmdFrqOn (SndCommandType& cmd)
{
	int	amplitude	= cmd.param3;
	int	midiKey		= prvFrqToMidiKey (cmd.param1);
	int	duration	= cmd.param2;

	PrvPlaySound (amplitude, midiKey, duration);
}

static void PrvCmdQuiet (void)
{
	SndCommand	sndCmd;

	sndCmd.cmd		= quietCmd;
	sndCmd.param1	= 0;
	sndCmd.param2	= 0;
	SndDoImmediate (GMacSndChanP, &sndCmd);
}

static SndChannelPtr PrvMakeSoundChannel (void)
{
	if (GMacSndChanP)
		return GMacSndChanP;
	
	SndChannelPtr	chanP = NULL;
	OSErr			osErr = SndNewChannel (&chanP, 0 /*squareWaveSynth*/, 0, NULL);
	if (osErr)
		return NULL;

	GMacSndChanP = chanP;

	return chanP;
}

static void PrvFreeSoundChannel (void)
{
	if (GMacSndChanP)
	{
		PrvCmdQuiet ();
		SndDisposeChannel (GMacSndChanP, false);
		GMacSndChanP = NULL;
	}
}

CallROMType Platform::SndDoCmd (SndCommandType& cmd)
{
	// Note: according to Sound.h, if a sound is playing when a new command
	// comes in, the previous sound is interrupted.  That's what the calls
	// to PrvCmdQuit are for below.  However, having those in there cause
	// the alarm sound to play too quickly.  Aparently, the call to
	// SysTaskDelay in PrvPlayAlarmSound is not taking the 150 msecs it's
	// supposed to.  So in order to get the expected effect, don't kill the
	// previous sound before embarking on a new one.

	if (PrvMakeSoundChannel())
	{
		switch (cmd.cmd)
		{			
			case sndCmdFreqDurationAmp:
				PrvCmdFreqDurationAmp (cmd);
				break;

			case sndCmdNoteOn:
//				PrvCmdQuiet ();
				PrvCmdNoteOn (cmd);
				break;

			case sndCmdFrqOn:
//				PrvCmdQuiet ();
				PrvCmdFrqOn (cmd);
				break;

			case sndCmdQuiet:
				PrvCmdQuiet ();
				break;
		}
	}

	return kSkipROM;
}

void Platform::StopSound()
{
}


void Platform::Beep()
{
	::SysBeep (5);
}


#pragma mark -

CPPCSocket::CPPCSocket (void) :
	fConnected (false),
	fShortPacketHack (true)
{
}

CPPCSocket::~CPPCSocket (void)
{
	deque<Handle>::iterator	iter = fBufferedPackets.begin();

	while (iter != fBufferedPackets.end())
	{
		::DisposeHandle(*iter);
		++iter;
	}

	fBufferedPackets.clear();
}

ErrCode 		
CPPCSocket::Open (void)
{
	return PPCOpenPorts (kEmulatorPortName, kDebuggerPortName,
						kFromEmulatorPortType, kToEmulatorPortType);
}

ErrCode 		
CPPCSocket::Close (void)
{
	return PPCClosePorts ();
}

ErrCode 		
CPPCSocket::Write (const void* buffer, long amountToWrite, long* amtWritten)
{
	if (amtWritten)
		*amtWritten = amountToWrite;

	ErrCode	err = PPCSendData (buffer, amountToWrite);

	if (err == kPPC_ErrBadState)
	{
		if (LogLLDebugger ())
			LogAppendMsg ("PPCSendData busy, queueing packet.");

		Handle	h = ::NewHandle (amountToWrite);
		memcpy (*h, buffer, amountToWrite);
		fBufferedPackets.push_back(h);
	}

	return err;
}

ErrCode 		
CPPCSocket::Read (void* buffer, long sizeOfBuffer, long* amtRead)
{
	if (sizeOfBuffer > fBufferedData.size ())
		sizeOfBuffer = fBufferedData.size ();

	if (amtRead)
		*amtRead = sizeOfBuffer;

	copy (fBufferedData.begin (), fBufferedData.begin () + sizeOfBuffer, (char*) buffer);
	fBufferedData.erase (fBufferedData.begin (), fBufferedData.begin () + sizeOfBuffer);

	return errNone;
}

Bool		
CPPCSocket::HasUnreadData (long timeout)
{
#pragma unused (timeout)

	return fBufferedData.size () > 0;
}

ErrCode 		
CPPCSocket::Idle (void)
{
	// WARNING!  Make sure we don't allocate any memory in this function or
	// the functions it calls.	This function may be called within the context
	// of the application that sent us the PPC data.  Any allocated memory
	// would thus be allocated from *that* application's heap, not ours.

	// Set a flag saying that we shouldn't allocate memory.  This flag is
	// checked in Platform::AllocateMemory and in a custom version of
	// operator new that we provide.

	if (!fConnected && PPCPortConnected (&gReceivePort))
	{
		fConnected = true;
	}
	else if (fConnected && !PPCPortConnected (&gReceivePort))
	{
		fConnected = false;
	}

	// See if the ports are still in place.

	if (PPCPortOpened (&gReceivePort) &&
		!PPCPortConnecting (&gReceivePort) &&
		!PPCPortConnected (&gReceivePort))
	{
		if (LogLLDebugger ())
			LogAppendMsg ("Found broken port. Called ClosePorts.");

		this->Close ();
		this->Open ();
	}

	// If still connected, see if there's data from the debugger.

	if (gReceivePort.fDataWaiting)
	{
		if (LogLLDebugger ())
			LogAppendMsg ("Handling received packet.");
		this->HandlePacket ();
		if (LogLLDebugger ())
			LogAppendMsg ("Handled received packet.");

		// Start listening for more data.
		PPCListenForData (&gReceivePort);
	}

	// If we have any queued data, try sending it.

	if (fBufferedPackets.size() > 0)
	{
		if (LogLLDebugger ())
			LogAppendMsg ("Found queued packet, resending.");

		Handle	hdl = fBufferedPackets[0];
		ErrCode	err;

		{
			StHandleLocker	locker(hdl);
			
			void*	buffer = *hdl;
			long	amountToWrite = ::GetHandleSize (hdl);

			err = PPCSendData (buffer, amountToWrite);
		}

		if (err == errNone)
		{
			fBufferedPackets.pop_front ();
			::DisposeHandle (hdl);
		}
		else
		{
			if (LogLLDebugger ())
				LogAppendMsg ("PPCSendData still busy.");
		}
	}

	return errNone;
}


void CPPCSocket::HandlePacket (void)
{
	// This is a temporary staging buffer.	When data is received from
	// another process, it's copied into here so that we can add it to
	// fBufferedData.  On the other end, when we have enough in the deque
	// that we can process it, the data is moved back into here so that
	// Debug::HandlePacket can grope it.

	unsigned char	buffer[ 10 +				// EmProxySlkPktHeaderType::GetSize ()
							sysPktMaxBodySize +	// EmProxySysPktBodyType::GetSize ()
							2];					// EmProxySlkPktFooterType::GetSize ()

	// Buffer up the data received from the other process.

	long	amt = PPCReceiveData (&buffer, sizeof (buffer));

	// See if this looks like the "Full Protocol" hack.

	if (strcmp ((char*) buffer, "Full Protocol") == 0)
	{
		fShortPacketHack = false;
		amt = 0;
	}

	// It's not, so insert the received data into our long-term buffer.

	fBufferedData.insert (fBufferedData.end (), &buffer[0], &buffer[amt]);

	// See if we have enough data to process.

	long	headerSize	= EmProxySlkPktHeaderType::GetSize ();
	long	bodySize	= EmProxySysPktBodyType::GetSize ();
	long	footerSize	= EmProxySlkPktFooterType::GetSize ();

	if (fBufferedData.size () > headerSize)
	{
		// Get the bodySize field to see if we have the entire body, too.

		copy (fBufferedData.begin (), fBufferedData.begin () + headerSize, buffer);
		EmAliasSlkPktHeaderType<LAS>	header (buffer);
		bodySize = header.bodySize;
	}

	// Get the total size of the number of bytes that need to be buffered
	// if we are to believe that we have received a full packet.

	long	totalSize = headerSize + bodySize + footerSize;

	if (fShortPacketHack)
	{
		totalSize -= footerSize;
	}

	// If we have at least that many bytes, process them.

	if (fBufferedData.size () >= totalSize)
	{
		try
		{
			SLP::EventCallback (this, CSocket::kDataReceived);
		}
		catch (const EmExceptionReset& e)
		{
			if (gHostDocument)
			{
				gHostDocument->SetPendingException (e);
			}
		}
	}

	gReceivePort.fDataWaiting = false;
}


Bool		
CPPCSocket::ShortPacketHack (void)
{
	return fShortPacketHack;
}
