/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 2000-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "TracerPlatform.h"

#include "ErrorHandling.h"		// Errors::ThrowIfError
#include "PreferenceMgr.h"
#include "Strings.r.h"

#include <CodeFragments.h>

// ===========================================================================
//		 Define Shared Library and Trace Function Name
// ===========================================================================

#define kTraceLibName		"\pPalmTraceLib"

#define kInitOutputName		"\ptracer_init_outputport"
#define kCloseOutpuName		"\ptracer_close_outputport"
#define kOutputStatusName	"\ptracer_get_outputport_status"
#define kTracerGetCapsName	"\ptracer_get_caps"
#define kTraceVTName		"\ptracer_trace_VT"
#define kTraceVTLName		"\ptracer_trace_VTL"
#define kTraceBName			"\ptracer_trace_B"

Tracer	gTracer;

// ===========================================================================
//		 Static Trace Function Pointer
// ===========================================================================

void (*tracer_init_outputport)			(const char*, const char*);
void (*tracer_close_outputport)			(void);

void (*tracer_trace_VT)					(UInt16, const char*, const va_list*);
void (*tracer_trace_VTL)				(UInt16, const char*, const va_list*);
void (*tracer_trace_B)					(UInt16, const void*, size_t);

long (*tracer_get_outputport_status)	(void);
void (*tracer_get_caps)					(char*, size_t*);


// ===========================================================================
//		 Static
// ===========================================================================

static CFragConnectionID	sConnectionID;

// ---------------------------------------------------------------------------
//		 Tracer::Tracer
// ---------------------------------------------------------------------------

Tracer::Tracer(void)
{
}

// ---------------------------------------------------------------------------
//		 Tracer::~Tracer
// ---------------------------------------------------------------------------

Tracer::~Tracer(void)
{
}

// ---------------------------------------------------------------------------
//		 Tracer::Initialize
// ---------------------------------------------------------------------------

static void PrvLoadFunction (ConstStr255Param symName, void* addr)
{
	unsigned char	symClass;
	OSErr	theErr = ::FindSymbol (sConnectionID, symName, (Ptr*) addr, &symClass);
	Errors::ThrowIfError (theErr);
}

void Tracer::Initialize (void)
{
	OSErr			theErr;
	Str255			theErrName;
	ProcPtr			theMainAddr;
	tracer_init_outputport			= NULL;
	tracer_close_outputport			= NULL;
	tracer_get_outputport_status	= NULL;
	tracer_get_caps					= NULL;
	tracer_trace_VT					= NULL;
	tracer_trace_VTL				= NULL;
	tracer_trace_B					= NULL;

	if (!UEnvironment::HasGestaltAttribute (gestaltCFMAttr, gestaltCFMPresent))
		return;

	try
	{
		// Load the Shared Library
		theErr = ::GetSharedLibrary (kTraceLibName, kCompiledCFragArch, kLoadCFrag, &sConnectionID, (Ptr *) &theMainAddr, theErrName);
		Errors::ThrowIfError (theErr);

		::PrvLoadFunction (kInitOutputName, &tracer_init_outputport);
		::PrvLoadFunction (kCloseOutpuName, &tracer_close_outputport);
		::PrvLoadFunction (kOutputStatusName, &tracer_get_outputport_status);
		::PrvLoadFunction (kTracerGetCapsName, &tracer_get_caps);
		::PrvLoadFunction (kTraceVTName, &tracer_trace_VT);
		::PrvLoadFunction (kTraceVTLName, &tracer_trace_VTL);
		::PrvLoadFunction (kTraceBName, &tracer_trace_B);
	}
	catch (ErrCode errCode)
	{
//		Errors::ReportIfError (kStr_GenericOperation, errCode, 0, false);
		Tracer::Dispose ();
	}

	// Request list of supported tracer Types
	LoadTracerTypeList ();

	// No trace user referenced at this time
	tracerRefCounter = 0;
}

// ---------------------------------------------------------------------------
//		 Tracer::Dispose
// ---------------------------------------------------------------------------

void Tracer::Dispose (void)
{
	OSErr		theErr;
	
	if (sConnectionID)
	{
		// Close any pending connection
		StopTracer();

		tracer_init_outputport			= NULL;
		tracer_close_outputport			= NULL;
		tracer_get_outputport_status	= NULL;
		tracer_get_caps					= NULL;
		tracer_trace_VT					= NULL;
		tracer_trace_VTL				= NULL;
		tracer_trace_B					= NULL;

		theErr = ::CloseConnection(&sConnectionID);
		sConnectionID = NULL;
		tracerRefCounter = 0;
	}	

	DisposeTracerTypeList();
}


// ---------------------------------------------------------------------------
//		 Tracer::InitOutputPort
// ---------------------------------------------------------------------------

void Tracer::InitOutputPort	(void)
{
	tracerRefCounter++;

	if (tracerRefCounter == 1)
	{
		if (tracer_init_outputport)
			tracer_init_outputport (GetTracerTypeInfo(runningTracerType)->type,
									GetTracerTypeInfo(runningTracerType)->paramCurVal);
	}
}


// ---------------------------------------------------------------------------
//		 Tracer::CloseOutputPort
// ---------------------------------------------------------------------------

void Tracer::CloseOutputPort (void)
{
	if (tracerRefCounter == 1)
	{
		// We only close the output port (the current tracer instance keeps alive)
		if (tracer_close_outputport)
			tracer_close_outputport ();
	}
	
	if (tracerRefCounter > 0)
		tracerRefCounter--;
}


// ---------------------------------------------------------------------------
//		 Tracer::OutputVT
// ---------------------------------------------------------------------------

void Tracer::OutputVT (unsigned short errModule, const char* formatString, va_list args)
{
	if (tracerRefCounter && tracer_trace_VT)
		tracer_trace_VT (errModule, formatString, &args);
}


// ---------------------------------------------------------------------------
//		 Tracer::OutputVTL
// ---------------------------------------------------------------------------

void Tracer::OutputVTL (unsigned short errModule, const char* formatString, va_list args)
{
	if (tracerRefCounter && tracer_trace_VTL)
		tracer_trace_VTL (errModule, formatString, &args);
}


// ---------------------------------------------------------------------------
//		 Tracer::OutputB
// ---------------------------------------------------------------------------

void Tracer::OutputB (unsigned short errModule, const void* buffer, size_t bufferLen)
{
	if (tracerRefCounter && tracer_trace_B)
		tracer_trace_B (errModule, buffer, bufferLen);
}


// ---------------------------------------------------------------------------
//		 Tracer::GetConnectionStatus
// ---------------------------------------------------------------------------

void Tracer::StopTracer(void)
{
	if (tracer_close_outputport)
		tracer_close_outputport ();

	tracerRefCounter = 0;
}

// ---------------------------------------------------------------------------
//		 Tracer::GetConnectionStatus
// ---------------------------------------------------------------------------

long Tracer::GetConnectionStatus(void)
{
	if (tracer_get_outputport_status)
		return tracer_get_outputport_status ();
	
	return 1; // 1 = Not connected
}

// ---------------------------------------------------------------------------
//		 Tracer::SetCurrentTracerTypeIndex
// ---------------------------------------------------------------------------

void Tracer::SetCurrentTracerTypeIndex (unsigned short tracerType, Bool paramChange)
{
	if (runningTracerType)
	{
		previousTracerType = runningTracerType;
	}

	runningTracerType = tracerType;

	// If there are trace subsystem users
	if (tracerRefCounter > 0)
	{
		// If we feel that we should close the current output port and open a new one
		if (tracerType != previousTracerType || paramChange)
		{
			// Establish connection to the new tracer
			// As a side effect this closes any pending connection, whatever the previous tracer Type was
			if (tracer_init_outputport)
				tracer_init_outputport (GetTracerTypeInfo(runningTracerType)->type,
										GetTracerTypeInfo(runningTracerType)->paramCurVal);
		}
	}
}

// ---------------------------------------------------------------------------
//		 Tracer::GetTracerCapabilities
// ---------------------------------------------------------------------------

void Tracer::GetTracerCapabilities (char* buffer, size_t* bufferLen)
{
	if (tracer_get_caps)
	{
		tracer_get_caps (buffer, bufferLen);
	}
	else
	{
		if (buffer && (*bufferLen > 1))
		{
			buffer[0] = 0;
			buffer[1] = 0;
			*bufferLen = 2;
		}
	}
}

// ---------------------------------------------------------------------------
//		 Tracer::GetLibraryVersionString
// ---------------------------------------------------------------------------

void Tracer::GetLibraryVersionString (char* buffer, size_t bufferLen)
{
	OSErr				err;
	FSSpec				libraryFSSPec;
	CFragConnectionID	connectionID = 0;	
	Str255				errMessage;
	Ptr					mainAddr;
	LFile				libraryFile;

	if (sConnectionID)
	{
		try
		{
			// I'm not sure how this is supposed to work.  And, indeed,
			// it doesn't.  GetDiskFragment returns error -50, with
			// errMessage saying that a "real" FSSpec needs to be passed
			// in.  I expect that the FSSpec is an input parameter, not
			// an output parameter.  Unfortunately, I can't find any other
			// way to find the file containing the library.  So, for now,
			// we get errors.

			err = ::GetDiskFragment (&libraryFSSPec, 0, kCFragGoesToEOF,
									kTraceLibName, kFindCFrag, &connectionID,
									&mainAddr, errMessage);		
			Errors::ThrowIfError(err);
			
			libraryFile.SetSpecifier (libraryFSSPec);
			libraryFile.OpenResourceFork(fsRdPerm);

			StResource		versionResource ('vers', 1);		
			VersRecHndl 	versHandle = (VersRecHndl) (Handle) versionResource;
			StHandleLocker	lock ((Handle) versHandle);
			LStr255			version;
			
			if (versHandle)
			{
				VersRecPtr	versptr			= *versHandle;
				version = versptr->reserved;
				
				strncpy(buffer, version.TextPtr(), bufferLen);
			}
			
			libraryFile.CloseDataFork ();
		}
		catch (ErrCode errCode)
		{
			strncpy (buffer, "Unable to get Palm Reporter version number", bufferLen);
		}
	}
	else
	{
		strncpy (buffer, "Palm Reporter Library not found", bufferLen);
	}
}

// ---------------------------------------------------------------------------
//		 Tracer::GetLibraryVersionNumber
// ---------------------------------------------------------------------------

unsigned long Tracer::GetLibraryVersionNumber (void)
{
	return GetLibraryVersionNumber();
}

// ---------------------------------------------------------------------------
//		 Tracer::IsLibraryLoaded
// ---------------------------------------------------------------------------

bool Tracer::IsLibraryLoaded(void)
{
	return sConnectionID != NULL;
}
