/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 2000-2001 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"


#include "EmSPISlaveSTMicro.h"

#include "EmLowMem.h"			// EmLowMem_GetGlobal
#include "EmMemory.h"			// CEnableFullAccess
#include "EmPalmStructs.h"		// EmAliasSysBatteryDataStruct

#include "M68VZ328Hwr.h"
#include "EmRegsVZPrv.h"

#include "EmSession.h"			// GetDevice
#include "AlphaSmartScanCodes.h"


 
#include "Logging.h"
#define PRINTF	if (1) ; else LogAppendMsg

	// commands received by ST during reset
#define STCommandSleepWhileConnected	0xED
#define STCommandEmergencySleep			0xEE
#define STCommandGoBackToSleep			0xEF
#define STCommandNOP					0x00

#define StateAcceptKeys					0
#define StateSendingKeyDowns			1
#define StateSendingKeyUps				2

// ---------------------------------------------------------------------------
//		 EmSPISlaveSTMicro::EmSPISlaveSTMicro
// ---------------------------------------------------------------------------

EmSPISlaveSTMicro::EmSPISlaveSTMicro (	void ) :
	EmSPISlave ()
{

	fScanCodeListCount = 0;
	fScanCodeNextIndex = 0;

	memset(rowData, 0, kAlphaSmartNumButtonCols);
	memset(rowDataSendKeep, 0, kAlphaSmartNumButtonCols);
	
	shiftButtonDown = false;
	fInterruptState = false;
	fDoingReset = false;

	fSendVersion = false;
	fVersionStringIndex = 0;
	fCmdStringIndex = 0;
	fkeyState = StateAcceptKeys;


		// on dana we want all the keys except the control-xxxx
		// which are shortcuts for poser
	gSession->SetLeaveKeysInQueue(true);


}


// ---------------------------------------------------------------------------
//		 EmSPISlaveSTMicro::~EmSPISlaveSTMicro
// ---------------------------------------------------------------------------

EmSPISlaveSTMicro::~EmSPISlaveSTMicro (void)
{
	fInterruptState = false;
	gSession->SetLeaveKeysInQueue(false);
}


// ---------------------------------------------------------------------------
//		 EmSPISlaveSTMicro::DoExchange
// ---------------------------------------------------------------------------

uint16 EmSPISlaveSTMicro::DoExchange (uint16 control, uint16 data)
{
	uint16	result = 0;

		// check to see if we are being reset
	if (fDoingReset)
	{
		fScanCodeListCount = 0;
		fScanCodeNextIndex = 0;
		fDoingReset = false;
		return 0x00AA;		// special reset indicator
	}

		// are we receiving the version command which
		// is a string of data, once we verify it is
		// the version command then client will ask for
		// data with 0 for data
	if (data || fSendVersion)
	{
		result = CheckForVersion(data);
		fScanCodeListCount = 0;
		fScanCodeNextIndex = 0;
		return result;
	}

	if (fScanCodeListCount == 0)
	{
			// a little state machine because
			// keys from keyboard must go down
			// then up.  This is to fix a problem
			// when the same key goes down two time
			// We need to make sure the 68k sees
			// the key go up and then down again.
		switch(fkeyState)
		{
			case StateAcceptKeys:
				break;
			case StateSendingKeyDowns:
				fkeyState = StateSendingKeyUps;
				break;
			case StateSendingKeyUps:
				fkeyState = StateAcceptKeys;
				break;
			default:
				break;

		}
		// go build scan list
		BuildScancodeArray();
		fScanCodeNextIndex = 0;
		if (fScanCodeListCount == 0)
		{
			result = 0;
		}
		else
		{
			result = fScanCodeListCount | 0x40;
		}
			// turn off the interrupt here even
			// though we should only turn it off
			// when we have no more keys down
			// The problem is that when the modifiers
			// are down the 68k will just stay in
			// in the interrup routine forever
			// so this gives it a little breathing
			// room.
		fInterruptState = false;
		if (fkeyState == StateSendingKeyUps)
		{
			fkeyState = StateAcceptKeys;
		}

		return result;
	}

	result = fScanCodeArray[fScanCodeNextIndex];

	++fScanCodeNextIndex;

		// have we sent all the scancodes in the array
	if (fScanCodeNextIndex == fScanCodeListCount)
	{
		fScanCodeNextIndex = 0;
		fScanCodeListCount = 0;
	}

	return result;
	
}

// ---------------------------------------------------------------------------
//		 EmSPISlaveSTMicro::CheckForVersion
// ---------------------------------------------------------------------------

#define VersionLength	4
#define CommandLength	8

uint16 EmSPISlaveSTMicro::CheckForVersion (uint16 data)
{

	UInt8			*pCmdString;
	UInt8			*pVersionString;
	uint16			result = 0;
	UInt8			cmdChar;

	pVersionString = (UInt8 *)"06.3";	// send to key mgr
	pCmdString = (UInt8 *)"Get Ver!";	// string that key mgr will send
										// to find the version
	if (fSendVersion)
	{
		result = (UInt8 )pVersionString[fVersionStringIndex];
		++fVersionStringIndex;
		if (fVersionStringIndex == VersionLength)
		{
			fSendVersion = false;
			fVersionStringIndex = 0;
			fCmdStringIndex = 0;
		}
	}
	else	// not sending version, see if receiving version command
	{
		cmdChar = (UInt8 )data;
		if (pCmdString[fCmdStringIndex] == cmdChar)
		{
			++fCmdStringIndex;
			if (fCmdStringIndex == CommandLength)
			{
				fSendVersion = true;
				fVersionStringIndex = 0;
			}

		}
		else	// no match so just go back to 0
		{
			fCmdStringIndex = 0;
		}

	}


	return result;
}



// ---------------------------------------------------------------------------
//		 EmSPISlaveSTMicro::BuildScancodeArray
//
// Go through the row data bits and see what scancode we need to put
// in our list.  Keys pressed from the keypad instead of a skin button
// need to be clear after putting them in our list.  That is communicated
// in the fRowDataSendKeep array.  If the bit is sent then send the key
// and don't clear the keydown bit in fRowData.  If it is a keypad key
// then the bit in the fRowSendKeep array will be 0 so then we clear it
// in the fRowData array.
// ---------------------------------------------------------------------------

void EmSPISlaveSTMicro::BuildScancodeArray (void)
{
	uint8	scanCode;
	uint8	iubNewKeys, vubBitCounter;
	uint8	vubBitUnderTest;
	uint8	iubColumnNumber;

	if (gSession->HasButtonEvent ())
	{
		HandleButtonToScanCode ();
	}

	if (gSession->HasKeyEvent())
	{			
		EmKeyEvent	event = gSession->PeekKeyEvent ();
		if (event.fControlDown == false)	// shortcuts for poser, we don't handle
		{
			HandleKeyToScanCode();	// must change this back 
		}
	}
			
		// now build a scancode list from the keys		
	for (iubColumnNumber = 0; iubColumnNumber != kAlphaSmartNumButtonCols; ++iubColumnNumber)
	{
		iubNewKeys = rowData[iubColumnNumber];

		vubBitCounter = 0;
		
		while (iubNewKeys)
		{
			vubBitUnderTest = 1 << vubBitCounter;
			if (iubNewKeys & vubBitUnderTest)
			{
					// Column is LSB portion of the icode
				scanCode = iubColumnNumber;	
					// row is the MSB portion of the icode
				scanCode |= (vubBitCounter << 4);
				fScanCodeArray[fScanCodeListCount] = scanCode;
				++fScanCodeListCount;

				EmAssert(fScanCodeListCount < MaxScanCodeArray);
				
					// should we clear this key after we send it
				if ((rowDataSendKeep[iubColumnNumber] & vubBitUnderTest) == 0)
				{
					rowData[iubColumnNumber] &= ~vubBitUnderTest;
				}

			}	
			
				// remove the tested bit
			iubNewKeys = iubNewKeys & ~vubBitUnderTest;
			++vubBitCounter;
		}
	
	
	}


}

// ---------------------------------------------------------------------------
//		 EmSPISlaveSTMicro::IRQ1Assert
// ---------------------------------------------------------------------------

bool EmSPISlaveSTMicro::IRQ1Assert (void)
{
	if (fInterruptState == true)
	{
		return true;	
	}

	bool foundKeyEvent = false;

		// check if there is a key event waiting for us
	if (gSession->HasKeyEvent())
	{
		EmKeyEvent	event = gSession->PeekKeyEvent ();
			// we don't want to steal a control-? which is for
			// a poser shortcut
		if (event.fControlDown)
		{
			foundKeyEvent = false;
		}
		else
		{
			foundKeyEvent = true;
		}
		
	}

	if (gSession->HasButtonEvent () || foundKeyEvent)
	{
		fInterruptState = true;
	}
	else
	{
		fInterruptState = false;
	}


	return fInterruptState;
}


// ---------------------------------------------------------------------------
//		 EmSPISlaveSTMicro::SetResetMode
// ---------------------------------------------------------------------------

void EmSPISlaveSTMicro::SetResetMode (void)
{
	fDoingReset = true;


}


// ---------------------------------------------------------------------------
//		 EmSPISlaveSTMicro::HandleKeyToScanCode
// ---------------------------------------------------------------------------

void EmSPISlaveSTMicro::HandleKeyToScanCode (void)
{
	uint8	scanCode;

	if (fkeyState != StateAcceptKeys)
	{
		return;
	}

 		
	EmKeyEvent	event = gSession->PeekKeyEvent ();
	
	switch (event.fKey)
	{
		case 'A':
		case 'a':
			scanCode = SC_A;
			break;
		case 'B':
		case 'b':
			scanCode = SC_B;
			break;
		case 'C':
		case 'c':
			scanCode = SC_C;
			break;
		case 'D':
		case 'd':
			scanCode = SC_D;
			break;
		case 'e':
		case 'E':
			scanCode = SC_E;
			break;
		case 'f':
		case 'F':
			scanCode = SC_F;
			break;
		case 'g':
		case 'G':
			scanCode = SC_G;
			break;
		case 'h':
		case 'H':
			scanCode = SC_H;
			break;
		case 'I':
		case 'i':
			scanCode = SC_I;
			break;
		case 'j':
		case 'J':
			scanCode = SC_J;
			break;
		case 'k':
		case 'K':
			scanCode = SC_K;
			break;
		case 'l':
		case 'L':
			scanCode = SC_L;
			break;
		case 'm':
		case 'M':
			scanCode = SC_M;
			break;
		case 'n':
		case 'N':
			scanCode = SC_N;
			break;
		case 'o':
		case 'O':
			scanCode = SC_O;
			break;
		case 'p':
		case 'P':
			scanCode = SC_P;
			break;
		case 'Q':
		case 'q':
			scanCode = SC_Q;
			break;
		case 'R':
		case 'r':
			scanCode = SC_R;
			break;
		case 's':
		case 'S':
			scanCode = SC_S;
			break;
		case 't':
		case 'T':
			scanCode = SC_T;
			break;
		case 'u':
		case 'U':
			scanCode = SC_U;
			break;
		case 'v':
		case 'V':
			scanCode = SC_V;
			break;
		case 'W':
		case 'w':
			scanCode = SC_W;
			break;
		case 'X':
		case 'x':
			scanCode = SC_X;
			break;
		case 'y':
		case 'Y':
			scanCode = SC_Y;
			break;
		case 'z':
		case 'Z':
			scanCode = SC_Z;
			break;
		case '1':
		case '!':
			scanCode = SC_1;
			break;
		case '2':
		case '@':
			scanCode = SC_2;
			break;
		case '3':
		case '#':
			scanCode = SC_3;
			break;
		case '4':
		case '$':
			scanCode = SC_4;
			break;
		case '5':
		case '%':
			scanCode = SC_5;
			break;
		case '6':
		case '^':
			scanCode = SC_6;
			break;
		case '7':
		case '&':
			scanCode = SC_7;
			break;
		case '8':
		case '*':
			scanCode = SC_8;
			break;
		case '9':
		case '(':
			scanCode = SC_9;
			break;
		case '0':
		case ')':
			scanCode = SC_ZERO;
			break;
		case '`':
		case '~':
			scanCode = SC_TILDE;
			break;
		case '-':
		case '_':
			scanCode = SC_DASH;
			break;
		case '=':
		case '+':
			scanCode = SC_PLUS;
			break;
		case '{':
		case '[':
			scanCode = SC_LEFT_BRACKET;
			break;
		case '}':
		case ']':
			scanCode = SC_RIGHT_BRACKET;
			break;
		case '|':
		case 0x5C:
			scanCode = SC_FORWARD_SLASH;
			break;
		case ';':
		case ':':
			scanCode = SC_SEMICOLON;
			break;
		
		case 0x1D:
			scanCode = SC_RIGHT_ARROW;
			break;
		case 0x1E:
			scanCode = SC_UP_ARROW;
			break;
		case 0x1C:
			scanCode = SC_LEFT_ARROW;
			break;
		case 0x1F:
			scanCode = SC_DOWN_ARROW;
			break;

		case ' ':
			scanCode = SC_SPACE;
			break;
		case '<':
		case ',':
			scanCode = SC_COMMA;
			break;
		case '>':
		case '.':
			scanCode = SC_PERIOD;
			break;
		case '?':
		case '/':
			scanCode = SC_BACK_SLASH;
			break;
		case 0x22:	// "
		case 0x27:	// '
			scanCode = SC_DOUBLE_QUOTE;
			break;

		case 0x09:	// TAB
			scanCode = SC_TAB;
			break;
		case 0x08:	// BACKSPACE
			scanCode = SC_BACKSPACE;
			break;
		case 0x0A:	// ENTER/RETURN
			scanCode = SC_ENTER;
			break;
		default:						
			//EmAssert (false);
			//scanCode = 0;
			//scanCode = 0x1D;
			return;
			break;
	}

		// is the key already set in the rowData
	if (rowData[scanCode & 0x0F] & (1 << ((scanCode & 0xF0) >> 4)))
	{
		return;	// just leave it on the queue
	}

	
		// pull the event off the queue
	event = gSession->GetKeyEvent ();


		// bits will be cleared by ST Micro
	rowData[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
	rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));

	if (shiftButtonDown == false)
	{

		scanCode = SC_LEFT_SHIFT;
		if (event.fShiftDown)
		{
			rowData[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
		}
		else
		{
			rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
		}

			// shift is shared between key and skin button
		rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));

	}


	fkeyState = StateSendingKeyDowns;


}

// ---------------------------------------------------------------------------
//		 EmSPISlaveSTMicro::HandleButtonToScanCode
// ---------------------------------------------------------------------------

void EmSPISlaveSTMicro::HandleButtonToScanCode (void)
{
	uint8	scanCode = 0xFF;
	
	EmButtonEvent	event = gSession->GetButtonEvent ();
		
	switch (event.fButton)
	{
		case kElement_None:				
			break;
		case kElement_PowerButton:		
			scanCode = SC_POWER;	
			break;
		case kElement_F1:		
			scanCode = SC_F1;	
			break;
		case kElement_F2:		
			scanCode = SC_F2;	
			break;
		case kElement_F3:		
			scanCode = SC_F3;	
			break;
		case kElement_F4:		
			scanCode = SC_F4;	
			break;
		case kElement_F5:		
			scanCode = SC_F5;	
			break;
		case kElement_F6:		
			scanCode = SC_F6;	
			break;
		case kElement_F7:		
			scanCode = SC_F7;	
			break;
		case kElement_F8:		
			scanCode = SC_F8;	
			break;
		case kElement_F9:		
		case kElement_App1Button:		
			scanCode = SC_F9;	
			break;
		case kElement_F10:		
		case kElement_App2Button:		
			scanCode = SC_F10;	
			break;
		case kElement_F11:		
		case kElement_App3Button:		
			scanCode = SC_F11;	
			break;
		case kElement_F12:		
		case kElement_App4Button:		
			scanCode = SC_F12;	
			break;
		case kElement_F13:		
			scanCode = SC_F13;	
			break;
		case kElement_F14:		
			scanCode = SC_F14;	
			break;
		case kElement_F15:		
			scanCode = SC_F15;	
			break;
		case kElement_F16:		
			scanCode = SC_F16;	
			break;

		case kElement_UpButton:		
			scanCode = SC_UP_ARROW;	
			break;
		case kElement_DownButton:		
			scanCode = SC_DOWN_ARROW;	
			break;
		case kElement_CapsLock:		
			scanCode = SC_CAPS_LOCK;	
			break;
		case kElement_PcDelete:		
			scanCode = SC_PC_DELETE;	
			break;
		case kElement_Escape:		
			scanCode = SC_ESC;	
			break;


		case kElement_Control:		
			scanCode = SC_CONTROL;
			if (event.fButtonIsDown)
			{
					// toggle bit
				if (rowData[scanCode & 0x0F] & (1 << ((scanCode & 0xF0) >> 4)))
				{
					rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
					rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
				}
				else
				{
					rowData[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
					rowDataSendKeep[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
				}
			}
			return;
			break;
		case kElement_LeftAlt:		
			scanCode = SC_LEFT_OPTION;	
			if (event.fButtonIsDown)
			{
					// toggle bit
				if (rowData[scanCode & 0x0F] & (1 << ((scanCode & 0xF0) >> 4)))
				{
					rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
					rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
				}
				else
				{
					rowData[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
					rowDataSendKeep[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
				}
			}
			return;
			break;
		case kElement_LeftShift:		
			scanCode = SC_LEFT_SHIFT;	
			if (event.fButtonIsDown)
			{
					// toggle bit
				if (rowData[scanCode & 0x0F] & (1 << ((scanCode & 0xF0) >> 4)))
				{
					rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
					rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
					shiftButtonDown = false;
				}
				else
				{
					rowData[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
					rowDataSendKeep[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
					shiftButtonDown = true;
				}
			}
			return;
			break;
		case kElement_RightShift:		
			scanCode = SC_RIGHT_SHIFT;	
			if (event.fButtonIsDown)
			{
					// toggle bit
				if (rowData[scanCode & 0x0F] & (1 << ((scanCode & 0xF0) >> 4)))
				{
					rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
					rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
					shiftButtonDown = false;
				}
				else
				{
					rowData[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
					rowDataSendKeep[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
					shiftButtonDown = true;
				}
			}
			return;
			break;
		case kElement_Command:		
			scanCode = SC_COMMAND;
			if (event.fButtonIsDown)
			{
					// toggle bit
				if (rowData[scanCode & 0x0F] & (1 << ((scanCode & 0xF0) >> 4)))
				{
					rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
					rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
				}
				else
				{
					rowData[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
					rowDataSendKeep[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
				}
			}
			return;
			break;
		case kElement_ClearModifiers:
				// clear modify bits
			scanCode = SC_CONTROL;	
			rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			scanCode = SC_LEFT_OPTION;	
			rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			scanCode = SC_LEFT_SHIFT;	
			rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			scanCode = SC_RIGHT_SHIFT;	
			rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			scanCode = SC_COMMAND;	
			rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
			return;
			break;
		default:						
			EmAssert (false);
			scanCode = 0;
			break;
	}


		// clear or set the bit that represents the key in our key matrix
		// scan code = upper nibble is row bit number + lower nibble is column number
	if (event.fButtonIsDown)
	{
		rowData[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
		//if (scanCode != SC_POWER)
		if (1)
		{
			rowDataSendKeep[scanCode & 0x0F] |= (1 << ((scanCode & 0xF0) >> 4));
		}
		if ((scanCode == SC_DOWN_ARROW) || (scanCode == SC_UP_ARROW))
		{
			rowData[SC_LEFT_OPTION & 0x0F] |= (1 << ((SC_LEFT_OPTION & 0xF0) >> 4));
			rowDataSendKeep[SC_LEFT_OPTION & 0x0F] |= (1 << ((SC_LEFT_OPTION & 0xF0) >> 4));
		}
	}
	else
	{
		rowData[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
		rowDataSendKeep[scanCode & 0x0F] &= (~(1 << ((scanCode & 0xF0) >> 4)));
		if ((scanCode == SC_DOWN_ARROW) || (scanCode == SC_UP_ARROW))
		{
			rowData[SC_LEFT_OPTION & 0x0F] &= (~(1 << ((SC_LEFT_OPTION & 0xF0) >> 4)));
			rowDataSendKeep[SC_LEFT_OPTION & 0x0F] &= (~(1 << ((SC_LEFT_OPTION & 0xF0) >> 4)));
		}
	}


}



